/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
	This file implements the setuid-root helper tool.
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <CommonCrypto/CommonDigest.h>
#include <SystemConfiguration/SCDynamicStoreCopySpecific.h>
#include "GBDebug.h"

@interface AuthHelperTool : NSObject {
@private
	NSString* _selfRepairPlistPath;
	NSString* _mainBundlePath;
}

- (instancetype)init;

- (void)exitWithError:(NSString*)msg;

@end

@implementation AuthHelperTool

- (instancetype)init
{
	if (!(self = [super init]))
		return nil;
	NSString *_selfRepairJobLabel = @"org.glimmerblocker.self-repair";
	_selfRepairPlistPath = [NSString stringWithFormat:@"/Library/LaunchDaemons/%@.plist", _selfRepairJobLabel];
	_mainBundlePath = @"/Library/PreferencePanes/GlimmerBlocker.prefPane";
	return self;
}


- (void)exitWithError:(NSString*)msg
{
	NSLog(@"AuthHelper failed: %@", msg);
	[NSException raise:msg format:@"AuthHelperFormat"];
}

- (void)readBytes:(size_t)numBytesNeeded
	  readingWhat:(NSString*)what
	   withBuffer:(char*)buffer
{
	size_t numBytesRead = 0;
	while (numBytesRead < numBytesNeeded) {
		if (feof(stdin))
			[self exitWithError:[NSString stringWithFormat:@"EOF on stdin while reading %@", what]];
		size_t x = fread(buffer + numBytesRead, 1, numBytesNeeded - numBytesRead, stdin);
		numBytesRead += x;
		if (ferror(stdin))
			[self exitWithError:[NSString stringWithFormat:@"error on stdin while reading %@", what]];
	}
}

- (NSData*)readData:(NSString*)description
{
	long long len = 0;
	NSAssert(sizeof(len) == 8, @"longlong != 64");
	[self readBytes:sizeof(len) readingWhat:description withBuffer:(char*)&len];
	if (len > 128 * 1024 * 1024) {
		NSString* fmt = @"Invalid NSData chunk size in stdin: %lld bytes";
		[self exitWithError:[NSString stringWithFormat:fmt, len]];
	}
	char* buffer = malloc(len);
	if (!buffer)
		[self exitWithError:[NSString stringWithFormat:@"Out of memory, can't malloc %lld bytes", len]];
	[self readBytes:len readingWhat:description withBuffer:buffer];
	NSData* data = [NSData dataWithBytes:buffer length:len];
	free(buffer);
	return data;
}

- (void)readAuth
{
	NSData* data = [self readData:@"AuthorizationExternalForm"];
	AuthorizationExternalForm extAuth;
	NSAssert([data length] == sizeof(extAuth), @"AuthDataSize");
	memcpy(&extAuth, [data bytes], sizeof(extAuth));
	AuthorizationRef authRef = NULL;
	OSStatus stat = AuthorizationCreateFromExternalForm(&extAuth, &authRef);
	if (stat != errAuthorizationSuccess)
		[self exitWithError:@"Authorization invalid/expired"];
	DebugNSLog(@"AuthHelper.auth ok");
}

- (int)executeCommandlineTool:(NSString*)cmd
					 withArgs:(NSArray*)args
		 withFatalDescription:(NSString*)fatal
{
	int status = [GBDebug executeCommandlineTool:cmd withArgs:args];
	if (status && fatal) {
		NSString* s = args ? [NSString stringWithFormat:@"%@ %@", cmd, [args componentsJoinedByString:@" "]] : cmd;
		NSLog(@"Got exit code %d: %@", status, s);
		[self exitWithError:fatal];
	}
	return status;
}

- (void)savePlist:(NSString*)xml
		   inPath:(NSString*)path
{
	NSFileManager* fm = [NSFileManager defaultManager];
	NSData* data = [xml dataUsingEncoding:NSUTF8StringEncoding];
	if ([[NSData dataWithContentsOfFile:path] isEqual:data]) {
		DebugNSLog(@"AuthHelper.write-plist: no changed.");
	} else {
		NSDictionary* attrs = @{NSFilePosixPermissions: @0644L};
		BOOL b = [fm createFileAtPath:path contents:data attributes:attrs];
		if (!b)
			[self exitWithError:@"Can't create plist"];
		DebugNSLog(@"AuthHelper.write-plist ok");
	}
	@try {
		[self executeCommandlineTool:@"launchctl"
							withArgs:@[@"load", path]
				withFatalDescription:@"launchctl won't load plist."];
	}
	@catch (NSException* ex) {
		[fm removeItemAtPath:path error:NULL];
		@throw ex;
	}
}

- (void)zapPlist:(NSString*)path
{
	NSFileManager* fm = [NSFileManager defaultManager];
	if (![fm fileExistsAtPath:path])
		return;
	[self executeCommandlineTool:@"launchctl"
						withArgs:@[@"unload", path]
			withFatalDescription:NULL];
	[fm removeItemAtPath:path error:NULL];
	if ([fm fileExistsAtPath:path])
		[self exitWithError:@"Failed deleting launchd plist file"];
}

- (NSString*)makeUpdaterPlistContents:(NSArray*)args
{
	NSFileManager* fm = [NSFileManager defaultManager];
	long secs;
	if ([fm fileExistsAtPath:CHANNEL_DEV_PATH])
		secs = 10 * 60 ;
	else if ([fm fileExistsAtPath:CHANNEL_BETA_PATH])
		secs =  18 * 60 * 60;
	else
		secs =  2 * 24 * 60 * 60;
	return [NSString stringWithFormat:@""
			"<?xml version='1.0' encoding='UTF-8'?>\n"
			"<!DOCTYPE plist PUBLIC '-//Apple Computer//DTD PLIST 1.0//EN' 'http://www.apple.com/DTDs/PropertyList-1.0.dtd'>\n"
			"<plist version='1.0'>\n"
			"<dict>\n"
			"   <key>Disabled</key>\n"
			"   <false/>\n"
			//
			"   <key>Label</key>\n"
			"   <string>%@</string>\n"
			//
			"   <key>ProgramArguments</key>\n"
			"   <array>\n"
			"   	<string>%@/Contents/MacOS/InstallerDaemon</string>\n"
			"   </array>\n"
			//
			"   <key>Umask</key>\n"
			"   <integer>18</integer>\n" // 022, files not writeable by group/other.
			//
			"   <key>ProcessType</key>\n"
			"   <string>Background</string>\n"
			//
			"   <key>StartInterval</key>\n"
			"   <integer>%ld</integer>\n"
			//
			"   <key>KeepAlive</key>\n"
			"   <false/>\n"
			//
			"   <key>WorkingDirectory</key>\n"
			"   <string>/tmp</string>\n"
			//
			//"   <key>StandardErrorPath</key>\n"
			//"   <string>/Library/Logs/GlimmerBlocker/Launchd.log</string>\n"
			//
			"</dict>\n"
			"</plist>\n",
			UPDATER_LAUNCHD_ID,
			[GBDebug updaterBundlePath],
			secs
			];
}

- (NSString*)makeProxyPlistContents
{
	return [NSString stringWithFormat:@""
			"<?xml version='1.0' encoding='UTF-8'?>\n"
			"<!DOCTYPE plist PUBLIC '-//Apple Computer//DTD PLIST 1.0//EN' 'http://www.apple.com/DTDs/PropertyList-1.0.dtd'>\n"
			"<plist version='1.0'>\n"
			"<dict>\n"
			"   <key>Disabled</key>\n"
			"   <false/>\n"
			"   <key>Label</key>\n"
			"   <string>%@</string>\n"
			"   <key>UserName</key>\n"
			"   <string>_glimmerblocker</string>\n"
			"   <key>InitGroups</key>\n"
			"   <true/>\n"
			"   <key>ProgramArguments</key>\n"
			"   <array>\n"
			"   	<string>%@/Contents/GlimmerBlockerProxy.app/Contents/MacOS/GlimmerBlockerProxy</string>\n"
			"       %@\n"
			"   </array>\n"
			"   <key>OnDemand</key>\n"
			"   <false/>\n"
			"   <key>Umask</key>\n"
			"   <integer>63</integer>\n" // 077, files only readable by _glimmerblocker
			"   <key>KeepAlive</key>\n"
			"   <true/>\n"
			"   <key>WorkingDirectory</key>\n"
			"   <string>%@/Contents/GlimmerBlockerProxy.app/Contents/Resources</string>\n"
			"   <key>StandardErrorPath</key>\n"
			"   <string>/Library/Logs/GlimmerBlocker/Launchd.log</string>\n"
			"</dict>\n"
			"</plist>\n",
			PROXY_LAUNCHD_ID,
			_mainBundlePath,
			[GBDebug fingerprintOptionForInstalledPanel:[NSBundle mainBundle]],
			_mainBundlePath];
}

- (void)stop:(NSString*)plistPath
{
	[self readAuth];
	// launchctl doesn't return error code if unloading fails....
	[self zapPlist:plistPath];
	[self zapPlist:_selfRepairPlistPath];
}

- (void)start:(NSString*)plistPath  withContents:(NSString*)contents
{
	[self readAuth];
	[self zapPlist:plistPath];
	[self savePlist:contents inPath:plistPath];
	//
	[self zapPlist:_selfRepairPlistPath];
}

- (NSString*)makeDir:(NSString*)path
		   withPerms:(int)perms
{
	NSFileManager* fm = [NSFileManager defaultManager];
	BOOL isDirectory = NO;
	if ([fm fileExistsAtPath:path isDirectory:&isDirectory]) {
		if (!isDirectory) {
			NSString* msg = [NSString stringWithFormat:@"Has non-directory item when expected directory: %@", path];
			[NSException raise:msg format:@"AuthHelperFormat"];
		}
		return path;
	}
	NSDictionary* attrs = @{NSFilePosixPermissions: [NSNumber numberWithLong:perms]};
	NSError *__autoreleasing error = NULL;
	if (![fm createDirectoryAtPath:path withIntermediateDirectories:YES attributes:attrs error:&error]) {
		NSString* msg = [NSString stringWithFormat:@"Can't create directory (%@): %@", error, path];
		[NSException raise:msg format:@"AuthHelperFormat"];
	}
	NSArray* args = @[@"_glimmerblocker:_glimmerblocker", path];
	[self executeCommandlineTool:@"chown"
						withArgs:args
			withFatalDescription:@"chmod of directory."];
	return path;
}

- (NSString*)makePrefsDir
{
	return [self makeDir:@"/Library/GlimmerBlocker" withPerms:0755];
}

- (void)savePrefs:(NSArray*)args
{
	[self readAuth];
	NSData* prefsData = [self readData:@"Prefs data"];
	NSData* expectedMD5 = [self readData:@"Prefs MD5"];
	// checking MD5 makes sure we got all of the data.
	NSData* actualMD5 = [GBDebug md5:prefsData];
	if (![actualMD5 isEqual:expectedMD5])
		[self exitWithError:[NSString stringWithFormat:@"Prefs data MD5 = %@ but expected %@", actualMD5, expectedMD5]];
	DebugNSLog(@"Got valid prefs");
	NSString* path = [NSString stringWithFormat:@"%@/PanelSettings.xml", [self makePrefsDir]];
	[GBDebug safeSaveFileAtPath:path contents:prefsData];
	NSFileManager* fm = [NSFileManager defaultManager];
	NSDictionary* attrs = @{NSFilePosixPermissions: @0644L};
	[fm setAttributes:attrs ofItemAtPath:path error:NULL];
	[self executeCommandlineTool:@"chown"
						withArgs:@[@"_glimmerblocker:_glimmerblocker", path]
			withFatalDescription:@"chmod of panel-prefs."];
	//
	for (NSString *arg in args) {
		if ([arg isEqualToString:@"--beta"]) {
			if ([fm fileExistsAtPath:CHANNEL_BETA_PATH])
				continue;
			if (![[NSData data] writeToFile:CHANNEL_BETA_PATH atomically:YES])
				NSLog(@"Can't create %@", CHANNEL_BETA_PATH);
			continue;
		}
		if ([arg isEqualToString:@"--final"]) {
			if (![fm fileExistsAtPath:CHANNEL_BETA_PATH])
				continue;
			NSError *error = NULL;
			[fm removeItemAtPath:CHANNEL_BETA_PATH error:&error];
			if (error)
				NSLog(@"Can't remove %@ due to %@", CHANNEL_BETA_PATH, error);
			continue;
		}
		NSLog(@"Unknown/unsupported --save-prefs argument: %@", arg);
	}
}

- (BOOL)isUpdaterAlreadyRunning
{
	for (NSRunningApplication* app in [[NSWorkspace sharedWorkspace] runningApplications]) {
		NSString* bid = [app bundleIdentifier];
		if ([bid isEqual:@"org.glimmerblocker.upgradeApp"])
			return YES;
	}
	return NO;
}

- (BOOL)isAdmin
{
	NSArray* args = @[@"-c", @"groups | perl -ne 'exit 42 if /\\badmin\\b/;'"];
	int status = [self executeCommandlineTool:@"/bin/sh" withArgs:args withFatalDescription:NULL];
	//NSLog(@"isAdmin: %d", status);
	return status == 42;
}

- (int)runUpgradeApp:(NSArray*)args
{
	if ([self isUpdaterAlreadyRunning]) {
		NSLog(@"Ignores notification helper request as updater already is running.");
		exit(501); // see UpdateDaemon.java for exit codes
	}
	// http://developer.apple.com/qa/qa2001/qa1133.html
	// user-id, message
	uid_t uid = -1;// = [userIdText intValue];
	NSString* userName = (NSString*)CFBridgingRelease(SCDynamicStoreCopyConsoleUser(NULL, &uid, NULL));
	DebugNSLog(@"Current user: %d = %@", uid, userName);
	if (uid <= 0 || ![userName length] || [userName isEqual:@"loginwindow"]) {
		// loginwindow might be 10.4 stuff, no such user in 10.5
		NSLog(@"Seems to have no logged-in user.");
		exit(502); // see UpdateDaemon.java for exit codes
	}
	if (setuid(uid) < 0)
		NSLog(@"Failed setting uid to %d", uid);
	if (getuid() != uid)
		[self exitWithError:[NSString stringWithFormat:@"Can't change uid to %d, is still %d", uid, getuid()]];
	if (REQUIRED_VALID_CODESIGN && ![GBDebug verifyCodesign])
		exit(503); // see UpdateDaemon.java for exit codes
	if (![self isAdmin])
		exit(504); // see UpdateDaemon.java for exit codes
	//
	// AuthorizationCopyRights succeeds only when the application is starting using the "open" command
	// instead of being started directly.
	//
	NSMutableArray* openArgs = [NSMutableArray arrayWithCapacity:10];
	[openArgs addObject:@"-n"]; // open new instance
	[openArgs addObject:@"-W"]; // wait until app quit.
	[openArgs addObject:@"-g"]; // open in background
	[openArgs addObject:@"-a"]; // open with app
	[openArgs addObject:[NSString stringWithFormat:@"%@/Contents/MacOS/GlimmerBlockerUpdater", [GBDebug updaterBundlePath]]];
	[openArgs addObjectsFromArray:args];
	int status = [self executeCommandlineTool:@"open" withArgs:openArgs withFatalDescription:NULL];
	return status;
}

- (void)checkForUpdates:(NSArray*)args
{
	NSString* resourcesPath = [NSString stringWithFormat:@"%@/Contents/Resources", [GBDebug updaterBundlePath]];
	NSMutableArray* openArgs = [NSMutableArray arrayWithCapacity:5];
	for (NSString* s in args) {
		if ([s isEqual:@"--background"])
			[openArgs addObject:[NSString stringWithFormat:@"%@/option-background.txt", resourcesPath]];
		else if ([s isEqual:@"--beta"])
			[openArgs addObject:[NSString stringWithFormat:@"%@/option-beta.txt", resourcesPath]];
		else if (![s isEqual:@"--foreground"] && ![s isEqual:@"--final"])
			NSLog(@"Unknown sub-option for --check-for-updates: '%@'", s);
	}
	int status = [self runUpgradeApp:openArgs];
	exit(status);
}

- (void)askFilterUpdate:(NSArray*)args
{
	NSMutableArray* openArgs = [NSMutableArray arrayWithCapacity:5];
	[openArgs addObject:args[0]];
	int status = [self runUpgradeApp:openArgs];
	NSLog(@"Ask filter return status: %d", status);
	exit(status);
}

- (void)javaSettings
{
	[self readAuth];
	NSURL* url = [NSURL fileURLWithPath:@"/Library/GlimmerBlocker/JavaSettings.xml"];
	NSData* data = [NSData dataWithContentsOfURL:url];
	NSFileHandle *stdoutHandle = [NSFileHandle fileHandleWithStandardOutput];
	[stdoutHandle writeData:data];
}

- (void)writeStringToStdout:(NSString*)s
{
	NSData* data = [s dataUsingEncoding:NSUTF8StringEncoding];
	NSFileHandle *stdoutHandle = [NSFileHandle fileHandleWithStandardOutput];
	[stdoutHandle writeData:data];
}

- (void)runWithArgs:(NSArray*)args
{
	if (geteuid()) {
		[self exitWithError:[NSString stringWithFormat:@"Not running as root but with uid = %d and euid = %d",
							 getuid(), geteuid()]];
	}
	if (getuid()) {
		if (setuid(0) < 0)
			NSLog(@"Failed setting uid to 0");
		if (getuid()) {
			[self exitWithError:[NSString stringWithFormat:@"Failed to set uid to 0, is still %d, euid = %d",
								 getuid(), geteuid()]];
		}
	}
	if ([args count] < 2)
		[self exitWithError:@"No arguments given"];
	NSString* cmd = args[1];
	args = [args subarrayWithRange:NSMakeRange(2, [args count] - 2)];
	if ([cmd isEqual:@"--plist-contents"])
		[self writeStringToStdout:[self makeProxyPlistContents]];
	else if ([cmd isEqual:@"--plist-path"])
		[self writeStringToStdout:[GBDebug proxyPlistPath]];
	else if ([cmd isEqual:@"--start"])
		[self start:[GBDebug proxyPlistPath] withContents:[self makeProxyPlistContents]];
	else if ([cmd isEqual:@"--stop"])
		[self stop:[GBDebug proxyPlistPath]];
	else if ([cmd isEqual:@"--updater-plist-path"])
		[self writeStringToStdout:[GBDebug updaterPlistPath]];
	else if ([cmd isEqual:@"--updater-plist-content"])
		[self writeStringToStdout:[self makeUpdaterPlistContents:args]];
	else if ([cmd isEqual:@"--updater-start"])
		[self start:[GBDebug updaterPlistPath] withContents:[self makeUpdaterPlistContents:args]];
	else if ([cmd isEqual:@"--updater-stop"])
		[self stop:[GBDebug updaterPlistPath]];
	else if ([cmd isEqual:@"--save-prefs"])
		[self savePrefs:args];
	else if ([cmd isEqual:@"--check-for-updates"])
		[self checkForUpdates:args];
	else if ([cmd isEqual:@"--ask-filter-update"])
		[self askFilterUpdate:args];
	else if ([cmd isEqual:@"--java-settings"])
		[self javaSettings];
	else
		[self exitWithError:[NSString stringWithFormat:@"Unknown arg: %@", cmd]];
}

@end

int main (int argc, const char * argv[])
{
	@autoreleasepool {
		NSArray *args = [[NSProcessInfo processInfo] arguments];
		AuthHelperTool* tool = [[AuthHelperTool alloc] init];
		@try {
			[tool runWithArgs:args];
		}
		@catch (NSException* ex) {
			if (![[ex reason] isEqual:@"AuthHelperFormat"])
				NSLog(@"Stops due to exception: '%@' / '%@'", [ex name], [ex reason]);
			exit(1);
		}
	}
	return 0;
}
