/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.connectionCache;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.List;
import java.util.TimeZone;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeUnit;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public abstract class ConnectionCache<INFO, CONN>
{
	@NotNull public final String cacheName;

	private final List<PrivateConnectionInfo> idleConnections = new ArrayList<>();
	private final List<PrivateConnectionInfo> busyConnections = new ArrayList<>();
	protected final int maxConnections;
	private int nextConnectionId = 100;
	private int numOpening;
	@SuppressWarnings("MismatchedQueryAndUpdateOfCollection")
	private final ArrayBlockingQueue<Object> connectionCachedNotificator = new ArrayBlockingQueue<>(1, true);

	public boolean debugFlag = false;
	public boolean shouldLogCreationAndDestroy = true;

	private volatile long oldestAllowedCreationTime = 0;

	protected ConnectionCache(@NotNull String cacheName,
	                          int maxConnections)
	{
		this.cacheName = cacheName;
		this.maxConnections = maxConnections;
		if (maxConnections <= 0)
			throw new IllegalArgumentException("maxConnections <= 0");
		if (debugFlag)
			log("Created");
	}

	public void execute(@NotNull ConnectionRunnable<CONN> r) throws Exception
	{
		@NotNull PrivateConnectionInfo info = get();
		try {
			r.run(info.connInfo.getConnection());
		}
		catch (Exception ex) {
			discardBusyConnection(info, "Exception while using");
			handleExceptionWhileUsingConnection(ex);
			throw ex;
		}
		catch (Throwable t) {
			discardBusyConnection(info, "Throwable while using");
			handleExceptionWhileUsingConnection(t);
			throw new Exception(t);
		}
		catchConnection(info);
	}

	public <T> T query(@NotNull ConnectionQuery<CONN, T> r) throws Exception
	{
		T obj;
		@NotNull PrivateConnectionInfo info = get();
		try {
			obj = r.run(info.connInfo.getConnection());
		}
		catch (Exception ex) {
			discardBusyConnection(info, "Exception while using");
			handleExceptionWhileUsingConnection(ex);
			throw ex;
		}
		catch (Throwable t) {
			discardBusyConnection(info, "Throwable while using");
			handleExceptionWhileUsingConnection(t);
			throw new Exception(t);
		}
		catchConnection(info);
		return obj;
	}

	public void execute(@NotNull TransientConnectionRunnable<INFO, CONN> r) throws Exception
	{
		r.notifyEnterScope(this);
		try {
			r.run();
			r.releaseConnection();
		}
		catch (Exception ex) {
			boolean hadConnection = r.hasConnection();
			r.discardBusyConnection();
			if (hadConnection)
				handleExceptionWhileUsingConnection(ex);
			throw ex;
		}
		catch (Throwable t) {
			boolean hadConnection = r.hasConnection();
			r.discardBusyConnection();
			if (hadConnection)
				handleExceptionWhileUsingConnection(t);
			throw new Exception(t);
		}
		finally {
			r.notifyLeaveScope();
		}
	}

	protected void handleExceptionWhileUsingConnection(@NotNull Throwable ex)
	{
		System.err.println("EXCEPTION WHILE USING CONNECTION:\n");
		ex.printStackTrace();
	}

	@NotNull
	protected abstract ConnectionInfo<CONN> createNewConnection() throws Exception;

	public void log(String s)
	{
		Calendar c = Calendar.getInstance();
		c.setTimeZone(TimeZone.getDefault());
		System.err.printf("====------>>>> %02d:%02d:%02d  %s: %s\n",
				c.get(Calendar.HOUR_OF_DAY), c.get(Calendar.MINUTE), c.get(Calendar.SECOND),
				cacheName, s);
	}

	public synchronized void debugDump(@NotNull String indent, @NotNull PrintWriter pw, boolean dumpCached)
	{
		for (PrivateConnectionInfo info : busyConnections) {
			String indent1 = String.format("%sBusy.%05d:%s", indent, info.id, (dumpCached ? "  " : ""));
			info.connInfo.debugDump(indent1, pw);
		}
		if (dumpCached) {
			for (PrivateConnectionInfo info : idleConnections) {
				String indent1 = String.format("%sCached.%05d: ", indent, info.id);
				info.connInfo.debugDump(indent1, pw);
			}
		}
	}

	@NotNull
	public String getCacheName()
	{
		return cacheName;
	}

	@SuppressWarnings({"UnusedDeclaration"})
	public synchronized int getNumCachedConnections()
	{
		return idleConnections.size();
	}

	public synchronized ConnectionCounts getConnectionCounts()
	{
		return new ConnectionCounts(busyConnections.size(), idleConnections.size(), numOpening);
	}

	@NotNull
	PrivateConnectionInfo get() throws Exception
	{
		long st0 = System.currentTimeMillis();
		while (true) {
			PrivateConnectionInfo info = popCachedHandle();
			if (info != null) {
				if (debugFlag)
					log("get -> id=" + info.id);
				return info;
			}
			boolean allowOpen;
			int numConnections;
			synchronized (this) {
				numConnections = busyConnections.size() + idleConnections.size() + numOpening;
				allowOpen = (numConnections < maxConnections);
				if (allowOpen)
					numOpening++;
			}
			if (!allowOpen) {
				if (System.currentTimeMillis() - st0 > 5000L)
					notifyTooManyConnectionsOpen(numConnections);
				connectionCachedNotificator.poll(1, TimeUnit.SECONDS);
				continue;
			}
			try {
				return createNewConnectionInfo();
			}
			finally {
				numOpening--;
			}

		}
	}

	long lastTooManyReport = 0;

	protected void notifyTooManyConnectionsOpen(int numConnections) throws Exception
	{
		long now = System.currentTimeMillis();
		if (now - lastTooManyReport > 5 * 60 * 1000L) {
			StringWriter sw = new StringWriter();
			PrintWriter pw = new PrintWriter(sw);
			debugDump("", pw, true);
			pw.flush();
			System.err.println("Connection cache state (" + cacheName + "):\n" + sw);
		}
		throw new SQLException(cacheName + ": Too many connections (" +
				numConnections + "/" + maxConnections + "). Please wait and try again later.");
	}

	final class PrivateConnectionInfo
	{
		/* package access for all attrs */
		final ConnectionInfo<CONN> connInfo;
		final int id;
		final long createTime;
		long cacheTime;

		PrivateConnectionInfo(int id, ConnectionInfo<CONN> connInfo)
		{
			this.id = id;
			this.connInfo = connInfo;
			this.createTime = System.currentTimeMillis();
		}
	}

	@NotNull
	private PrivateConnectionInfo createNewConnectionInfo() throws Exception
	{
		long st0 = System.currentTimeMillis();
		ConnectionInfo<CONN> info = createNewConnection();
		synchronized (this) {
			PrivateConnectionInfo privateInfo = new PrivateConnectionInfo(nextConnectionId++, info);
			long dur = System.currentTimeMillis() - st0;
			if (debugFlag)
				log("get -> created new: id=" + privateInfo.id);
			busyConnections.add(privateInfo);
			if (shouldLogCreationAndDestroy) {
				log("created." + privateInfo.id + " -> " + idleConnections.size() + " cached, "
						+ busyConnections.size() + " alive (" + dur + " msecs)");
			}
			return privateInfo;
		}
	}

	@Nullable
	private PrivateConnectionInfo popCachedHandle()
	{
		while (true) {
			PrivateConnectionInfo info;
			boolean ageIsOk;
			synchronized (this) {
				int index = idleConnections.size() - 1;
				if (index < 0)
					return null;
				info = idleConnections.get(index);
				idleConnections.remove(index);
				ageIsOk = (info.createTime >= oldestAllowedCreationTime);
			}
			if (ageIsOk) {
				if (info.connInfo.isReusable()) {
					synchronized (this) {
						busyConnections.add(info);
					}
					return info;
				}
				discardConnection(info, "Not reusable");
			} else {
				log("popCached." + info.id + ": Discards too old connection");
				discardConnection(info, "Too old");
			}
		}
	}

	public int markAllCurrentAsNonCacheable()
	{
		int numFlushed;
		List<PrivateConnectionInfo> discards;
		synchronized (this) {
			oldestAllowedCreationTime = System.currentTimeMillis();
			discards = new ArrayList<>(idleConnections);
			idleConnections.clear();
			numFlushed = discards.size();
			numFlushed += busyConnections.size(); // as they are all marked as obsolete.
		}
		for (PrivateConnectionInfo info : discards)
			discardConnection(info, "All obsoleted");
		return numFlushed;
	}

	public interface DiscardTester
	{
		boolean isWanted(long cacheAge, int numIdle);
	}

	public int discardOldConnections(@NotNull DiscardTester tester, long maxAllowedTimeSinceCreation)
	{
		if (idleConnections.isEmpty())
			return 0;
		List<PrivateConnectionInfo> discards = new ArrayList<>(2);
		List<PrivateConnectionInfo> candidates = new ArrayList<>(2);
		synchronized (this) {
			long now = System.currentTimeMillis();
			for (PrivateConnectionInfo info : idleConnections) {
				if (now - info.createTime > maxAllowedTimeSinceCreation)
					discards.add(info);
				else
					candidates.add(info);
			}
			Collections.sort(candidates, (a, b) -> Long.compare(b.cacheTime, a.cacheTime));
			for (int i = candidates.size() - 1; i >= 0; i--) {
				PrivateConnectionInfo info = candidates.get(i);
				if (tester.isWanted(now - info.cacheTime, i + 1))
					break;
				discards.add(info);
			}
			idleConnections.removeAll(discards);
		}
		for (PrivateConnectionInfo info : discards) {
			long now = System.currentTimeMillis();
			log("Purges connection: " + info.id + ", age = " +
					(now - info.createTime) / 1000L + " secs" +
					", idle = " + (now - info.cacheTime) / 1000L + " secs.");
			discardConnection(info, "Explicit too old");
		}
		return discards.size();
	}

	protected void createNewSpareConnection() throws Exception
	{
		numOpening++;
		PrivateConnectionInfo info;
		try {
			info = createNewConnectionInfo();
		}
		finally {
			numOpening--;
		}
		catchConnection(info);
	}

	void catchConnection(@NotNull PrivateConnectionInfo info)
	{
		// Don't synchronize the complete method to avoid having the lock
		// while calling info.connInfo.isReusable() as that method might
		// e.g. make a remote call to a database, and thus block other connections
		// from being cached while that method is executed.
		// And do not call discardConnection from inside synchronized.
		if (info.connInfo == null)
			return;
		int idx;
		synchronized (this) {
			idx = busyConnections.indexOf(info);
		}
		if (idx < 0) {
			logBug(info, "catchConnection with connection NOT in busyConnections");
			discardConnection(info, "catchConnection with connection NOT in busyConnections");
			return;
		}
		if (info.createTime < oldestAllowedCreationTime) {
			if (debugFlag || shouldLogCreationAndDestroy)
				log("catchConnection(" + info.id + "): is older than oldestAllowedCreationTime");
			synchronized (this) {
				busyConnections.remove(info);
			}
			discardConnection(info, "Older than oldestAllowedCreationTime");
			return;
		}
		if (debugFlag)
			log("catch " + info.id);
		if (!info.connInfo.isReusable()) {
			if (debugFlag) {
				log("catchConnection(" + info.id + "): couldn't be prepared for reuse");
			}
			synchronized (this) {
				busyConnections.remove(info);
			}
			discardConnection(info, "Not reusable");
			return;
		}
		synchronized (this) {
			int oldSize = idleConnections.size();
			info.cacheTime = System.currentTimeMillis();
			busyConnections.remove(info);
			idleConnections.add(info);
			if (debugFlag) {
				log("catchConnection(" + info.id + "): caches it: " + oldSize +
						" -> " + idleConnections.size());
			}
		}
		if (maxConnections != Integer.MAX_VALUE)
			connectionCachedNotificator.offer(this);
	}

	void discardBusyConnection(@NotNull PrivateConnectionInfo info,
	                           @NotNull String why)
	{
		synchronized (this) {
			int idx = busyConnections.indexOf(info);
			if (idx >= 0) {
				busyConnections.remove(idx);
			} else {
				logBug(info, "discardBusyConnection with connection NOT in busyConnections");
			}
		}
		discardConnection(info, why + " (busy)");
	}

	private void discardConnection(@NotNull PrivateConnectionInfo info,
	                               @NotNull String why)
	{
		// Do not synchronize the complete method, as this would make all other
		// connections block while abort() is executed, and abort() can be slow
		// if it e.g. executes a remote "rollback()".
		synchronized (this) {
			if (busyConnections.contains(info)) {
				logBug(info, "discardConnection with connection in busyConnections: " + why);
				busyConnections.remove(info);
			}
			if (idleConnections.contains(info)) {
				logBug(info, "discardConnection with connection in cachedConnections: " + why);
				idleConnections.remove(info);
			}
		}
		if (debugFlag)
			log("discardConnection " + info.id + ": " + why);
		try {
			info.connInfo.abort();
		}
		catch (Exception ex) {
			log("Got exception while aborting connection: " + ex.toString() + ": " + why);
			ex.printStackTrace();
		}
		finally {
			if (shouldLogCreationAndDestroy) {
				synchronized (this) {
					log("destroyed." + info.id + " -> " + idleConnections.size() + " cached, "
							+ busyConnections.size() + " alive: " + why);
				}
			}
		}
	}

	@SuppressWarnings({"ThrowableInstanceNeverThrown"})
	private void logBug(@NotNull PrivateConnectionInfo info, @NotNull String desc)
	{
		String msg = "@@@@ BUG " + desc + ", id = " + info.id;
		log(msg);
		new Exception(msg).printStackTrace();
	}

	public static String age(long t)
	{
		long d = System.currentTimeMillis() - t;
		d /= 1000;
		long h = d / (60 * 60);
		d -= h * (60 * 60);
		long m = d / 60;
		long s = d - m * 60;
		StringBuilder sb = new StringBuilder(10);
		//
		if (h < 10)
			sb.append('0');
		sb.append(h);
		sb.append(':');
		if (m < 10)
			sb.append('0');
		sb.append(m);
		sb.append(':');
		if (s < 10)
			sb.append('0');
		sb.append(s);
		return sb.toString();
	}
}
