/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.annotation.attribute.sub;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import org.jabhts.annotation.attribute.collection.ACollection;

/**
 * Specifies that a set of css attributes are generated by the method/field.
 * The {@code String} returned by the method/field must contain the full attribute specification,
 * including their name and value, e.g. "{@code color: blue; background-color: green; margin: 5px}".
 * The string is html-encoded before it is output, so you must not html-escape e.g. quotes.
 * <p/>
 * It is similar to ACollection, but for css attributes inside a {@code style} attribute.
 * <p/>
 * The html snippet<br>
 * &nbsp; {@code <div id=Logo style='color: green; border 2px dotted blue'>}<br>
 * is output as<br>
 * &nbsp; {@code <img id=Logo style='color: green; margin: 5px; padding 2px; border: 1px solid black'>}<br>
 * if you declare a method like:<br>
 * &nbsp; {@code @SCollection(styles = {"margin", "padding", "border"})}<br>
 * &nbsp; {@code String&nbsp;getLogo() { return "margin: 5px; padding 2px; border: 1px solid black"; }}<br>
 * or a field like:<br>
 * &nbsp; {@code @SCollection(styles = {"margin", "padding", "border"}, id = "Logo")}<br>
 * &nbsp; {@code String&nbsp;variable = "margin: 5px; padding 2px; border: 1px solid black";}
 * <p/>
 * You don't need to specify the {@code style} attribute in the template. The html snippet:<br>
 * &nbsp; {@code <div id=Logo>}<br>
 * is output as<br>
 * &nbsp; {@code <img id=Logo style='margin: 5px; padding 2px; border: 1px solid black'>}<br>
 * if the same field/method is specified.
 * <p/>
 * Must be applied to a String field or a method which returns a String and takes no parameters.
 *
 * @see SValue
 * @see ACollection
 */
@Target({ElementType.METHOD, ElementType.FIELD})
@Retention(RetentionPolicy.RUNTIME)
public @interface SCollection
{
	/**
	 * Makes the annotation match html elements which contains the value in the 'id' attribute or as a name in the 'class' attribute.
	 * Leading underscores in the names specified in the html are ignored when comparing the names.
	 * <p/>
	 * When the parameter is specified, the 'id', 'className' or 'tag' parameters must not be specified.
	 */
	String[] value() default {};

	/**
	 * Makes the annotation match html elements which contains the value in the 'id' attribute.
	 * Leading underscores in the 'id' attribute value are ignored when comparing the names.
	 * <p/>
	 * When the parameter is specified, the 'value' parameter must not be specified. It is allowed to specify
	 * the 'className' and 'tag' parameters.
	 */
	String[] id() default {};

	/**
	 * Makes the annotation match html elements which contains the value as a name in the 'class' attribute.
	 * Leading underscores in the names specified in the html are ignored when comparing the names.
	 * <p/>
	 * When the parameter is specified, the 'value' parameter must not be specified. It is allowed to specify
	 * the 'id' and 'tag' parameters.
	 */
	String[] className() default {};

	/**
	 * Makes the annotation match html elements whose tag is the specified value.
	 * <p/>
	 * When the parameter is specified, the 'value' parameter must not be specified. It is allowed to specify
	 * the 'id' and 'className' parameters.
	 */
	String[] tag() default {};

	/**
	 * Specifies whether the compiler should emit a warning (the default) if the annotation doesn't match any html element at all.
	 */
	boolean warnIfUnused() default true;

	/**
	 * The name of the css styles which the field/method includes.
	 */
	String[] styles();
}