/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.buildsystem;

import java.io.File;
import java.util.HashSet;
import java.util.Set;
import org.jabhts.translator.globals.CompilerOptions;

public class CommandLineOptionsDecoder
{
	public File scanBaseDir;
	public String scanBaseDottedPackageName;

	public File tmpDirF = new File("tmp");
	public File sourceInfoF;
	public File generatedSourceBaseDirF;
	public String generatedSourceEncoding = "UTF-8";
	public File generatedTemplateDataBaseDirF;
	public boolean forceRecompilation, checkForNewFiles, verbose, clean;
	public boolean saveSourceToDisk = true;
	public boolean traceProgram = false;

	public Set<String> compileClassFilter;

	public CompilerOptions options;

	public void main(String[] args) throws Exception
	{
		int idx = 0;
		while (idx < args.length && args[idx].startsWith("-")) {
			String s = args[idx++];
			switch (s) {
				case "-clean":
					clean = true;
					checkForNewFiles = true;
					break;
				case "-verbose":
					verbose = true;
					break;
				case "-force":
					forceRecompilation = true;
					break;
				case "-new":
					checkForNewFiles = true;
					break;
				case "-trace":
					traceProgram = true;
					break;
				case "-encoding":
					generatedSourceEncoding = args[idx++];
					break;
				case "-tmp":
					tmpDirF = new File(args[idx++]);
					break;
				case "-only":
					if (compileClassFilter == null)
						compileClassFilter = new HashSet<>();
					compileClassFilter.add(args[idx++]);
					break;
				default:
					throw new InvalidCommandLineOptionException("unknown option: " + s);
			}
		}
		if (idx > args.length - 2)
			throw new InvalidCommandLineOptionException("Too few arguments specified");
		sourceInfoF = new File(tmpDirF, "tcompile.dat");
		generatedTemplateDataBaseDirF = new File(tmpDirF, "jannotemp-data");
		generatedSourceBaseDirF = new File(args[idx++]);
		scanBaseDir = new File(args[idx++]);
		scanBaseDottedPackageName = idx < args.length ? args[idx] : "";
		options = new CompilerOptions(generatedSourceEncoding, sourceInfoF,
				generatedSourceBaseDirF, generatedTemplateDataBaseDirF,
				forceRecompilation, checkForNewFiles, verbose, saveSourceToDisk,
				compileClassFilter, traceProgram);
	}

	public void usage(String error)
	{
		System.err.println("Error: " + error);
		System.err.println("Usage:  <options> generated-source-basedir  scan-basedir  scan-basepackage");
		System.err.println("where possible options include:");
		System.err.println("  -new           Scan for new files rather than just check known files");
		System.err.println("  -force         Force recompilation of all files instead of only modified files");
		System.err.println("  -encoding <X>  Encoding used for generated java source files. Default is UTF-8.");
		System.err.println("  -only <X>      Compile only files matching <X>. Can be specified multiple times.");
		System.err.println("  -verbose       Produce verbose logging");
		System.err.println("  -trace         Verbose log of pcode generation");
		System.exit(1);
	}
}
