/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import org.jabhts.runtime.lib.TemplateData;
import org.jabhts.translator.buildsystem.CompilationInfo;
import org.jabhts.translator.compiler.application.scope.ScopeImpl;
import org.jabhts.translator.compiler.codeGeneration.impl.ProgramData;
import org.jabhts.translator.compiler.codeGeneration.impl.ProgramEmit;
import org.jabhts.translator.compiler.codeGeneration.impl.ProgramImpl;
import org.jabhts.translator.compiler.htmlLoader.HtmlDocument;
import org.jabhts.translator.compiler.htmlLoader.HtmlLoader;
import org.jabhts.translator.compiler.htmlLoader.HtmlTransformer;
import org.jabhts.translator.compiler.tree.TreeEmit;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.FileUtils;
import org.jabhts.translator.compiler.utils.LogLevel;
import org.jabhts.translator.globals.CompilerOptions;
import org.jabhts.translator.globals.FileInfo;
import org.jabhts.translator.globals.PersistentCompilationInfo;
import org.jabhts.translator.globals.ProgressInfo;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class AnnotationCompiler
{
	public final CompilerOptions options;
	public final PersistentCompilationInfo pci;
	public final FileInfo javaFI;
	public final Class<?> loadedJavaClass;
	public final CompilerLog log;
	public final CompilationInfo compilationInfo;

	public TemplateLocation htmlLocation;
	public HtmlDocument doc;
	public Node rootE;
	public boolean includeDoctype, onlyChilds;
	public ScopeImpl rootScope, annotationScope;
	public ProgramData programData;
	public String fullClassName;
	public TemplateData templateData;

	public AnnotationCompiler(CompilerOptions options, PersistentCompilationInfo pci,
	                          FileInfo javaFI, Class<?> loadedJavaClass,
	                          CompilerLog log, CompilationInfo compilationInfo)
	{
		this.options = options;
		this.pci = pci;
		this.javaFI = javaFI;
		this.loadedJavaClass = loadedJavaClass;
		this.log = log;
		this.compilationInfo = compilationInfo;
	}

	public void compile(ProgressInfo pi) throws CompilerException
	{
		findHtmlTemplate();
		if (htmlLocation == null) {
			pi.numErrors++;
			return;
		}
		loadHtmlTemplate();
		if (doc == null) {
			pi.numErrors++;
			return;
		}
		postProcessHtmlTemplate(doc);
		findRootElement(doc);
		if (rootE == null) {
			pi.numErrors++;
			return;
		}
		log.log(LogLevel.Verbose, "Root: " + rootE.getNodeName());
		pci.sourceFiles.add(javaFI);
		pci.sourceFiles.add(doc.fileInfo);
		//
		makeScopes();
		emitTree();
		emitPostCompilationWarnings();
		if (log.hasError()) {
			pi.numErrors++;
			return;
		}
		emitProgram(pi);
		File templateDataF = new File(options.generatedTemplateDataBaseDirF, fullClassName.replace('.', '/') + ".tmpldata");
		if (options.saveSourceToDisk)
			saveTemplateData(pi, templateDataF);
		pci.sourceFiles.add(new FileInfo(templateDataF));

	}

	protected void postProcessHtmlTemplate(HtmlDocument doc) throws CompilerException
	{
		try {
			for (HtmlTransformer t : options.htmlTransformers)
				t.transform(doc);
		}
		catch (CompilerException ex) {
			throw ex;
		}
		catch (Exception ex) {
			throw new CompilerException(ex.getMessage(), ex);
		}
	}

	protected void findRootElement(HtmlDocument doc) throws CompilerException
	{
		String fragment = compilationInfo.annot.fragment();
		onlyChilds = compilationInfo.annot.fragmentChilds();
		includeDoctype = false;
		if (fragment.isEmpty()) {
			if (onlyChilds) {
				rootE = findBodyElement();
				if (rootE == null) {
					log.log(LogLevel.Error,
							"Can't find <body> element in template.");
					return;
				}
			} else {
				rootE = doc.doc.getDocumentElement();
				includeDoctype = true;
			}
			return;
		}
		List<Element> rootList = doc.findMatchingElements(fragment);
		if (rootList.isEmpty()) {
			log.log(LogLevel.Error,
					"Can't find specified fragment node in html: " + compilationInfo.annotDescription);
			return;
		}
		if (rootList.size() > 1) {
			log.log(LogLevel.Error,
					"Specified fragment node matches multiple html elements: " + compilationInfo.annotDescription);
			return;
		}
		rootE = rootList.get(0);
	}

	protected Node findBodyElement()
	{
		for (Node n = doc.doc.getDocumentElement().getFirstChild(); n != null; n = n.getNextSibling()) {
			if (n.getNodeName().equalsIgnoreCase("body"))
				return n;
		}
		return null;
	}

	protected void findHtmlTemplate() throws CompilerException
	{
		String fn = FileUtils.basename(compilationInfo.annot.filename());
		if (fn.isEmpty()) {
			if (compilationInfo.scopeClass != null)
				fn = compilationInfo.scopeClass.getSimpleName();
			else
				fn = FileUtils.basename(javaFI.f.getName());
		}
		File dirF = javaFI.f.getParentFile();
		String packageName = loadedJavaClass.getPackage().getName();
		TemplateLocation loc = new TemplateLocation(dirF, packageName, fn);
		if (loc.htmlF != null) {
			htmlLocation = loc;
			return;
		}
		dirF = new File(dirF, "templates");
		packageName += ".templates";
		loc = new TemplateLocation(dirF, packageName, fn);
		if (loc.htmlF != null) {
			htmlLocation = loc;
			return;
		}
		log.log(LogLevel.Error, "Can't find html template: " + fn + ".html/.htm/.xhtml");
	}

	protected void loadHtmlTemplate() throws CompilerException
	{
		doc = new HtmlLoader(log).load(htmlLocation.htmlF);
	}

	protected void makeScopes() throws CompilerException
	{
		rootScope = new ScopeImpl(log, loadedJavaClass, compilationInfo);
		rootScope.process();
		makeSubScope(compilationInfo.scopeClass);
	}

	private ScopeImpl makeSubScope(Class<?> cls) throws CompilerException
	{
		if (cls == loadedJavaClass)
			return rootScope;
		ScopeImpl parentScope = makeSubScope(cls.getEnclosingClass());
		return parentScope.getSubScope(cls, null, null);
	}

	protected void emitTree() throws CompilerException
	{
		ProgramImpl program = new ProgramImpl(options, log, rootScope);
		annotationScope = enterScope(program, compilationInfo.scopeClass);
		TreeEmit te = new TreeEmit(options, program, log);
		te.emit(doc, includeDoctype, onlyChilds, rootE, annotationScope, compilationInfo.annotDescription);
		for (ScopeImpl scope = annotationScope; scope != rootScope; scope = scope.parent) {
			//log.log(LogLevel.Notice, "LEAVES AUTO-SCOPE: " + scope.scopeClass.getSimpleName());
			program.addLeaveScopeInstance(scope);
		}
		program.addReturn();
		programData = program.postProcess();
	}

	protected ScopeImpl enterScope(ProgramImpl program, Class<?> cls) throws CompilerException
	{
		if (cls == loadedJavaClass)
			return rootScope;
		ScopeImpl parentScope = enterScope(program, cls.getEnclosingClass());
		//log.log(LogLevel.Notice, "ENTERS AUTO-SCOPE: " + cls.getSimpleName());
		ScopeImpl scope = parentScope.getSubScope(cls, null, null);
		program.addCreateScopeInstance(scope);
		return scope;
	}

	protected void emitProgram(ProgressInfo pi) throws CompilerException
	{
		String packageName = htmlLocation.packageName;
		File tdF = new File(htmlLocation.dirF, "templatedata");
		if (tdF.exists())
			packageName += ".templatedata";
		fullClassName = packageName + "." + compilationInfo.tpname;
		ProgramEmit pe = new ProgramEmit(options, log, programData, packageName, fullClassName);
		pe.main(pi);
		templateData = pe.templateData;
		pci.sourceFiles.add(new FileInfo(pe.getGeneratedJavaF()));
	}

	protected void emitPostCompilationWarnings() throws CompilerException
	{
		annotationScope.emitPostCompilationWarnings();
	}

	protected void saveTemplateData(ProgressInfo pi, File templateDataF) throws CompilerException
	{
		try {
			File dirF = templateDataF.getParentFile();
			if (!dirF.exists() && !dirF.mkdirs())
				throw new IOException("Can't create directory: " + dirF.getPath());
			byte[] newData = templateData.toBytes(fullClassName);
			if (templateDataF.exists()) {
				byte[] oldData = FileUtils.readFile(templateDataF);
				if (Arrays.equals(newData, oldData)) {
					if (options.verbose)
						log.log(LogLevel.Notice, "tmpldata is unchanged");
					return;
				}
			}
			if (options.verbose)
				log.log(LogLevel.Notice, "Saves TemplateData in " + templateDataF.getPath());
			FileUtils.saveFile(templateDataF, newData);
			pi.numTemplateDataUpdated++;
		}
		catch (IOException ex) {
			throw new CompilerException("Can't save template data: " + CompilerUtils.getSomeMessage(ex), ex);
		}
	}

	static class TemplateLocation
	{
		final File dirF, htmlF;
		final String packageName;

		TemplateLocation(File dirF, String packageName, String fn) throws CompilerException
		{
			this.dirF = dirF;
			File foundF = null;
			for (String s : suffixes) {
				File f = new File(dirF, fn + s);
				if (!f.exists())
					continue;
				if (foundF != null)
					throw new CompilerException("Has both " + foundF.getName() + " and " + f.getName());
				foundF = f;
			}
			this.htmlF = foundF;
			this.packageName = packageName;
		}
	}

	static String[] suffixes = {".html", ".htm", ".xhtml"};
}
