/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.application.baseclass;

import org.jabhts.translator.compiler.application.scope.TargetingInfo;
import org.jabhts.translator.compiler.utils.CompilerException;

public abstract class AbstractValueApplication extends Application
{
	public final CallbackType callbackType;
	public final Class<?> resultType;

	public enum CallbackType
	{
		Field, Method, VoidMethod
	}

	protected AbstractValueApplication(TargetingInfo target,
	                                   boolean allowVoidMethod, Class<?>... allowedTypes) throws CompilerException
	{
		super(target);
		if (method == null && field == null)
			throw makeInvalidAnnotationApplianceException("Annotation must be applied to field or method: " + annotationToString());
		Class<?> actualType = (method != null) ? method.getReturnType() : field.getType();
		if (method != null) {
			if (method.getParameterTypes().length > 0)
				throw makeInvalidAnnotationApplianceException("Method may not take any arguments: " + annotationToString());
			if (allowVoidMethod && actualType.getName().equals("void")) {
				callbackType = CallbackType.VoidMethod;
				resultType = null;
				return;
			}
		}
		callbackType = (method != null) ? CallbackType.Method : CallbackType.Field;
		Class<?> match = findMatchingType(actualType, allowedTypes);
		if (match != null) {
			resultType = match;
			return;
		}
		StringBuilder sb = new StringBuilder();
		sb.append(method != null ? "Method must return " : "Field must be ");
		if (allowedTypes.length == 1)
			sb.append("a ");
		for (int i = 0; i < allowedTypes.length; i++) {
			Class<?> t = allowedTypes[i];
			if (i > 0)
				sb.append(i + 1 == allowedTypes.length ? " or " : ", ");
			String s = t.getSimpleName();
			sb.append(s);
			if (t.isPrimitive())
				sb.append("/").append(s.toLowerCase());
		}
		sb.append(method != null ? " but returns " : " but is ");
		sb.append(actualType.getSimpleName());
		throw makeInvalidAnnotationApplianceException(sb.toString());
	}

	private static Class<?> findMatchingType(Class<?> actualType, Class<?>[] allowedTypes)
	{
		String s = actualType.getSimpleName();
		for (Class<?> t : allowedTypes) {
			if (t == Integer.class && s.equals("int"))
				return t;
			if (t == Long.class && s.equals("long"))
				return t;
			if (t == Boolean.class && s.equals("boolean"))
				return t;
			if (t.isAssignableFrom(actualType))
				return t;
		}
		return null;
	}
}
