/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.compiler.tree;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.jabhts.translator.compiler.application.attribute.sub.SubAttributeEmitter;
import org.jabhts.translator.compiler.codeGeneration.api.Program;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerLog;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.InternalCompilerException;
import org.jabhts.translator.compiler.utils.LogLevel;

class StyleSubAttributesHandler extends AbstractSubAttributesHandler
{
	StyleSubAttributesHandler(Program program, CompilerLog log,
	                          List<SubAttributeEmitter<?>> subEmitters) throws InternalCompilerException
	{
		super(program, log, subEmitters);
	}

	String itemPrefix;

	void main(String attrValue) throws CompilerException
	{
		itemPrefix = " " + subEmitters.get(0).lcAttributeName + "='";
		if (!attrValue.isEmpty()) {
			Set<String> handledItems = new HashSet<>();
			for (SubAttributeEmitter<?> p : subEmitters) {
				for (String s : p.lcSubAttributeNames)
					handledItems.add(s);
			}
			new StyleParser(attrValue, handledItems);
		}
		if (subEmitters.size() == 1 && !itemPrefix.contains(";") && subEmitters.get(0).canGenerateFullAttribute()) {
			// if only one conditional and we have no static value, then make the complete element
			// attribute inside the conditional so we avoid "style=''" when the conditional is false
			subEmitters.get(0).generateFullAttribute();
		} else {
			program.addEmitString(itemPrefix);
			itemPrefix = "";
			for (SubAttributeEmitter<?> p : subEmitters) {
				p.generateSubAttribute(itemPrefix);
				itemPrefix = "; ";
			}
		}
		if (itemPrefix.contains(";"))
			program.addEmitString("'");
	}

	protected void parseStyles(String attrValue) throws CompilerException
	{
		try {
			int idx = 0;
			while (idx < attrValue.length()) {
				while (idx < attrValue.length() && Character.isWhitespace(attrValue.charAt(idx)))
					idx++;
			}
		}
		catch (Exception ex) {
			log.log(LogLevel.Error, "Can't parse style='" + attrValue + "': " + CompilerUtils.getSomeMessage(ex));
		}
	}

	final class StyleParser
	{
		final String attrValue;
		int idx;

		StyleParser(String attrValue, Set<String> handledItems)
		{
			this.attrValue = attrValue;
			try {
				while (true) {
					char ch = next(null);
					if (ch == 0)
						break;
					if (Character.isWhitespace(ch))
						continue;
					String name = parseName(ch);
					String value = parseValue(name);
					if (handledItems.contains(name.toLowerCase()))
						continue;
					program.addEmitString(itemPrefix);
					itemPrefix = "; ";
					program.addEmitString(name);
					program.addEmitString(": ");
					program.addEmitString(value);
				}
			}
			catch (Exception ex) {
				log.log(LogLevel.Error, "Can't parse style='" + attrValue + "': " + CompilerUtils.getSomeMessage(ex));
			}
		}

		void skipWhiteSpace()
		{
			while (idx < attrValue.length() && Character.isWhitespace(attrValue.charAt(idx)))
				idx++;
		}

		char next(String msg) throws CompilerException
		{
			if (idx < attrValue.length())
				return attrValue.charAt(idx++);
			if (msg != null)
				error(msg);
			return 0;
		}

		String parseName(char ch) throws CompilerException
		{
			StringBuilder sb = new StringBuilder();
			do {
				sb.append(ch);
				ch = next("end of style attribute when expecting ':' after name");
				if (ch == ';')
					error("';' when expecting ':' after name");
			} while (ch != ':');
			String s = sb.toString().trim();
			for (int i = 0; i < s.length(); i++) {
				ch = s.charAt(i);
				if (ch != '-' && !Character.isLetter(ch)) {
					log.log(LogLevel.Notice, "Weird css style name: '" + s + "'");
					break;
				}
			}
			return s;
		}

		String parseValue(String name) throws CompilerException
		{
			StringBuilder sb = new StringBuilder();
			while (true) {
				char ch = next(null);
				if (ch == 0)
					break;
				if (ch <= 32 && sb.length() == 0)
					continue;
				if (ch == ';')
					break;
				sb.append(ch);
				if (ch == '\\') {
					sb.append(next("backslash-escaped character"));
					continue;
				}
				if (ch != '"' && ch != '\'')
					continue;
				char quote = ch;
				do {
					ch = next("inside string");
					sb.append(ch);
					if (ch == '\\')
						sb.append(next("backslash-escaped character inside string"));
				} while (ch != quote);
			}
			String s = sb.toString().trim();
			if (s.isEmpty())
				error("empty style value for '" + name + "'");
			return s;
		}

		void error(String msg) throws CompilerException
		{
			throw new CompilerException(msg + " at character " + idx + " of " + attrValue.length());
		}

	}
}
