/* Copyright (c) 2006-2018 Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.jabhts.translator.globals;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InvalidClassException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.jabhts.translator.buildsystem.SourceInfo;
import org.jabhts.translator.compiler.htmlLoader.HtmlTransformer;
import org.jabhts.translator.compiler.htmlLoader.HtmlTransformerMetaTags;
import org.jabhts.translator.compiler.utils.CompilerException;
import org.jabhts.translator.compiler.utils.CompilerUtils;
import org.jabhts.translator.compiler.utils.FileUtils;

public class CompilerOptions
{
	public final String sourceEncoding;
	public final File sourceInfoF;
	public final File generatedSourceBaseDirF;
	public final File generatedTemplateDataBaseDirF;
	public final ClassLoader classloader = getClass().getClassLoader();
	public final boolean forceRecompilation, checkForNewFiles, verbose, saveSourceToDisk;
	public final boolean traceProgram;

	public Map<File, SourceInfo> sourceInfoMap;
	public Set<String> compileClassFilter;

	public TreeEmitModificator treeModificator = new TreeEmitModificator();

	public List<HtmlTransformer> htmlTransformers = new ArrayList<>();

	public CompilerOptions(String sourceEncoding, File sourceInfoF,
	                       File generatedSourceBaseDirF, File generatedTemplateDataBaseDirF,
	                       boolean forceRecompilation, boolean checkForNewFiles, boolean verbose,
						   boolean saveSourceToDisk,
	                       Set<String> compileClassFilter, boolean traceProgram)
	{
		this.sourceEncoding = sourceEncoding;
		this.sourceInfoF = sourceInfoF;
		this.generatedSourceBaseDirF = generatedSourceBaseDirF;
		this.generatedTemplateDataBaseDirF = generatedTemplateDataBaseDirF;
		this.forceRecompilation = forceRecompilation;
		this.checkForNewFiles = checkForNewFiles;
		this.verbose = verbose;
		this.saveSourceToDisk = saveSourceToDisk;
		this.compileClassFilter = compileClassFilter;
		this.traceProgram = traceProgram;
		this.htmlTransformers.add(new HtmlTransformerMetaTags());
	}


	public void loadSourceInfo() throws Exception
	{
		if (sourceInfoF.exists()) {
			//noinspection UnusedCatchParameter
			try {
				byte[] buffer = FileUtils.readFile(sourceInfoF);
				try (ObjectInputStream is = new ObjectInputStream(new ByteArrayInputStream(buffer))) {
					sourceInfoMap = (SourceInfoMap)is.readObject();
					if (sourceInfoMap != null)
						return;
				}
			}
			catch (InvalidClassException ex) {
				System.err.println("Incompatible saved SourceInfoMap: " + sourceInfoF.getPath());
			}
			catch (Exception ex) {
				System.err.println("Can't read existing SourceInfoMap ("
						+ CompilerUtils.getSomeMessage(ex) + "): " + sourceInfoF.getPath());
			}
		}
		sourceInfoMap = new SourceInfoMap();
	}

	static class SourceInfoMap extends HashMap<File, SourceInfo>
	{
		// class created to avoid unchecked-cast warning from is.readObject()
		private static final long serialVersionUID = 4593253319314046083L;
	}

	public void saveSourceInfo() throws Exception
	{
		System.err.println("Saves sourceInfoMap");
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ObjectOutputStream os = new ObjectOutputStream(baos);
		os.writeObject(sourceInfoMap);
		os.close();
		File f = sourceInfoF.getParentFile();
		if (f != null && !f.exists() && !f.mkdirs())
			throw new IOException("Can't create dir: " + f.getPath());
		FileUtils.saveFile(sourceInfoF, baos.toByteArray());
	}

	public static class TreeEmitModificator
	{
		public String modifyScriptElementContent(String js) throws CompilerException
		{
			return js;
		}

		public String modifyStyleElementContent(String css) throws CompilerException
		{
			return css;
		}

		public String modifyCommentContent(String s) throws CompilerException
		{
			return s;
		}
	}
}
