/* Copyright (c) 2006-2011, Peter Speck / Vitality. All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Vitality nor the names of its contributors may be
 *       used to endorse or promote products derived from this software without
 *       specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PETER SPECK / VITALITY AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL PETER SPECK / VITALITY AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR  SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.vitality.lib.foundation.xml.parse;

import java.io.File;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import com.sun.org.apache.xml.internal.serialize.XMLSerializer;
import org.jetbrains.annotations.NotNull;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.ext.DefaultHandler2;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.xml.create.XmlCreateUtils;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;

public class SaxToDomParser extends DefaultHandler2
{
	public Document doc;
	public XmlArgs docXA;
	protected List<Element> elementStack = new ArrayList<>();
	protected Exception fatalErrorException;
	protected Element currentElement;
	protected boolean debugFlag;

	public void parse(@NotNull File f) throws Exception
	{
		getParser().parse(f, this);
	}

	@SuppressWarnings({"UnusedDeclaration"})
	public void debugDumpDocument() throws IOException
	{
		XMLSerializer serializer = new XMLSerializer();
		serializer.setOutputCharStream(new OutputStreamWriter(System.out, "UTF-8"));
		serializer.serialize(doc);
		System.out.println();
	}

	@NotNull
	public SAXParser getParser() throws Exception
	{
		synchronized (getClass()) {
			if (factory == null)
				factory = SAXParserFactory.newInstance();
			return factory.newSAXParser();
		}
	}

	@SuppressWarnings({"UnusedDeclaration"})
	protected void endedElement(String qName) throws Exception
	{
		// for you to override.
	}

	public void zapCurrentElement()
	{
		if (currentElement == null)
			return;
		Node parent = currentElement.getParentNode();
		if (parent == null)
			return;
		parent.removeChild(currentElement);
	}

	public void startElement(String uri, String localName, String qName, Attributes attrs)
	{
		if (debugFlag) {
			printIndent();
			System.err.println("startElement  " + uri + ";" + localName + ";" + qName);
		}
		Element e = doc.createElementNS(uri, qName);
		int num = attrs.getLength();
		for (int i = 0; i < num; i++)
			e.setAttributeNS(attrs.getURI(i), attrs.getQName(i), attrs.getValue(i));
		if (currentElement == null)
			doc.appendChild(e);
		else
			currentElement.appendChild(e);
		elementStack.add(e);
		currentElement = e;
	}

	public void endElement(String uri, String localName, String qName) throws SAXException
	{
		int size = elementStack.size();
		elementStack.remove(--size);
		Element e = (size == 0) ? null : elementStack.get(--size);
		if (debugFlag) {
			printIndent();
			System.err.println("endElement  " + uri + ";" + localName + ";" + qName);
		}
		try {
			endedElement(qName);
		}
		catch (SAXException ex) {
			throw ex;
		}
		catch (Exception ex) {
			ex.printStackTrace();
			throw new SAXException(ExceptionUtils.getExMsg(ex), ex);
		}
		currentElement = e;
	}

	protected void printIndent()
	{
		for (int i = elementStack.size(); i > 0; i--)
			System.err.print(". ");
	}

	@Override
	public void characters(char[] chars, int offset, int length) throws SAXException
	{
		if (isWhitespace(chars, offset, length) && whitespaceIsIgnorable(chars, offset, length))
			return;
		if (currentElement == null) {
			throw new SAXException("Non-ignorable chars outside document.");
		}
		String s = new String(chars, offset, length);
		makeTextChild(s);
	}

	protected void makeTextChild(String s)
	{
		Text t = doc.createTextNode(s);
		currentElement.appendChild(t);
	}

	@SuppressWarnings({"UnusedDeclaration"})
	protected boolean whitespaceIsIgnorable(char[] chars, int offset, int length) throws SAXException
	{
		return true;
	}

	protected boolean isWhitespace(char[] chars, int offset, int length) throws SAXException
	{
		while (length > 0) {
			char ch = chars[offset++];
			if (ch > 32)
				return false;
			length--;
		}
		return true;
	}

	public void warning(SAXParseException ex)
	{
		System.err.println("[Warning] " + getLocationString(ex) + ": " + ex.getMessage());
	}

	public void error(SAXParseException ex)
	{
		System.err.println("[Error] " + getLocationString(ex) + ": " + ex.getMessage());
		if (fatalErrorException == null)
			fatalErrorException = ex;
	}

	public void fatalError(SAXParseException ex) throws SAXException
	{
		System.err.println("[Fatal Error] " + getLocationString(ex) + ": " + ex.getMessage());
		throw ex;
	}

	public static String getLocationString(SAXParseException ex)
	{
		StringBuilder sb = new StringBuilder();
		String systemId = ex.getSystemId();
		if (systemId != null) {
			int index = systemId.lastIndexOf('/');
			if (index != -1)
				systemId = systemId.substring(index + 1);
			sb.append(systemId);
			sb.append(':');
		}
		sb.append(ex.getLineNumber());
		sb.append(':');
		sb.append(ex.getColumnNumber());
		return sb.toString();
	}

	private static SAXParserFactory factory = null;

	@Override
	public void startDocument() throws SAXException
	{
		try {
			doc = XmlCreateUtils.createDocument();
			docXA = new XmlArgs(doc);
		}
		catch (ParserConfigurationException ex) {
			throw new SAXException("Can't create new document: " + ExceptionUtils.getExMsg(ex), ex);
		}
	}
}
