#!/usr/bin/perl

use strict;
use Socket;
use POSIX 'setsid';
use XML::XPath;

$| = 1; # can't use stderr as AuthorizationExecuteWithPrivileges only reads stdout.

delete @ENV{'PATH', 'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};
$ENV{"SHELL"} = "/bin/sh";
$ENV{"IFS"} = "";
$ENV{"PATH"} = "/bin:/usr/bin:/usr/sbin:/sbin";
$ENV{"CDPATH"} = "";

my $verbose = defined($ENV{VERBOSE});
print "GlimmerBlocker installer: uid = $<, euid = $>\n";
system("id");
$< = 0;
die "Not running as root" if $<;
chdir('/tmp') || die "Can't chdir('/tmp')\n"; # have pwd to be a place _glimmerblocker user can access so sudo _glimmerblocker doesn't warn.

my $runningBundlePath = "/Library/PreferencePanes/GlimmerBlocker.prefPane";
my $runningFromSystemPreferences = undef;
while (@ARGV) {
	my $opt = shift(@ARGV);
	if ($opt eq "-v") {
		$verbose = 1;
	} elsif ($opt eq "-syspref") {
		$runningFromSystemPreferences = 1;
		print "Running from System Preferences.\n";
	} elsif ($opt eq "-bundle") {
		$_ = shift(@ARGV) || die "Missing -bundle argument.";
		die unless m{^(/.*[^/])\/*$};
		$runningBundlePath = $1; # untaint.
	} else {
		warn "Unknown arg: $opt\n";
	}
}
die "-bundle <path> not specified." unless defined $runningBundlePath;
print "Current panel bundle path: $runningBundlePath\n";

my $panelFilename = "GlimmerBlocker.prefPane";
my $panelPath = "/Library/PreferencePanes/$panelFilename";
my $authHelper = "$panelPath/Contents/AuthHelper.app/Contents/MacOS/AuthHelper";
my $settingsPath = "/Library/GlimmerBlocker";
my $panelSettingsPath = "$settingsPath/PanelSettings.xml";
my $javaSettingsPath = "$settingsPath/JavaSettings.xml";
my $lastUpdateCheckPath = "$settingsPath/LastUpdateCheck";
my $proxyBinary = "$panelPath/Contents/GlimmerBlockerProxy.app/Contents/MacOS/GlimmerBlockerProxy";
my $javaJarDir = "$panelPath/Contents/GlimmerBlockerProxy.app/Contents/Resources/Java";

# doesn't die if fails
sub shsys
{
	my ($cmd) = @_;
	print "# runs: $cmd\n";
	my $x = system($cmd);
	print "# exit code $x" if $x;
	return $x;
}

sub sys
{
	my (@cmd) = @_;
	print "# execs: ", join(" ", @cmd), "\n";
	my $x = system(@cmd);
	return unless $x;
	die "system(".join(" ", @cmd).") failed with exitcode ".($x / 256)."\n";
}

sub getHash
{
	my ($cmd) = @_;
	my %h;
	open(DSCL, "$cmd |") || die;
	while (<DSCL>) {
		chomp;
		if (/^(\S+?):?\s+(\S+)/) {
			$h{$1} = $2;
		} elsif (/^(\S+)/) {
			$h{$1} = 1;
		} else {
			print "Weird value from '$cmd': '$_'\n";
		}
	}
	close(DSCL);
	return \%h;
}

sub findUnusedId
{
	my ($hR) = @_;
	my %id;
	$id{$_} = 1 foreach values %$hR;
	for (my $i = 50; 1; $i++) {
		return $i unless $id{$i};
	}
}

sub makeGroup
{
	my $hR = getHash("dscl . -list /Groups PrimaryGroupID");
	if ($$hR{_glimmerblocker}) {
		print "Has already _glimmerblocker group\n";
		return;
	}
	my $id = findUnusedId($hR);
	print "Creates group _glimmerblocker with PrimaryGroupID = $id\n";
	my $g = "/Groups/_glimmerblocker";
	sys(("dscl", ".", "-create", $g));
	sys(("dscl", ".", "-create", $g, "PrimaryGroupID", $id));
	sys(("dscl", ".", "-append", $g, "RecordName", "glimmerblocker"));
	sys(("dscl", ".", "-create", $g, "RealName", "GlimmerBlocker-proxy"));
}

sub makeUser
{
	my $hR = getHash("dscl . -read /Groups/_glimmerblocker PrimaryGroupID");
	my $groupId = $$hR{PrimaryGroupID} || die "Can't get PrimaryGroupID of _glimmerblocker group";
	$hR = getHash("dscl . -list /Users UniqueID");
	if ($$hR{_glimmerblocker}) {
		print "Has already _glimmerblocker user\n";
		return;
	}
	my $id = findUnusedId($hR);
	print "Creates user _glimmerblocker with UniqueID = $id\n";
	my $g = "/Users/_glimmerblocker";
	sys(("dscl", ".", "-create", $g));
	sys(("dscl", ".", "-create", $g, "UniqueID", $id));
	sys(("dscl", ".", "-create", $g, "PrimaryGroupID", $groupId));
	sys(("dscl", ".", "-append", $g, "RecordName", "glimmerblocker"));
	sys(("dscl", ".", "-create", $g, "RealName", "GlimmerBlocker-proxy"));
	sys(("dscl", ".", "-create", $g, "NFSHomeDirectory", "$panelPath/Contents/Resources"));
	sys(("dscl", ".", "-create", $g, "UserShell", "/usr/bin/false"));
}

sub setPerms
{
	my ($path, $perm) = @_;
	return unless -e $path;
	if ($perm =~ /\D/) {
		sys(("chmod", sprintf("0%o", $perm), $path));
	} else {
		printf "chmod 0%o  %s\n", $perm, $path if $verbose;
		chmod($perm, $path) || die sprintf("Can't chmod 0%o %s", $perm, $path);
	}
}

sub createFolder
{
	my ($path, $owner, $perm, $force) = @_;
	if (-e $path) {
		return unless $force;
	} else {
		mkdir($path, $perm) || die "Can't create dir: $path";
		sys(("chown", $owner, $path));
		return;
	}
	if ($force > 1) {
		print STDERR "createFolder.setDeepPerms($owner, $perm, $force): $path\n";
		shsys("find '$path' -type d -print0 | xargs -0 chmod ".sprintf("0%o", $perm));
		shsys("find '$path' -type f -print0 | xargs -0 chmod ".sprintf("0%o", $perm & 0666));
	} else {
		print STDERR "createFolder.setPerms($owner, $perm, $force): $path\n";
		setPerms($path, $perm);
	}
	shsys("find '$path' -print0 | xargs -0 chown $owner");
}

sub prefByXPath
{
	my ($xpath) = @_;
	if (! -e $panelSettingsPath) {
		print STDERR "prefByXPath: Missing settings file for: '$xpath'\n";
		return '';
	}
	my $r = XML::XPath->new(filename => $panelSettingsPath);
	my $nodes = $r->find($xpath);
	unless ($nodes->isa('XML::XPath::NodeSet')) {
		print STDERR "prefByXPath: xpath didn't return nodeset: '$xpath'\n";
		return "";
	}
	unless ($nodes->size) {
		print STDERR "prefByXPath: No node matches: '$xpath'\n";
		return '';
	}
	my $v = ($nodes->get_nodelist)[0]->string_value;
	print STDERR "prefByXPath: Node with value '$v' matches: '$xpath'\n";
	return $v;
}

sub stopLaunchdDaemons
{
	my $any = 1;
	foreach my $dir (qw( /Library/LaunchAgents /Library/LaunchDaemons )) {
		opendir(DIR, $dir) or next;
		while (defined($_ = readdir(DIR))) {
			next unless /^(org\.glimmerblocker.*\.plist)$/;
			print STDERR "Stops launchd daemon: $1\n";
			my $path = "$dir/$1";
			sys(("launchctl", "unload", $path));
			unlink($path);
			$any = 1;
		}
		closedir(DIR);
	}
	sleep 1 if $any;
}

sub killSystemPreferencesApplication
{
	if ($runningFromSystemPreferences) {
		print "Detaches from System Preferences process: pid = $$, ppid = ", getppid(), "\n";
		if (setsid() < 0) {
			print "Failed to setsid() nr 1\n";
		}
		exit if fork();
		if (setsid() < 0) {
			print "Failed to setsid() nr 2\n";
		}
		exit if fork();
		if (setsid() < 0) {
			print "Failed to setsid() nr 3\n";
		}
		sleep 1;
		print "Now detacted from SysPref parent process: pid = $$, ppid = ", getppid(), "\n";
#		if (fork()) {
#			print "Tells System Preferences to quit using AppleScript (helper pid $$).\n";
#			shsys(("osascript", "-e",
#						"tell application \"System Preferences\"\n".
#						"   ignoring application responses\n".
#						"      quit\n".
#						"   end ignoring\n".
#						"end tell\n"));
#			exit;
#		}
	}
# The updater application should already have quitted it,
# but make really sure it doesn't override the upgraded PanelPreferences.xml
	unless (open(FH, "ps axo pid=,user=,comm= |")) {
		print "Can't pipe from ps-find-syspanel.\n";
		return;
	}
	my $num = 0;
	while (<FH>) {
		next unless m!^\s*(\d+)\s+(\S+)\s+/Applications/System Preferences.app/Contents/MacOS!;
		print "Kills System Preferences.app with pid $1 user ($2)\n";
#kill 1, $1;
#sleep 1;
#kill 9, $1;
		$num++;
	}
	print "Found $num running System Preferences apps.\n";
	close(FH);
}

###############################################################################################
#
print "MSG:Stops old proxy.\n";
stopLaunchdDaemons();
#
print "MSG:Prepares system.\n";
makeGroup();
makeUser();
#
createFolder("/Library/Logs", "root:admin", 0775, 0);
createFolder("/Library/Logs/GlimmerBlocker", "_glimmerblocker:_glimmerblocker", 0755, 2);
my $log = "/Library/Logs/GlimmerBlocker/GlimmerBlocker.log";
unless (-e $log) {
	open(FH, ">$log") && close(FH);
}
setPerms($log, 0644);
shsys("chown _glimmerblocker:_glimmerblocker $log");
shsys("chmod 666 /Library/Logs/GlimmerBlocker/Launchd.log");
#
# Can't store preferences in /Library/Preferences because it has drwxrwxr-x
# so all admin users can change all files at will.
createFolder($settingsPath, "_glimmerblocker:_glimmerblocker", 0755, 1);
createFolder("$settingsPath/Filter subscriptions", "_glimmerblocker:_glimmerblocker", 0755, 2);
createFolder($lastUpdateCheckPath, "_glimmerblocker:admin", 0777, 2);
setPerms($javaSettingsPath, 0600);
setPerms($panelSettingsPath, 0644);
#
unless ($runningBundlePath eq $panelPath) {
	print "\n";
	printf "Payload size: %.1f MB.\n", (-s $runningBundlePath) / 1024.0 / 1024.0;
	print "MSG:Relocates Panel bundle.\n";
	sys(("rsync", "--archive", "--delete-after", "$runningBundlePath/", $panelPath));
}
foreach my $payloadName (qw( GlimmerBlockerProxy GlimmerBlockerInstaller 
			AuthHelper NetworkConfigurationListener NetworkSettingsUpdater )) {
	my $tar = "$panelPath/$payloadName.tar";
	next unless -e $tar;
	print "MSG:Installs $payloadName\n";
	sys(("tar", "xf", $tar, "-C", "$panelPath/Contents"));
	sys(("rm", $tar));
}
print "MSG:Sets permissions.\n";
shsys("chown -R _glimmerblocker:_glimmerblocker $panelPath");
shsys("chown root:admin $authHelper");
shsys("chmod 6555 $authHelper");
unlink("/Library/GlimmerBlocker/LastUpdateCheck.txt"); # obsolete.

my $jre = "$panelPath/Contents/GlimmerBlockerProxy.app/Contents/PlugIns/jre/";
shsys("du -sh $panelPath/Contents/* $panelPath/Contents/GlimmerBlockerProxy.app/Contents/* $jre");

###############################################################################################

sub needsUpgradeLittleSnitch
{
	my $path = "/Library/Little Snitch/Little Snitch UIAgent.app/Contents/Info.plist";
	return 0 unless -s $path;
	open(FH, $path) || return 0;
	$_ = join("", <FH>);
	close(FH);
	return m!<key>CFBundleVersion</key>\s*<string>385</string>! ? 1 : 0;
}   

sub openPreferencesPanel
{
	print "Opens System Preferences.\n";
	shsys("open $panelPath");
}

sub getPlistPath
{
	my ($opt) = @_;
	my $r = `$authHelper $opt`;
	unless ($r =~ m{^((/[-.\w]+)+)$}) {
		die "Could not decode AuthHelper $opt output = '$r'";
	}
	return $1;
}

sub savePlist
{
	my ($pathOpt, $contentsOpt) = @_;
	my $plistPath = getPlistPath($pathOpt);
	my $plistContents = `$authHelper $contentsOpt`;
	my $plistDir = $plistPath;
	$plistDir =~ s{/[^/]+$}{};
	unless (-e $plistDir) {
		sys("mkdir -p $plistDir");
		sys("chmod 755 $plistDir");
		shsys("chown root:wheel $plistDir"); # ok to fail in 10.6
	}
	sys("chmod u+w $plistDir");
	open(FH, ">$plistPath") || die "Can't create $plistPath for writing.";
	print FH $plistContents;
	close(FH);
	sys("chown root:wheel $plistPath");
	sys("chmod 644 $plistPath");
	shsys("launchctl unload $plistPath");
	sys("launchctl load $plistPath");
}

sub startServer
{
	print "\n";
	print "Launches java server.\n";
	print "MSG:Starts proxy server.\n";
	savePlist("--plist-path", "--plist-contents");
}

sub runJavaProgram
{
	my ($mainClass, $env) = @_;
	$mainClass =~ s!\.!/!g;
	print "Running $mainClass\n";
	$env = "" unless $env;
	my $cmd = "time sudo -u _glimmerblocker $env $proxyBinary --main-class $mainClass";
	my $x = shsys($cmd);
	unless ($x) {
		print "Done running $mainClass\n";
		return;
	}
	print "Failed ($x) executing: $cmd\n";
	print "CAN'T FIND JAVA INSTALLATION.\n";
	exit(1);
}

sub upgradePanelPreferences
{
	runJavaProgram("org.glimmerblocker.proxy.PanelPreferencesUpgrader");
}

sub savePanelPrefsXml
{
	my ($xml) = @_;
	open(FH, ">$panelSettingsPath") || die "Can't create $panelSettingsPath for writing: $!";
	print FH $xml;
	close(FH);
	sys("chown _glimmerblocker:_glimmerblocker $panelSettingsPath");
	sys("chmod 644 $panelSettingsPath");
}

sub wantsServerToRun
{
	return 1 if -e getPlistPath("--plist-path");
	return prefByXPath('prefs/proxy-server/@enabled');
}

sub doneInstalling
{
	unless ($runningBundlePath eq $panelPath) {
		print "\n";
		print "MSG:Zaps wrongly located bundle.\n";
		sys(("rm", "-rf", $runningBundlePath));
	}
	if (prefByXPath('prefs/updater/@check-in-background')) {
		print "Makes updater plist\n";
		savePlist("--updater-plist-path", '--updater-plist-content');
	} else {
		print "Automatick check for updates not enabled, skips making updater plist\n";
	}
	sleep 2; # give Java a bit of time to get running.
	openPreferencesPanel(); # if needsUpgradeLittleSnitch() || $runningFromSystemPreferences;
	print "\n";
	print "Installer done, exits.\n";
	exit(0);
}

killSystemPreferencesApplication();
sys("ls -ld / /Library /Library/PreferencePanes $panelPath $javaJarDir");
print "#\n";
print "# OBS: If the following command fails,\n";
print "       you probably have a permission problem with one of the above directories.\n";
print "#\n";
sys("sudo -u _glimmerblocker sh -c 'id;pwd;cd $javaJarDir;pwd;ls -l'");

print "\n\n\n";
runJavaProgram("org.glimmerblocker.proxy.TestJava", "DYLD_PRINT_LIBRARIES=1 DYLD_PRINT_RPATHS=1 DYLD_BIND_AT_LAUNCH=1 DYLD_PRINT_APIS=1");
print "\n\n\n";

if (-e $panelSettingsPath) {
	print "\n";
	my $run = wantsServerToRun();
	print "", ($run ? "Wants" : "Does not want"), " server to run.\n";
	print "MSG:Updates settings.\n";
	upgradePanelPreferences();
	openPreferencesPanel() if needsUpgradeLittleSnitch() || $runningFromSystemPreferences;
	print "postflight done of upgrade, has panel-prefs, so just quit\n";
	if ($run) {
		sleep 2;
		startServer();
	}
	doneInstalling();
}

print "\n";
print "Missing panel prefs -> creates initial file\n";
print "MSG:Creates initial settings.\n";
# initial installation. Create prefs file and launch java server.
open(FH, "/dev/random") || die "Can't open /dev/random for reading";
my $adminKey;
read(FH, $adminKey, 20);
$adminKey = unpack("H*", $adminKey);
my $port = 8228;
my $xml = <<"XML";
<?xml version="1.0" encoding="UTF-8"?>
<prefs version="0">
    <filters>
    </filters>
    <proxy-server listen-port="$port" info-admin-key="$adminKey" enabled="1"></proxy-server>
    <suspects show-details="1" enabled="1"/>
	<history enabled="1"/>
    <updater check-in-background="1"/>
</prefs>
XML
savePanelPrefsXml($xml);
upgradePanelPreferences();
sys("sync");
#
startServer();
#
print "\n";
print "Synchronizes network proxy settings.\n";
print "MSG:Updates 'Network' proxy settings.\n";
sys("$panelPath/Contents/NetworkSettingsUpdater.app/Contents/MacOS/NetworkSettingsUpdater $port");
#
# wait for java server to start before we're really done
# so the user don't get the "proxy not responding" error if he's quick to launch Safari.
print "\n";
print "MSG:Waits for proxy to become ready.\n";
my $iaddr = inet_aton("127.0.0.1") || die "Failed creating iaddr";
my $paddr = sockaddr_in($port, $iaddr) || die "Failed creating paddr";
my $proto = getprotobyname('tcp') || die "Failed getting tcp proto";
foreach (1 .. 20) {
	socket(SOCK, PF_INET, SOCK_STREAM, $proto) || die "Failed creating socket";
	my $ok = connect(SOCK, $paddr);
	close(SOCK);
	if ($ok) {
		print "Java proxy is now responding at localhost:$port\n";
		last;
	}
	print "Waits for java proxy to respond at localhost:$port\n";
	sleep 1;
}
doneInstalling();
# EOF
