/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "Filter.h"
#import "GBDebug.h"
#import "AlertHelper.h"
#import "Prefs.h"
#import "XmlDoc.h"
#import "Xml.h"
#import "ServerCommunication.h"
#import "FilterTabController.h"
#import "FilterPublisher.h"
#import "FilterListView.h"
#import "AutoZap.h"

NSString* updatePolicies[] = {
	@"auto", @"notify", @"manual", NULL
};

NSString* safetyPolicies[] = {
	@"simple", @"modification", NULL
};

//==============================================================================================

@interface LabelTextField : NSTextField
{
}
@property Filter* filter;

- (instancetype)initWithFrame:(NSRect)frameRect
		 withFilter:(Filter*)filter;

@end

@implementation LabelTextField

- (instancetype)initWithFrame:(NSRect)frameRect
				   withFilter:(Filter*)filter
{
	if (!(self = [super initWithFrame:frameRect]))
		return nil;
	_filter = filter; // not retained, as it would create a ref-counter loop.
	return self;
}


- (void)textDidEndEditing:(NSNotification *)aNotification
{
	NSString*s = [self stringValue];
	[_filter endEditing];
	if ([s length]) {
		[_filter setName:s];
		[[_filter comm] saveDirtyPrefsAndNotify];
	}
	[_filter makeTableFirstResponder];
}

@end

//==============================================================================================

@interface Filter()
@property FilterTabController* filterTabController;
@property FilterListView* filterListView;
@property ServerCommunication* comm;
@property AlertHelper* alertHelper;
@property Prefs* prefs;
@property NSXMLElement* filterE;
@property NSXMLElement* filterContainerE; // only subscriptions
@property NSXMLElement* filterDataE;
@property BOOL enabled;
@property BOOL isSubscription, subscriptionIsMissingDownload;
@property BOOL safetySimpleRulesOnly;
@property NSTimeInterval lastMouseDownTime;
@property BOOL mouseUpCanStartEditing;
@property BOOL mouseUpIsDoubleClick;
@property NSPoint mouseClickLocation;
@property NSMutableDictionary* subscriptionRuleStates;
@property NSString* subscriptionDataFilePath;
@property NSTimeInterval subscriptionDataFileModdate;
@property NSArray *topLevelObjects;

@property NSButton *checkbox;
@property NSTextField* label;
@property NSButton* globeButton;
@property NSButton* warningButton;
@property NSProgressIndicator* spinner;

@property NSURLConnection* updateConnection;
@property NSMutableData* updateData;
@property NSURL* updateUrl;


@property IBOutlet NSWindow*		editSheet;
@property (weak) IBOutlet NSTextField*	editNameTextField;
@property (weak) IBOutlet NSTextField*	editUrlTextField;
@property (weak) IBOutlet NSTextField*	editUrlErrorField;
@property (weak) IBOutlet NSMatrix*		editUpdatesMatrix;
@property (weak) IBOutlet NSTextField*	editUpdateDatesTextField;
@property (weak) IBOutlet NSTextField*	editUpdateErrorTextField;
@property (weak) IBOutlet NSButton*		editUpdateNowBtn;
@property (weak) IBOutlet NSMatrix*		editSafetyMatrix;
@property (weak) IBOutlet NSTextField*	editSafetyWarningTextField;
@property (weak) IBOutlet NSButton*		editOkBtn;


- (NSString*) getPathArgs:(NSString*)cmd;
- (void)updateViews;

- (void)releaseUpdateWithCancel:(BOOL)cancel;
- (void)makeGlobeImageView:(NSImage*)image;

- (void)enabledAction:(id)sender;
- (void)clickedGlobe:(id)sender;

- (void)updateEditingSubscriptionDates;
-(void)startEditingSubscriptionOneShot:(id)sender;
@end

@implementation Filter

- (instancetype)initWithFilterElement:(NSXMLElement*)filterE
							 withComm:(ServerCommunication*)comm
					withTabController:(FilterTabController*)tabController
				   withFilterListView:(FilterListView*)filterListView
							  atIndex:(NSUInteger)idx
{
	if (!(self = [super initWithFrame:NSMakeRect(0, 0, 265, 19)]))
		return nil;
	_comm = comm;
	_prefs = [comm prefs];
	_alertHelper = [_prefs alertHelper];
	_filterE = filterE;
	_filterId = [[Xml getAttribute:_filterE withName:@"id"] intValue];
	_url = [Xml getAttribute:_filterE withName:@"url"];
	_filterTabController = tabController; // not retained as it always lives longer.
	if ([[filterE elementsForName:@"filter-data"] count]) {
		_isSubscription = NO;
		NSArray* a = [_filterE elementsForName:@"filter-data"];
		if ([a count] != 1) {
			NSLog(@"Invalid <filter> in panel settings: %@", filterE);
			return nil;
		}
		_filterDataE = a[0];
	} else {
		_isSubscription = YES;
		NSString* pathArgs = [self getPathArgs:@"/filters/subscription/get"];
		DebugNSLog(@"Getting sub info: %@", pathArgs);
		XmlDoc* doc = [_comm curlToXml:pathArgs withActionDescription:NULL];
		if (!doc) {
			_filterContainerE = [Xml createElementWithName:@"filter-container"];
			_filterDataE = [Xml createElementWithName:@"filter-data"];
			[Xml setAttribute:@"(Subscription not downloaded)" withName:@"name" inElement:_filterDataE];
			_subscriptionIsMissingDownload = YES;
		} else {
			_filterContainerE = [doc byXPath:@"filter-container"];
			_filterDataE = [doc byXPath:@"filter-container/filter-data"];
			if (!_filterDataE) {
				NSLog(@"Invalid subscription xml: missing filter-data root child.");
				return nil;
			}
		}
		_subscriptionRuleStates = [NSMutableDictionary dictionaryWithCapacity:10];
	}
	[self readAttributesFromXml];
	if (_isSubscription) {
		_subscriptionDataFilePath = [NSString stringWithFormat:@"/Library/GlimmerBlocker/Filter subscriptions/%d.xml", _filterId];
		NSDictionary* attrs = [[NSFileManager defaultManager] attributesOfItemAtPath:_subscriptionDataFilePath error:NULL];
		_subscriptionDataFileModdate = [attrs[NSFileModificationDate] timeIntervalSince1970];
	}
	_filterListView = filterListView;
	[_filterListView addRow:self atIndex:idx];
	//
	_checkbox = [[NSButton alloc] initWithFrame:NSMakeRect(3, 1, 16, 16)]; // size by mag.glass
	[_checkbox setButtonType:NSSwitchButton];
	[_checkbox setTitle:@""];
	[_checkbox setTarget:self];
	[_checkbox setAction:@selector(enabledAction:)];
	[self addSubview:_checkbox];
	//
	if (![self isSubscription]) {
		[self updateViews];
		return self;
	}
	//
	_warningButton = [[NSButton alloc] initWithFrame:NSMakeRect(226, 2, 15, 15)];
	[_warningButton setImagePosition:NSImageOnly];
	[_warningButton setButtonType:NSMomentaryLightButton];
	[_warningButton setBordered:NO];
	[_warningButton setImage:[_filterTabController warningImage]];
	[self addSubview:_warningButton];
	[_warningButton setHidden:!_updateError];
	[_warningButton setTarget:self];
	[_warningButton setAction:@selector(startEditingSubscriptionOneShot:)];
	[_warningButton sendActionOn:NSLeftMouseDownMask];
	//
	_spinner = [[NSProgressIndicator alloc] initWithFrame:NSMakeRect(225, 1, 17, 17)];
	[_spinner setStyle:NSProgressIndicatorSpinningStyle];
	[_spinner setControlSize:NSSmallControlSize];
	[_spinner setDisplayedWhenStopped:NO];
	[_spinner sizeToFit];
	//DebugNSLog(@"spinner size: %@", NSStringFromSize([_spinner frame].size));
	[self addSubview:_spinner];
//	if (idx == 1)
//		[self performSelector:@selector(startSpinner) withObject:NULL afterDelay:0.01];
	// don't know why it must be delayed.
	[self updateViews];
	return self;
}

- (void)startSpinner
{
	[_spinner startAnimation:self];
}

- (void)dealloc
{
	DebugNSLog(@"Deallocs filter: %@", _name);
	[self releaseUpdateWithCancel:YES];
	[_filterListView removeRow:self];
	_filterTabController = NULL; // not retained
	//
	//
}

- (void)makeGlobeImageView:(NSImage*)image
{
	if ([_globeButton image] == image)
		return;
	if (!_globeButton) {
		_globeButton = [[NSButton alloc] initWithFrame:NSMakeRect(245, 1, 17, 17)];
		[_globeButton setImagePosition:NSImageOnly];
		[_globeButton setButtonType:NSMomentaryLightButton];
		[_globeButton setBordered:NO];
		[self addSubview:_globeButton];
		[_globeButton setTarget:self];
		[_globeButton setAction:@selector(clickedGlobe:)];
		[_globeButton sendActionOn:NSLeftMouseDownMask];
	}
	[_globeButton setImage:image];
}

- (NSImage*)getInfoImage
{
	if (_updateUrl)
		return NULL;
	if (_updateError)
		return [_filterTabController warningImage];
	if (_filterContainerE && [[Xml nodesForXPath:@"available-update-last-modified" withRoot:_filterContainerE] count])
		return [_filterTabController updateAvailableImage];
	return NULL;
}

- (void)updateViews
{
	[_checkbox setEnabled:[_prefs hasGlimmerAuth]];
	[_checkbox setState:[self enabled]];
	NSImage* infoImage = [self getInfoImage];
	[_warningButton setHidden:!infoImage];
	if (infoImage)
		[_warningButton setImage:infoImage];
	if (_updateUrl)
		[_spinner startAnimation:self];
	else
		[_spinner stopAnimation:self];
}

- (void)decodeSubscriptionRuleStates:(NSString*)encoded withState:(NSNumber*)state
{
	if (![encoded length])
		return;
	for (NSString* s in [encoded componentsSeparatedByString:@";"])
		_subscriptionRuleStates[s] = state;
}

- (void)readSubscriptionRuleStates
{
	if (![self isSubscription])
		return;
	[_subscriptionRuleStates removeAllObjects];
	[self decodeSubscriptionRuleStates:[Xml getAttribute:_filterE withName:@"rules-enabled"]
							 withState:@(NSOnState)];
	[self decodeSubscriptionRuleStates:[Xml getAttribute:_filterE withName:@"rules-disabled"]
							 withState:@(NSOffState)];
}

- (void)setUpdateError:(NSString*)s
{
	DebugNSLog(@"setUpdateError(%@): %@", _name, s);
	_updateError = s;
}

- (void)readAttributesFromXml
{
	_enabled = [[Xml getAttribute:_filterE withName:@"enabled"] intValue];
	//
	_name = [Xml getAttribute:_filterDataE withName:@"name"];
	//
	_description = [Xml stringForXPath:@"description" withRoot:_filterDataE];
	if (![_description length])
		_description = [Xml getAttribute:_filterDataE withName:@"description"]; // old format stores comments in attribute.
	//
	_comments = [Xml stringForXPath:@"comments" withRoot:_filterE];
	if (![_comments length])
		_comments = [Xml getAttribute:_filterE withName:@"comments"]; // old format stores comments in attribute.
	//
	_safetySimpleRulesOnly = _isSubscription && ![[Xml getAttribute:_filterE withName:@"safety-policy"] isEqual:@"modification"];
	//
	[self readSubscriptionRuleStates];
	if ([self isSubscription]) {
		[self makeGlobeImageView:[_filterTabController subscriptionGlobeImage]];
		NSString* tip = [NSString stringWithFormat:@"Subscription:\n%@", [self url]];
		NSString* s = [Xml dateFromMillisecsXPath:@"last-modified/@msecs" withRoot:_filterContainerE];
		if (s)
			tip = [NSString stringWithFormat:@"%@\n\nLast modification by author:\n        %@", tip, s];
		s = [Xml dateFromMillisecsXPath:@"last-update-check/@msecs" withRoot:_filterContainerE];
		if (s)
			tip = [NSString stringWithFormat:@"%@\nLast check for updates:\n        %@", tip, s];
		tip = [tip stringByReplacingOccurrencesOfString:@" " withString:@"\u00a0"]; // tried to avoid line breaking...
		[_globeButton setToolTip:tip];
		[self setUpdateError:[Xml getAttribute:_filterContainerE withName:@"network-error"]];
	} else if ([self url]) {
		if ([self isDirty])
			[self makeGlobeImageView:[_filterTabController publishGlobeDirtyImage]];
		else
			[self makeGlobeImageView:[_filterTabController publishGlobeImage]];
		NSString* tip = [NSString stringWithFormat:@"Published at:\n%@", [self url]];
		NSString* s = [Xml dateFromMillisecsXPath:@"last-publish-date/@msecs" withRoot:_filterE];
		if (s)
			tip = [NSString stringWithFormat:@"%@\n\nTime of latest publication:\n        %@", tip, s];
		s = [Xml dateFromMillisecsXPath:@"first-dirty-date/@msecs" withRoot:_filterE];
		if (s)
			tip = [NSString stringWithFormat:@"%@\nHas unpublished changes since:\n        %@", tip, s];
		tip = [tip stringByReplacingOccurrencesOfString:@" " withString:@"\u00a0"]; // tried to avoid line breaking...
		[_globeButton setToolTip:tip];
	} else {
		[_globeButton setToolTip:@""];
	}
	//
	[self updateViews];
}

- (NSNumber*)getRuleSubscriptionState:(NSString*)ruleId
{
	NSNumber* n = _subscriptionRuleStates[ruleId];
	return n ? n : @(NSMixedState);
}

- (void)setRuleSubscriptionState:(NSNumber*)state forRuleId:(NSString*)ruleId
{
	if ([state intValue] == NSMixedState)
		[_subscriptionRuleStates removeObjectForKey:ruleId];
	else
		_subscriptionRuleStates[ruleId] = state;
	NSMutableString* enabled = [NSMutableString stringWithCapacity:100];
	NSMutableString* disabled = [NSMutableString stringWithCapacity:100];
	for (NSXMLElement* ruleE in [[self filterDataE] elementsForName:@"rule"]) {
		ruleId = [Xml getAttribute:ruleE withName:@"rule-id"];
		if (!ruleId)
			continue;
		state = _subscriptionRuleStates[ruleId];
		if (!state)
			continue;
		NSMutableString* ms = ([state intValue] == NSOnState) ? enabled : disabled;
		if ([ms length])
			[ms appendString:@";"];
		[ms appendString:ruleId];
	}
	[Xml setAttribute:enabled withName:@"rules-enabled" inElement:_filterE];
	[Xml setAttribute:disabled withName:@"rules-disabled" inElement:_filterE];
	[_prefs markAsDirty];
	[_comm saveDirtyPrefsAndNotify];
}

//---------------------------------------
- (BOOL)deleteFilter
{
	if (![self isSubscription]) {
		[_filterE detach];
		[_filterListView removeRow:self];
		[_prefs markAsDirty];
		return YES;
	}
	if (![_comm isServerRunning]) {
		[_alertHelper prepareWarningAlertWithTitle:@"Can't delete subscriptions when GlimmerBlocker is not activated"
							   withInformativeText:@"You must activate GlimmerBlocker before deleting the filter"];
		[_alertHelper showPreparedAlert];
		return NO;
	}
	NSString* pathArgs = [self getPathArgs:@"!/filters/subscription/remove"];
	XmlDoc* doc = [_comm curlToXml:pathArgs withActionDescription:@"Failed deleting subscription"];
	if (!doc)
		return NO;
	[_filterE detach];
	[_filterListView removeRow:self];
	[_prefs markAsDirty];
	[_comm saveDirtyPrefsAndNotify];
	return YES;
}

//---------------------------------------

- (NSXMLElement*)ruleWithId:(int)ruleId
{
	if (!_filterDataE)
		return NULL;
	for (NSXMLElement* ruleE in [_filterDataE elementsForName:@"rule"]) {
		if ([Xml getIntAttribute:ruleE withName:@"rule-id"] == ruleId)
			return ruleE;
	}
	return NULL;
}

- (BOOL)isDirty
{
	if (![self url] || [self isSubscription])
		return NO;
	return [[_filterE elementsForName:@"first-dirty-date"] count] > 0;
}

- (void)notifyDirty
{
	[_prefs markAsDirty];
	if (![self url] || [self isSubscription])
		return;
	[Xml singletonDateElement:[NSDate date] forXPath:@"first-dirty-date" fromParent:_filterE updateExisting:NO];
	[self readAttributesFromXml];
}

- (void)notifyPublished
{
	NSAssert([self url] && ![self isSubscription], @"is-subscription");
	for (NSXMLElement* e in [_filterE elementsForName:@"first-dirty-date"])
		[e detach];
	[Xml singletonDateElement:[NSDate date] forXPath:@"last-publish-date" fromParent:_filterE updateExisting:YES];
}

- (void)notifyUpdatedAuthorizationState
{
	[self updateViews];
}
//---------------------------------------

- (void)filterEnabled:(BOOL)enabled
{
	if (enabled == _enabled)
		return;
	_enabled = enabled;
	[Xml setAttribute:(enabled ? @"1" : @"") withName:@"enabled" inElement:_filterE];
	[_prefs markAsDirty];
}

- (void)enabledAction:(id)sender
{
	[self filterEnabled:[sender state]];
	[_filterTabController updateTabViewWithForceUpdate:NO];
	[_comm saveDirtyPrefsAndNotify];
	if ([sender state]) {
		// make the server check for an update if the filter hasn't been updated for a while.
		NSURLResponse* response = NULL;
		[_comm curlToData:@"!/filters/subscription/check-updates?softcheck=1" withResponse:&response withErrorRef:NULL];
	}
}
//---------------------------------------

- (int)filterId
{
	return _filterId;
}

//---------------------------------------

- (NSString*)name
{
	return _name ? _name : @"";
}

- (void)setName:(NSString*)s
{
	NSAssert([self canEditName], @"setName without canEditName");
	if ([s isEqual:[self name]])
		return;
	_name = s;
	[_label setStringValue:_name];
	[Xml setAttribute:_name withName:@"name" inElement:_filterDataE];
	[self notifyDirty];
}

//---------------------------------------

- (NSString*)description
{
	return _description ? _description : @"";
}

//---------------------------------------

- (NSString*)comments
{
	return _comments ? _comments : @"";
}

- (void)setComments:(NSString*)s
{
	NSAssert([self canEditComments], @"setComments without canEditComments");
	if ([s isEqual:[self comments]])
		return;
	_comments = s;
	[Xml setCData:_comments forXPath:@"comments" withRoot:_filterE];
	[Xml setAttribute:NULL withName:@"comments" inElement:_filterE]; // storing in attribute is old format.
	[self notifyDirty];
}

//---------------------------------------

+ (NSString*)extractHostFromStringUrl:(NSString*)url
{
	NSString* host;
	if ([url hasPrefix:@"http://"])
		host = [url substringFromIndex:7];
	else if ([url hasPrefix:@"https://"])
		host = [url substringFromIndex:8];
	else
		return NULL;
	NSInteger len = [host length];
	for (int i = 0; i < len; i++) {
		unichar ch = [host characterAtIndex:i];
		if (ch == ':' || ch == '/')
			return [host substringToIndex:i];
	}
	return host;
}

- (NSString*)publishHost
{
	return _isSubscription ? NULL : [Filter extractHostFromStringUrl:_url];
}

- (BOOL)isPublished
{
	return !_isSubscription && _url;
}

- (BOOL)isMobileMePublished
{
	NSString* host = [self publishHost];
	return [host isEqual:@"idisk.mac.com"] || [host isEqual:@"idisk.me.com"];
}

- (BOOL)isWebDavPublished
{
	return [self publishHost] && ![self isMobileMePublished];
}

- (NSString*)url
{
	return _url;
}

- (void)setUrl:(NSString*)url
{
	_url = url;
	[Xml setAttribute:url withName:@"url" inElement:_filterE];
	[self notifyDirty];
	[self setNeedsDisplay:YES];
}

- (BOOL)canEditName
{
	return ![self isSubscription];
}

- (BOOL)canEditComments
{
	return ![self isSubscription];
}

- (BOOL)canEditRules
{
	return ![self isSubscription];
}

- (BOOL)canDelete
{
	NSString* url = [Xml getAttribute:_filterE withName:@"url"];
	return !url || ![url hasPrefix:@"http://glimmerblocker.org/site/filters/default/"];
}

- (void)startEditingPart2
{
	[[self window] makeFirstResponder:_label];
}

- (void)startEditingName
{
	if (_label)
		return;
	[_filterListView setSelectionByView:self];
	// must be 1 pixel to the left of the drawString.
	_label = [[LabelTextField alloc] initWithFrame:NSMakeRect(25, 2, 200, 17) withFilter:self];
	[_label setEditable:YES];
	[_label setSelectable:YES];
	[_label setDrawsBackground:YES];
	[_label setBezeled:NO];
	[_label setBordered:YES]; // If the border is omitted, the focus frame is not drawn????
	[_label setAlignment:NSLeftTextAlignment];
	[_label setStringValue:[self name]];
	[_label setDelegate:self];
	[self addSubview:_label];
	[self performSelector:@selector(startEditingPart2) withObject:NULL afterDelay:0.1];
}

- (BOOL)isEditing
{
	return !!_label;
}

- (void)endEditing
{
	if (!_label)
		return;
	DebugNSLog(@"Filter endEditing");
	[_label removeFromSuperview];
	_label = NULL;
}

- (BOOL)control:(NSControl*)control
	   textView:(NSTextView*)fieldEditor doCommandBySelector:(SEL)commandSelector
{
	if (_label && commandSelector == @selector(insertNewline:)) {
		[self endEditing];
		return YES;
	}
	return NO;
}

- (void)drawLabelUsingFocusColor:(BOOL)usingFocusColor
{
	NSMutableDictionary* attrs = [NSMutableDictionary dictionaryWithCapacity:3];
	attrs[NSFontAttributeName] = [NSFont systemFontOfSize:[NSFont systemFontSize]];
	NSColor* color;
	if (usingFocusColor)
		color = [NSColor alternateSelectedControlTextColor];
	else if (_subscriptionIsMissingDownload)
		color = [NSColor redColor];
	else
		color = [NSColor blackColor];
	attrs[NSForegroundColorAttributeName] = color;
	NSRect labelRect = NSMakeRect(26, 0, 200, 19);
	NSString* s = [self name];
	[s drawInRect:labelRect withAttributes:attrs];
}

- (void)drawRect:(NSRect)r
{
	[super drawRect:r];
	if (!_label)
		[self drawLabelUsingFocusColor:[self shouldDrawUsingFocusColor]];
}

- (void)mouseUp:(NSEvent *)theEvent
{
	NSTimeInterval dur = [theEvent timestamp] - _lastMouseDownTime;
	if (dur > 1) // any NSxxx ?
		return;
	if ([self isSubscription]) {
		if (_mouseUpIsDoubleClick)
			[self startEditingSubscription:NO];
		return;
	}
	if (_mouseUpCanStartEditing && [self canEditName] && [_prefs hasGlimmerAuth])
		[self startEditingName];
}

+ (NSString*)draggingDataType
{
	return @"dk.vitality.glimmerblocker.filter";
}

- (void)mouseDown:(NSEvent *)theEvent
{
	//DebugNSLog(@"mouseDown: %@", theEvent);
	_lastMouseDownTime = [theEvent timestamp];
	BOOL tableWasFirstResponder = [_filterListView isFirstResponder];
	[[self window] makeFirstResponder:_filterListView];
	BOOL wasSelected = [self selected];
	[_filterListView setSelectionByView:self];
	_mouseClickLocation = [self convertPoint:[theEvent locationInWindow] fromView:NULL];
	NSUInteger cc = [theEvent clickCount];
	_mouseUpCanStartEditing = (cc == 2 || (cc == 1 && wasSelected && tableWasFirstResponder));
	_mouseUpIsDoubleClick = (cc == 2);
}

- (void)mouseDragged:(NSEvent *)theEvent
{
	//DebugNSLog(@"Mouse dragged: %@", theEvent);
	NSPoint pos = [self convertPoint:[theEvent locationInWindow] fromView:NULL];
	if (ABS(pos.x - _mouseClickLocation.x) >= 10 || ABS(pos.y - _mouseClickLocation.y) < 5)
		return;
	DebugNSLog(@"Starts drag");
	// Make image snapshot
	NSSize imgSize = [self bounds].size;
	NSImage* img2 = [[NSImage alloc] initWithSize:imgSize];
	[img2 lockFocus];
	[self drawLabelUsingFocusColor:NO];
	[img2 unlockFocus];
	// Lighten it up, so it seems transparent
	NSImage* image = [[NSImage alloc] initWithSize:imgSize];
	[image lockFocus];
	[img2 drawAtPoint:NSMakePoint(0, 0) fromRect:NSZeroRect operation:NSCompositePlusLighter fraction:0.5f];
	[image unlockFocus];
	// Start the dragging operation
	[_filterListView startDrag:self withImage:image withEvent:theEvent];
}

- (void)makeTableFirstResponder
{
	//DebugNSLog(@"makeTableFirstResponder");
	[[self window] makeFirstResponder:_filterListView];
}

- (void)releaseUpdateWithCancel:(BOOL)cancel
{
	if (!_updateUrl)
		return;
	if (cancel)
		[_updateConnection cancel];
	_updateConnection = NULL;
	_updateData = NULL;
	_updateUrl = NULL;
}

- (NSString*) getPathArgs:(NSString*)cmd
{
	NSString* url = [Xml getAttribute:_filterE withName:@"url"];
	NSAssert(url, @"Must be a subscription");
	return [NSString stringWithFormat:@"%@?filter-id=%d", cmd, _filterId];
}

- (void)startUpdateSubscription
{
	if (_updateUrl) {
		NSLog(@"Ignores update-subscription request as already updating it: %@", _url);
		return;
	}
	NSLog(@"Update subscription: %@", _url);
	NSAssert([self isSubscription], @"Not a subscription");
	NSString* pathArgs = [self getPathArgs:@"!/filters/subscription/update"];
	_updateUrl = [NSURL URLWithString:[_comm getAdminUrl:pathArgs]];
	NSURLRequest* req = [NSURLRequest requestWithURL:_updateUrl
										 cachePolicy:NSURLRequestReloadIgnoringLocalCacheData // want fresh data
									 timeoutInterval:60.0];
	_updateData = [NSMutableData data];
	_updateConnection = [[NSURLConnection alloc] initWithRequest:req delegate:self];
	if (!_updateConnection) {
		[self releaseUpdateWithCancel:NO];
		[self setUpdateError:@"Communication with proxy failed."];
	}
	[self updateViews];
}

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response
{
	// called for each response, i.e. for each redirect
	DebugNSLog(@"FilterSub.didReceiveResponse: %@", response);
	[_updateData setLength:0];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{
	[_updateData appendData:data];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
	NSString* a = [error localizedDescription];
	NSString* b = [error userInfo][NSURLErrorFailingURLStringErrorKey];
	[self setUpdateError:[NSString stringWithFormat:@"%@ %@", a, b]];
	DebugNSLog(@"FilterSub.didFailWithError: %@", _updateError);
	[self releaseUpdateWithCancel:NO];
	[self updateViews];
}

- (void)didGetNewSubscriptionData
{
	_subscriptionIsMissingDownload = NO;
	[self readAttributesFromXml];
	[_filterTabController notifyUpdatedFilter:self];
	[self updateEditingSubscriptionDates];
}

- (void)decodeUpdateResponse
{
	XmlDoc* doc = [[XmlDoc alloc] initWithURL:_updateUrl withData:_updateData];
	if (![doc rootE]) {
		[self setUpdateError:@"Could not get proper xml response from server"];
		return;
	}
	[self setNeedsDisplay:YES];
	NSString* s = [Xml getAttribute:[doc rootE] withName:@"error"];
	if (s) {
		[self setUpdateError:s];
		return;
	}
	NSXMLElement* filterContainerE = [doc byXPath:@"filter-container"];
	NSXMLElement* filterDataE = [doc byXPath:@"filter-container/filter-data"];
	if (!filterContainerE || !filterDataE)
		return; // not updated.
	_filterDataE = filterDataE;
	_filterContainerE = filterContainerE;
	[_prefs markAsDirty];
	[self didGetNewSubscriptionData];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection
{
	[self setUpdateError:NULL];
	[self decodeUpdateResponse];
	[self releaseUpdateWithCancel:NO];
	[self updateViews];
}

- (void)clickedGlobeAlertCallbackWithReturnCode:(NSInteger)returnCode
{
	if (returnCode != NSAlertFirstButtonReturn)
		return;
	FilterPublisher *publisher = [[FilterPublisher alloc] initWithFilterTabController:_filterTabController
																		   withFilter:self];
	[AutoZap add:publisher];
}

// callback from ImageButton
- (void)clickedGlobe:(id)sender
{
	[_filterListView setSelectionByView:self];
	if ([self isSubscription]) {
		[self startEditingSubscription:NO];
		//[_alertHelper prepareInformativeAlertWithTitle:[self name] withInformativeText:_url];
		//[_alertHelper showPreparedAlert];
		return;
	}
	NSString* info = [self url];
	NSString* s = [Xml dateFromMillisecsXPath:@"last-publish-date/@msecs" withRoot:_filterE];
	if (s)
		info = [NSString stringWithFormat:@"%@\n\nTime of latest publication:\n        %@", info, s];
	NSString* dirty = [Xml dateFromMillisecsXPath:@"first-dirty-date/@msecs" withRoot:_filterE];
	if (dirty)
		info = [NSString stringWithFormat:@"%@\n\nHas unpublished changes since:\n        %@", info, dirty];
	[_alertHelper prepareInformativeAlertWithTitle:@"Published filter:" withInformativeText:info];
	// ensure some reasonable min-width of alert so URL isn't wrapped.
	[_alertHelper setAccessoryView:[[NSView alloc] initWithFrame:NSMakeRect(0, 0, 425, 1)]];
	NSButton* okBtn = [_alertHelper addButtonWithTitle:(dirty ? @"Publish changes" : @"Publish")];
	if (!dirty)
		[okBtn setKeyEquivalent:@""];
	if (![_prefs hasGlimmerAuth])
		[okBtn setEnabled:NO];
	[_alertHelper addButtonWithTitle:@"Cancel"];
	[_alertHelper setDelegate:self withSelector:@selector(clickedGlobeAlertCallbackWithReturnCode:)];
	[_alertHelper showPreparedAlert];
}

- (void)checkReloadSubscriptionFilterData
{
	if (!_isSubscription)
		return;
	NSDictionary* attrs = [[NSFileManager defaultManager] attributesOfItemAtPath:_subscriptionDataFilePath error:nil];
	NSTimeInterval moddate = [attrs[NSFileModificationDate] timeIntervalSince1970];
	if (moddate == _subscriptionDataFileModdate)
		return;
	_subscriptionDataFileModdate = moddate;
	DebugNSLog(@"Subscription changed on disk (%d.xml): %@", _filterId, [self name]);
	NSString* pathArgs = [self getPathArgs:@"/filters/subscription/get"];
	DebugNSLog(@"Getting sub info: %@", pathArgs);
	XmlDoc* doc = [_comm curlToXml:pathArgs withActionDescription:NULL];
	if (!doc)
		return;
	NSXMLElement* filterContainerE = [doc byXPath:@"filter-container"];
	NSXMLElement* filterDataE = [doc byXPath:@"filter-container/filter-data"];
	if (!filterContainerE || !filterDataE) {
		NSLog(@"Invalid subscription xml: missing filter-data root child.");
		return;
	}
	//
	_filterContainerE = filterContainerE;
	_filterDataE = filterDataE;
	[self didGetNewSubscriptionData];
}


#pragma mark ------------------- Subscription edit

- (void)checkSubscriptionUrlSyntax
{
	NSString* s = [_editUrlTextField stringValue];
	NSString* errorMsg = [Xml getHttpUrlSyntaxError:s];
	if (!errorMsg && ![s hasSuffix:@".xml"])
		errorMsg = @"URL must end with '.xml'";
	[_editUrlErrorField setStringValue:(errorMsg ? errorMsg : @"")];
	[_editOkBtn setEnabled:(!errorMsg && [_prefs hasGlimmerAuth])];
}

- (void)controlTextDidChange:(NSNotification *)aNotification
{
	if (_editUrlTextField == [aNotification object])
		[self checkSubscriptionUrlSyntax];
}

- (void)updateEditingSubscriptionDates
{
	NSMutableString* sb = [NSMutableString stringWithCapacity:0];
	//
	NSString* s = [Xml dateFromMillisecsXPath:@"last-modified/@msecs" withRoot:_filterContainerE];
	[sb appendString:(s ? s : @"")];
	//
	[sb appendString:@"\n"];
	s = [Xml dateFromMillisecsXPath:@"last-update-check/@msecs" withRoot:_filterContainerE];
	[sb appendString:(s ? s : @"")];
	//
	[sb appendString:@"\n"];
	s = [Xml dateFromMillisecsXPath:@"subscribed-date/@msecs" withRoot:_filterContainerE];
	[sb appendString:(s ? s : @"")];
	//
	[_editUpdateDatesTextField setStringValue:sb];
	//
	s = [Xml getAttribute:_filterContainerE withName:@"network-error"];
	if (s) {
		[_editUpdateErrorTextField setTextColor:[_editSafetyWarningTextField textColor]];
		s = [@"Last update error:\n" stringByAppendingString:s];
	} else {
		s = [Xml dateFromMillisecsXPath:@"available-update-last-modified/@msecs" withRoot:_filterContainerE];
		if (s) {
			[_editUpdateErrorTextField setTextColor:[NSColor colorWithDeviceRed:0 green:0.5f blue:0 alpha:1]];
			s = [@"Available update:\n" stringByAppendingString:s];
		}
	}
	[_editUpdateErrorTextField setStringValue:(s ? s : @"")];
}

- (void)startEditingSubscription:(BOOL)oneshot
{
	if (!_editSheet) {
		_topLevelObjects = [AutoZap loadNibNamed:@"FilterSubscriptionEdit" bundle:[_comm mainBundle] owner:self];
		NSAssert(_editSheet, @"editSheet is null");
	}
	[_editNameTextField setStringValue:[self name]];
	[_editUrlTextField setStringValue:[self url]];
	[_editUrlTextField setDelegate:self];
	[self updateEditingSubscriptionDates];
	[GBDebug setVertMatrix:_editUpdatesMatrix
					values:updatePolicies
			 selectedValue:[Xml getAttribute:_filterE withName:@"update-policy"]
			  defaultValue:@"notify"];
	[GBDebug setVertMatrix:_editSafetyMatrix
					values:safetyPolicies
			 selectedValue:[Xml getAttribute:_filterE withName:@"safety-policy"]
			  defaultValue:@"simple"];
	BOOL editable = [_prefs hasGlimmerAuth];
	[_editUrlTextField setEnabled:editable];
	[_editUpdatesMatrix setEnabled:editable];
	[_editSafetyMatrix setEnabled:editable];
	[_editUpdateNowBtn setEnabled:editable];
	[_editOkBtn setEnabled:editable];
	[_editUpdateNowBtn setState:oneshot];
	[self checkSubscriptionUrlSyntax];
	[NSApp beginSheet: _editSheet
	   modalForWindow: [self window]
		modalDelegate: nil
	   didEndSelector: nil
		  contextInfo: nil];
}

-(void)startEditingSubscriptionOneShot:(id)sender
{
	[_filterListView setSelectionByView:self];
	[self startEditingSubscription:YES];
}

- (IBAction)editCancelBtnAction:(id)sender
{
	[NSApp endSheet:_editSheet];
	[_editSheet orderOut:self];
}

- (IBAction)editOkBtnAction:(id)sender
{
	[self editCancelBtnAction:self];
	[self setUrl:[_editUrlTextField stringValue]];
	[Xml setAttribute:updatePolicies[[_editUpdatesMatrix selectedRow]] withName:@"update-policy" inElement:_filterE];
	[Xml setAttribute:safetyPolicies[[_editSafetyMatrix selectedRow]] withName:@"safety-policy" inElement:_filterE];
	[_prefs markAsDirty];
	[self readAttributesFromXml];
	[_comm saveDirtyPrefsAndNotify];
	[_filterTabController redrawRulesTable];
	if ([_editUpdateNowBtn state])
		[self startUpdateSubscription];
}

@end

@implementation Filter(NSDraggingSource)

- (NSDragOperation)draggingSourceOperationMaskForLocal:(BOOL)isLocal
{
	NSLog(@"Filter.draggingSourceOperationMaskForLocal.%d", isLocal);
	return isLocal ? NSDragOperationNone : NSDragOperationNone;
}

- (BOOL)ignoreModifierKeysWhileDragging
{
	return YES;
}

- (void)draggedImage:(NSImage *)anImage beganAt:(NSPoint)aPoint
{
	NSLog(@"beganAt: %@", NSStringFromPoint(aPoint));
}

@end
