/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#import "InstallationHelper.h"
#import <PreferencePanes/PreferencePanes.h>
#import "AutoZap.h"
#import "GBDebug.h"

@interface InstallationHelper ()

@property NSString* copyrightMessageHtml;
@property NSPreferencePane* preferencePane;
@property NSBundle* bundle;
@property AuthorizationRef authRef;
@property NSString* progressText;
@property BOOL invalidCodeSignature;
@property NSArray *topLevelObjects;

@property (weak) IBOutlet NSView*				fixView;
@property (weak) IBOutlet NSTextField*			fixMessageTextField;
@property (weak) IBOutlet NSButton*				fixBtn;

@property (weak) IBOutlet NSTextField*			pleaseWaitTextField;
@property (weak) IBOutlet NSProgressIndicator*	progressIndicator;
@property (weak) IBOutlet NSTextField*			progressTextField;

@property IBOutlet NSTextView*					footerCopyrightTextView;
@property IBOutlet NSTextView*					footerUrlTextView;




- (void)loadUI;
- (void)showFixInstallation:(NSString*)reason;
- (void)showSimpleMessage:(NSString*)reason;
- (void)showReinstallMessage:(NSString*)reason;

@end

@implementation InstallationHelper

-(instancetype)initWithCopyrightMessage:(NSString*)copyrightMessageHtml
					 withPreferencePane:(NSPreferencePane*)preferencePane
{
	if (!(self = [super init]))
		return NULL;
	_preferencePane = preferencePane;
	_bundle = [preferencePane bundle];
	_copyrightMessageHtml = copyrightMessageHtml;
	return self;
}


- (BOOL)hasDir:(NSString*)path
{
	NSFileManager* fm = [NSFileManager defaultManager];
	BOOL dir = NO;
	return [fm fileExistsAtPath:path isDirectory:&dir] && dir;
}

- (BOOL)hasRelDir:(NSString*)relPath
{
	return [self hasDir:[NSString stringWithFormat:@"%@/%@", [_bundle bundlePath], relPath]];
}

- (BOOL)checkLittleSnitch
{
	NSString* path = @"/Library/Little Snitch/Little Snitch UIAgent.app/Contents/Info.plist";
	if (![[NSFileManager defaultManager] fileExistsAtPath:path])
		return YES;
	NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:path];
	NSString* vers = dict[@"CFBundleVersion"];
	NSLog(@"Little Snitch Core System Version: '%@'", vers);
	if (![vers isEqual:@"385"])
		return YES;
	NSString* msg = @"You must upgrade Little Snitch to version 2.0.5 or later\n"
		"before you can use GlimmerBlocker.\n"
		"\n\n"
		"You're using LittleSnitch version 2.0.4 with 'Core System Version' 385 which often breaks internet connectivity.";
	NSLog(@"GB: %@", msg);
	[self showSimpleMessage:msg];
	return NO;
}

- (void)setFixBtnSize:(NSControlSize)controlSize
		   withDeltaY:(CGFloat)dy
{
	CGFloat fontSize = [NSFont systemFontSizeForControlSize:controlSize];
	NSCell *theCell = [_fixBtn cell];
	if (!theCell)
		NSLog(@"FixBtn has no cell");
	NSFont *theFont = [NSFont fontWithName:[[theCell font] fontName] size:fontSize];
	[theCell setFont:theFont];
	[theCell setControlSize:controlSize];
	[_fixBtn sizeToFit];
	NSRect f = [_fixBtn frame];
	f.origin.x = ([_fixView frame].size.width - f.size.width) / 2;
	f.origin.y += dy;
	[_fixBtn setFrame:f];
}

- (BOOL)validate:(NSString*)glimmerBlockerVersionText
{
	NSString* panelPath = @"/Library/PreferencePanes/GlimmerBlocker.prefPane";
	if ([[_bundle bundlePath] hasPrefix:@"/Users/"]) {
		NSLog(@"Invalid installation path: %@", [_bundle bundlePath]);
		NSLog(@"Required path:             %@", panelPath);
		[self showFixInstallation:@"GlimmerBlocker is installed for a single user,\nbut must be installed for all users."];
		return NO;
	}
	if (![[_bundle bundlePath] isEqual:panelPath]) {
		NSLog(@"Invalid installation path: %@", [_bundle bundlePath]);
		NSLog(@"Required path:             %@", panelPath);
		[self showFixInstallation:@"GlimmerBlocker is not installed with the proper filename/path."];
		return NO;
	}
	if (![self checkLittleSnitch])
		return NO;
	if (![self hasRelDir:@"Contents/GlimmerBlockerProxy.app"] || ![self hasRelDir:@"Contents/GlimmerBlockerProxy.app"]) {
		[self showFixInstallation:@"The proxy files are not extracted."];
		return NO;
	}
	if (REQUIRED_VALID_CODESIGN && ![GBDebug verifyCodesign]) {
		[self showReinstallMessage:@"The installation is not valid but has been tampered with."];
		[_fixBtn setTitle:@"Show detailed error report"];
		[_fixBtn setHidden:NO];
		// http://devworld.apple.com/documentation/Cocoa/Conceptual/ControlCell/Articles/ManipulateCellControl.html
		[self setFixBtnSize:NSMiniControlSize withDeltaY:-39];
		//
		_invalidCodeSignature = YES;
		return NO;
	}
	NSFileManager* fm = [NSFileManager defaultManager];
	if (![fm fileExistsAtPath:@"/Library/GlimmerBlocker/PanelSettings.xml"]) {
		[self showFixInstallation:@"The settings file is missing."];
		return NO;
	}
	if (![self hasDir:@"/Library/Logs/GlimmerBlocker"]) {
		[self showFixInstallation:@"The log folder is missing."];
		return NO;
	}
	if (![self hasDir:@"/Library/GlimmerBlocker/Filter subscriptions"]) {
		[self showFixInstallation:@"The subscriptions folder is missing."];
		return NO;
	}
	NSArray* args = @[@"-c", @"dscl . -read /Users/_glimmerblocker > /dev/null"];
	int status = [GBDebug executeCommandlineTool:@"/bin/sh" withArgs:args];
	if (status) {
		NSLog(@"dscl failed listing _glimmerblocker user");
		[self showFixInstallation:@"The GlimmerBlocker proxy is not installed."];
		return NO;
	}
	if (![GBDebug fingerprintForRunningServerIsValid:_bundle]) {
		NSLog(@"Running server has incompatible fingerprint.");
		[self showFixInstallation:@"The GlimmerBlocker proxy needs to be\nrestarted after the upgrade."];
		[_fixBtn setTitle:@"  Restart proxy  "]; // for some unknown reason, [sizeToFit] makes it too narrow.
		[self setFixBtnSize:NSRegularControlSize withDeltaY:0];
		return NO;
	}
	NSString* path = [panelPath stringByAppendingString:@"/Contents/Info.plist"];
	NSString* vers = @"???";
	if ([[NSFileManager defaultManager] fileExistsAtPath:path]) {
		NSDictionary *dict = [NSDictionary dictionaryWithContentsOfFile:path];
		vers = dict[@"CFBundleVersion"];
	}
	if (![vers isEqual:glimmerBlockerVersionText]) {
		NSLog(@"GB version number on disk doesn't match loaded panel:");
		NSLog(@"      Loaded panel: %@", glimmerBlockerVersionText);
		NSLog(@"   Version on disk: %@", vers);
		[self showSimpleMessage:@"GlimmerBlocker has been upgraded\nsince this panel was opened."];
		[_pleaseWaitTextField setStringValue:@"Please quit and relaunch System Preferences."];
		return NO;
	}
	if (NO) {
		NSLog(@"FORCES INVALID INSTALLATION PATH.");
		[self showFixInstallation:@"The sky is not blue today."];
		return NO;
	}
	return YES;
}

- (void)loadUI
{
	if (_fixView)
		return;
	_topLevelObjects = [AutoZap loadNibNamed:@"Installation" bundle:[_preferencePane bundle] owner:self];
	NSAssert(_fixView, @"fixView is null");
	NSView* mainView = [_preferencePane mainView];
	if (mainView) {
		DebugNSLog(@"GB Panel has already mainView, adds InstallView as only visible subview.");
		for (NSView* view in [mainView subviews])
			[view setHidden:YES];
		[mainView addSubview:_fixView];
	} else {
		[_preferencePane setMainView:_fixView];
	}
	NSAttributedString *as = [[NSAttributedString alloc] initWithHTML:[_copyrightMessageHtml dataUsingEncoding:NSUTF8StringEncoding] documentAttributes:NULL];
	NSTextStorage* ts = [_footerCopyrightTextView textStorage];
	[ts replaceCharactersInRange:NSMakeRange(0, [ts length])
			withAttributedString:as];
	//
	[_fixView setFrameOrigin:NSMakePoint(0, 0)];
	[_footerCopyrightTextView setEditable:NO];
	[_footerUrlTextView setEditable:NO];
	[_pleaseWaitTextField setStringValue:@""];
	[_progressTextField setStringValue:@""];
}

- (void)showFixInstallation:(NSString*)message
{
	NSLog(@"showFixInstallation: %@", message);
	[self loadUI];
	[_fixMessageTextField setStringValue:message];
}

- (void)showSimpleMessage:(NSString*)reason
{
	[self loadUI];
	[_fixBtn setHidden:YES];
	[_fixMessageTextField setStringValue:reason];
	NSRect f = [_fixMessageTextField frame];
	CGFloat d = 70;
	f.origin.y -= d;
	f.size.height += d;
	[_fixMessageTextField setFrame:f];
}

- (void)showReinstallMessage:(NSString*)reason
{
	NSLog(@"showReinstallMessage: %@", reason);
	NSString* msg = [NSString stringWithFormat:@"GlimmerBlocker needs to be reinstalled:\n"
					 "%@\n"
					 "\n"
					 "Please download and install the latest version from http://glimmerblocker.org",
					 reason];
	[self showSimpleMessage:msg];
}

//============================================================================

- (void)updateProgressText
{
	[_progressTextField setStringValue:_progressText];
	[[_progressTextField window] display]; // force window to update GraphicsCard pixels right now.
}

- (void)installerDone
{
	_progressText = @"";
	[self updateProgressText];
	[_progressIndicator stopAnimation:self];
	[_pleaseWaitTextField setStringValue:@"Please quit and relaunch System Preferences."];
}

- (void)runInstaller
{
	NSString* bundlePath = [_bundle bundlePath];
	const char* cmd = [[NSString stringWithFormat:@"%@/Contents/Resources/InstallerScript.pl", bundlePath] UTF8String];
	const char* pid = [[NSString stringWithFormat:@"%d", getpid()] UTF8String];
	const char* args[] = { "-syspref", pid, "-bundle", [bundlePath UTF8String], NULL };
	FILE *msgPipe = NULL;
	NSLog(@"Calls AuthorizationExecuteWithPrivileges");
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
	OSStatus status = AuthorizationExecuteWithPrivileges(_authRef,
														 cmd,
														 kAuthorizationFlagDefaults,
														 (char**)args,
														 &msgPipe);
#pragma GCC diagnostic warning "-Wdeprecated-declarations"
	if (status == errAuthorizationSuccess) {
		NSLog(@"AuthorizationExecuteWithPrivileges returned success");
		char buffer[128];
		NSMutableData* msgBuffer = [NSMutableData dataWithCapacity:80];
		while (1) {
			long num = read(fileno(msgPipe), buffer, sizeof(buffer));
			if (num < 1)
				break;
			[msgBuffer appendBytes:buffer length:num];
			NSInteger len = [msgBuffer length];
			int idx = 0;
			while (idx < len) {
				char ch = ((const char*)[msgBuffer bytes])[idx];
				if (ch != '\n') {
					idx++;
					continue;
				}
				NSString* s = [[NSString alloc] initWithBytes:[msgBuffer bytes] length:idx encoding:NSUTF8StringEncoding];
				NSLog(@"%@", s);
				if ([s hasPrefix:@"MSG:"]) {
					_progressText = [s substringFromIndex:4];
					[self performSelectorOnMainThread:@selector(updateProgressText)
										   withObject:NULL
										waitUntilDone:YES];
				}
				[msgBuffer replaceBytesInRange:NSMakeRange(0, idx + 1) withBytes:"" length:0];
				len -= idx + 1;
				idx = 0;
			}
		}
		NSLog(@"installer done.");
	}
	if (msgPipe)
		fclose(msgPipe);
	if (status != errAuthorizationSuccess) {
		NSLog(@"AuthorizationExecuteWithPrivileges failed with error code %ld", (long)status);
	} else {
		BOOL dir = NO;
		BOOL e = [[NSFileManager defaultManager] fileExistsAtPath:@"/Library/PreferencePanes/GlimmerBlocker.prefPane"
												  isDirectory:&dir];
		NSLog(@"Installer done: %d + %d",  e, dir);
	}
	[self performSelectorOnMainThread:@selector(installerDone)
						   withObject:NULL
						waitUntilDone:NO];
}

- (void)prepareAuth
{
	AuthorizationRef authRef = NULL;
	AuthorizationFlags authFlags = kAuthorizationFlagDefaults;
	OSStatus status = AuthorizationCreate(NULL, kAuthorizationEmptyEnvironment, authFlags, &authRef);
	if (status != errAuthorizationSuccess) {
		NSLog(@"GlimmerBlockerInstaller: AuthorizationCreate failed");
		return;
	}
	NSLog(@"GlimmerBlockerInstaller: AuthorizationCreate ok");
	AuthorizationItem myItems = {kAuthorizationRightExecute, 0, NULL, 0};
	AuthorizationRights myRights = {1, &myItems};
	authFlags = kAuthorizationFlagInteractionAllowed | kAuthorizationFlagExtendRights;
	status = AuthorizationCopyRights(authRef, &myRights, NULL, authFlags, NULL );
	if (status != errAuthorizationSuccess) {
		NSLog(@"GlimmerBlockerInstaller: AuthorizationCopyRights failed: %ld", (long)status);
		return;
	}
	NSLog(@"GlimmerBlockerInstaller: AuthorizationCopyRights ok");
	_authRef = authRef;
}

- (BOOL)testCurrentTmpDir
{
	FILE* tmpfileF = tmpfile(); // returns NULL if we can't create a file in $TMPDIR
	if (tmpfileF) {
		fclose(tmpfileF);
		NSLog(@"  works: got tmpfile().");
		return YES;
	} else {
		return NO;
	}
}

- (BOOL)createTmpDir
{
	NSLog(@"Finds usable TMPDIR value:");
	// If the $TMPDIR directory doesn't exist (or is unwriteable),
	// AuthorizationExecuteWithPrivileges fails with errAuthorizationInternal = -60008
	// so we must ensure it points to an existing directory which this process has write permissions to.
	//--------------------------------------------------------------------
	char* origTmpDirPath = getenv("TMPDIR");
	if (origTmpDirPath) {
		NSLog(@"   Tries tmpfile() using default $TMPDIR = '%s'.", origTmpDirPath);
		if ([self testCurrentTmpDir])
			return YES;
	} else {
		NSLog(@"   TMPDIR is unset.");
	}
	//--------------------------------------------------------------------
	NSString* s = NSTemporaryDirectory();
	if (s) {
		NSLog(@"   Tries using TMPDIR = NSTemporaryDirectory() = '%@'", s);
		setenv("TMPDIR", [s UTF8String], 1);
		if ([self testCurrentTmpDir])
			return YES;
	} else {
		NSLog(@"   NSTemporaryDirectory() returned NULL.");
	}
	//--------------------------------------------------------------------
	NSLog(@"   Tries using TMPDIR = '/tmp'");
	setenv("TMPDIR", "/tmp", 1);
	if ([self testCurrentTmpDir])
		return YES;
	//--------------------------------------------------------------------
	s = NSHomeDirectory();
	if (s) {
		NSLog(@"   Tries using TMPDIR = NSHomeDirectory() = '%@'", s);
		setenv("TMPDIR", [s UTF8String], 1);
		if ([self testCurrentTmpDir])
			return YES;
	} else {
		NSLog(@"   NSHomeDirectory() returned NULL.");
	}
	//--------------------------------------------------------------------
	NSLog(@"Can't create tmp files at all!");
	return NO;
}

//============================================================================
static InstallationHelper *retainMe;

- (void)runProxyDebugDone
{
	[_fixBtn setEnabled:YES];
	[_progressIndicator stopAnimation:self];
	[_pleaseWaitTextField setStringValue:@"Please email the TextEdit document to feedback@glimmerblocker.org"];
	retainMe = NULL;
}

- (void)runProxyDebug
{
	NSArray* args = @[@"-c", @"/Library/PreferencePanes/GlimmerBlocker.prefPane/Contents/GlimmerBlockerProxy.app/Contents/MacOS/GlimmerBlocker --make-report --textedit 2>&1 | open -f"];
	[GBDebug executeCommandlineTool:@"/bin/bash" withArgs:args];
	retainMe = self;
	[self performSelectorOnMainThread:@selector(runProxyDebugDone)
						   withObject:NULL
						waitUntilDone:NO];
}

//============================================================================

- (IBAction)fixBtnAction:(id)sender
{
	if (_invalidCodeSignature) {
		[_fixBtn setEnabled:NO];
		[_pleaseWaitTextField setStringValue:@"Creates TextEdit document with all the details."];
		[_progressIndicator startAnimation:self];
		[self performSelector:@selector(runProxyDebug)
				   withObject:NULL
				   afterDelay:0.5];
		return;
	}
	NSLog(@"GB install action.");
	if (![self createTmpDir]) {
		[_fixBtn setEnabled:NO];
		[_pleaseWaitTextField setStringValue:@"Can't create temporary files at all!!!  No fix possible."];
		return;
	}
	[self prepareAuth];
	if (!_authRef) {
		NSLog(@"Could not authenticate.");
		return;
	}
	[_fixBtn setEnabled:NO];
	[_pleaseWaitTextField setStringValue:@"Please wait!"];
	_progressText = @"Please wait…";
	[self updateProgressText];
	[_progressIndicator startAnimation:self];
	//
	[self performSelector:@selector(runInstaller)
			   withObject:NULL
			   afterDelay:0.5];
}

@end
