/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.admin;

import javax.net.ssl.HttpsURLConnection;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.proxyRequest.ProxyRequestHandler;
import org.glimmerblocker.proxy.util.DocumentPrinter;
import org.glimmerblocker.proxy.util.XmlParserHelper;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.exception.UserErrorException;

public abstract class AbstractAdminXmlHandler extends AbstractAdminHandler
{
	@NotNull public final Document doc;
	@NotNull public final Element rootE;

	protected AbstractAdminXmlHandler(@NotNull ProxyRequestHandler req) throws Exception
	{
		super(req);
		this.doc = XmlParserHelper.createDocument();
		this.rootE = doc.createElement("root");
		doc.appendChild(rootE);
	}

	protected final void main() throws Exception
	{
		try {
			processAdminXml();
		}
		catch (Throwable ex) {
			String msg = ExceptionUtils.getExMsg(ex);
			if (!(ex instanceof UserErrorException)) {
				ln("Got unexpected admin exception: " + msg);
				ln(ExceptionUtils.getStackTraceAsString(ex));
			}
			rootE.setAttribute("error", msg);
			Element stacktraceE = doc.createElement("error-stacktrace");
			stacktraceE.appendChild(doc.createCDATASection(ExceptionUtils.getStackTraceAsString(ex)));
		}
		Headers h = new Headers(HttpsURLConnection.HTTP_OK, "OK", "text/xml; charset=UTF-8", null, null);
		DocumentPrinter dp = new DocumentPrinter(doc);
		ResultData rd = new ResultData(h, dp.printToString());
		req.sendResultData(rd);
	}

	protected abstract void processAdminXml() throws Exception;

	// ================================================================================


	public static void setOptAttr(@NotNull Element elem, @NotNull String name, @Nullable String value)
	{
		if (value == null || value.length() == 0)
			return;
		setAttr(elem, name, value);
	}

	public static void setAttr(@NotNull Element elem, @NotNull String name, @Nullable String value)
	{
		if (name.length() == 0)
			throw new NullPointerException("name is null");
		elem.setAttribute(name, value);
	}

	public static void setOptAttr(@NotNull Element elem, @NotNull String name, boolean value)
	{
		if (value)
			setAttr(elem, name, "1");
	}

	public static void setAttr(@NotNull Element elem, @NotNull String name, int value)
	{
		setAttr(elem, name, Integer.toString(value));
	}

	public static void setAttr(@NotNull Element elem, @NotNull String name, long value)
	{
		setAttr(elem, name, Long.toString(value));
	}

	public static void setAttr(@NotNull Element elem, @NotNull String name, double value)
	{
		setAttr(elem, name, Double.toString(value));
	}

	public static void removeAttr(@NotNull Element elem, @NotNull String name)
	{
		elem.removeAttribute(name);
	}

	public static void removeAllChildren(@NotNull Node parentNode)
	{
		while (true) {
			Node child = parentNode.getFirstChild();
			if (child == null)
				return;
			parentNode.removeChild(child);
		}
	}

	public static void removeNode(@Nullable Node n)
	{
		if (n == null)
			return;
		Node parentNode = n.getParentNode();
		if (parentNode != null)
			parentNode.removeChild(n);
	}

	public static void removeNodes(@Nullable Node[] list)
	{
		if (list == null)
			return;
		for (Node n : list)
			removeNode(n);
	}

	public Element addChildWithTextNode(@NotNull Element parentE, @NotNull String tagName, @Nullable String value)
	{
		if (tagName.length() == 0)
			throw new RuntimeException("Empty tagname not valid");
		Element e = doc.createElement(tagName);
		if (value != null && value.length() > 0)
			e.appendChild(doc.createTextNode(value));
		parentE.appendChild(e);
		return e;
	}

	public void addCommentNode(@NotNull Element elem, @Nullable String value)
	{
		if (value == null || value.length() == 0)
			return;
		elem.appendChild(doc.createComment(value));
	}

	public void addCDataNodes(@NotNull Element elem, @Nullable String value)
	{
		if (value == null || value.length() == 0)
			return;
		int startIndex = 0;
		int len = value.length();
		while (startIndex < len) {
			int idx = value.indexOf("]]>");
			if (idx < 0) {
				elem.appendChild(doc.createCDATASection(value.substring(startIndex)));
				return;
			}
			String s = value.substring(startIndex, idx);
			elem.appendChild(doc.createCDATASection(s));
			elem.appendChild(doc.createTextNode("]]>"));
			startIndex = idx + 3;
		}
	}
}
