/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.admin;

import java.io.StringWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Map;
import javax.net.ssl.HttpsURLConnection;
import org.glimmerblocker.proxy.admin.filter.AdminFilterCheckUpdates;
import org.glimmerblocker.proxy.admin.filter.AdminFilterSubAdd;
import org.glimmerblocker.proxy.admin.filter.AdminFilterSubGet;
import org.glimmerblocker.proxy.admin.filter.AdminFilterSubRemove;
import org.glimmerblocker.proxy.admin.filter.AdminFilterSubUpdate;
import org.glimmerblocker.proxy.admin.filter.AdminFilterValidate;
import org.glimmerblocker.proxy.admin.misc.AdminDumpHttpCache;
import org.glimmerblocker.proxy.admin.misc.AdminHelloWorld;
import org.glimmerblocker.proxy.admin.misc.AdminHistoryList;
import org.glimmerblocker.proxy.admin.misc.AdminLogoLarge;
import org.glimmerblocker.proxy.admin.misc.AdminReloadPanelSettings;
import org.glimmerblocker.proxy.admin.misc.AdminSiteProxyValidate;
import org.glimmerblocker.proxy.admin.misc.AdminStopNotification;
import org.glimmerblocker.proxy.admin.misc.AdminSuspectList;
import org.glimmerblocker.proxy.admin.misc.AdminTestRunning;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.proxyRequest.ProxyRequestHandler;
import org.jabhts.runtime.html.HtmlPrintWriter;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public class AdminDispatcher
{
	public AdminDispatcher(@NotNull GlimmerGlobals globals) throws Exception
	{
		this.globals = globals;
		publicAdmin = new HashMap<>();
		infoAdmin = new HashMap<>();
		secureAdmin = new HashMap<>();
		//
		add(secureAdmin, "/reload-panel-settings", AdminReloadPanelSettings.class);
		add(secureAdmin, "/stop-notification", AdminStopNotification.class);
		add(infoAdmin, "/test-running", AdminTestRunning.class);
		add(secureAdmin, "/http-cache/dump", AdminDumpHttpCache.class);
		//
		add(infoAdmin, "/filters/validate-edit", AdminFilterValidate.class);
		//
		add(infoAdmin, "/filters/subscription/get", AdminFilterSubGet.class);
		add(secureAdmin, "/filters/subscription/check-updates", AdminFilterCheckUpdates.class);
		//
		add(secureAdmin, "/filters/subscription/add", AdminFilterSubAdd.class);
		add(secureAdmin, "/filters/subscription/remove", AdminFilterSubRemove.class);
		add(secureAdmin, "/filters/subscription/update", AdminFilterSubUpdate.class);
		//
		add(infoAdmin, "/suspect/list", AdminSuspectList.class);
		//
		add(infoAdmin, "/history/list", AdminHistoryList.class);
		//
		add(infoAdmin, "/site-proxy/validate", AdminSiteProxyValidate.class);
		//
		add(publicAdmin, "/logo/large.png", AdminLogoLarge.class);
		add(publicAdmin, "/hello", AdminHelloWorld.class);
	}

	private final GlimmerGlobals globals;
	private final Map<String, Constructor<? extends AbstractAdminHandler>>
			publicAdmin, infoAdmin, secureAdmin;

	public static final String pathPrefix = "/=gb=/";
	public static final String publicAdminKey = "-public-";

	public String getPublicUrlBase()
	{
		return "http://127.0.0.1:" + globals.httpServer.getListenPort() + pathPrefix + publicAdminKey;
	}

	public void handle(@NotNull ProxyRequestHandler req, @NotNull String path) throws Exception
	{
		if (path.startsWith(pathPrefix)) {
			String p = path.substring(pathPrefix.length());
			if (handle(req, p, publicAdminKey, publicAdmin))
				return;
			if (handle(req, p, req.globals.infoAdminKey, infoAdmin))
				return;
			if (handle(req, p, req.globals.secureAdminKey, secureAdmin))
				return;
			req.ln("Invalid admin key");
			req.sendResultData(ResultData.FORBIDDEN);
			return;
		}
		if (req.requestPath.equals("/favicon.ico")) {
			req.sendResultData(ResultData.checkerboardPatternGif);
			return;
		}
		req.sendResultData(ResultData.NOT_FOUND);
	}

	public boolean handle(@NotNull ProxyRequestHandler req,
	                      @NotNull String subPath,
	                      @Nullable String mapPrefix,
	                      @NotNull Map<String, Constructor<? extends AbstractAdminHandler>> map) throws Exception
	{

		if (mapPrefix == null
				|| subPath.length() + 1 < mapPrefix.length()
				|| !subPath.startsWith(mapPrefix)
				|| subPath.charAt(mapPrefix.length()) != '/') {
			return false;
		}
		subPath = subPath.substring(mapPrefix.length());
		Constructor<? extends AbstractAdminHandler> cnstr = map.get(subPath);
		if (cnstr != null) {
			AbstractAdminHandler handler = cnstr.newInstance(req);
			handler.main();
			return true;
		}
		req.ln("Unknown admin subPath: " + subPath);
		StringWriter sw = new StringWriter();
		HtmlPrintWriter pw = new HtmlPrintWriter(sw);
		pw.println("<div style='font-weight: bold; color: red'>Unknown admin URL: ");
		pw.htmlEncode(req.request.path);
		pw.println("</div>");
		pw.flush();
		Headers h = new Headers(HttpsURLConnection.HTTP_NOT_FOUND, "Unknown URL",
				"text/html; charset=UTF-8", null, null);
		ResultData rd = new ResultData(h, sw.toString().getBytes("UTF-8"));
		req.sendResultData(rd);
		return true;
	}

	public void add(@NotNull Map<String, Constructor<? extends AbstractAdminHandler>> map,
	                @NotNull String path,
	                @NotNull Class<? extends AbstractAdminHandler> cls) throws Exception
	{
		if (!Modifier.isPublic(cls.getModifiers()))
			throw new Exception("Admin class is not public: " + cls.getName());
		Constructor<? extends AbstractAdminHandler> cnstr;
		try {
			cnstr = cls.getConstructor(ProxyRequestHandler.class);
		}
		catch (Exception ex) {
			throw new Exception("Admin handler does not have required constructor: " + cls.getName(), ex);
		}
		if (!Modifier.isPublic(cnstr.getModifiers()))
			throw new Exception("Admin handler constructor is not public: " + cls.getName());
		map.put(path, cnstr);
	}
}
