/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import java.util.List;
import java.util.Map;
import org.glimmerblocker.proxy.TestServer;
import org.glimmerblocker.proxy.admin.AdminDispatcher;
import org.glimmerblocker.proxy.globals.history.HistoryCache;
import org.glimmerblocker.proxy.globals.log.DatePrefixedLogger;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.RollingLogfile;
import org.glimmerblocker.proxy.globals.network.NetworkManager;
import org.glimmerblocker.proxy.globals.network.dns.DnsResolver;
import org.glimmerblocker.proxy.globals.subscriptions.FilterSubscriptions;
import org.glimmerblocker.proxy.globals.subscriptions.SubscriptionsAutoUpdater;
import org.glimmerblocker.proxy.globals.suspect.SuspectsManager;
import org.glimmerblocker.proxy.http.client.BodyCollector;
import org.glimmerblocker.proxy.http.client.cache.HttpClientCache;
import org.glimmerblocker.proxy.http.client.connection.HttpResponseInfo;
import org.glimmerblocker.proxy.http.client.curl.CurlBase;
import org.glimmerblocker.proxy.http.client.stream.TimeoutableInputStream;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.HeadersBuilder;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.http.server.HttpServer;
import org.glimmerblocker.proxy.rhino.runtime.RhinoRuntimeCache;
import org.glimmerblocker.proxy.util.ResourceUtils;
import org.glimmerblocker.proxy.util.XmlParserHelper;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.io.SimpleUnixExec;
import dk.vitality.lib.foundation.misc.StringUtils;
import dk.vitality.lib.foundation.ref.FatRef;
import dk.vitality.lib.foundation.xml.xpath.XmlArgs;

/**
 * Intended for preferences and more global customization without requiring you to subclass
 * more internal classes.
 */
@SuppressWarnings({"ConstantConditions"})
public class GlimmerGlobals implements Logger
{
	/**
	 * General log debug info.
	 */
	public boolean debugFlag = false;

	/**
	 * Log verbose info about streaming to log
	 */
	public boolean debugVerbose = false;

	/**
	 * Log info about filters
	 */
	public boolean debugFilters = false;

	/**
	 * Log debug info for http server requests
	 */
	public boolean debugHttpServer;

	/**
	 * Log debug info for http client requests/transactions.
	 */
	public final FatRef<Boolean> debugHttpClient = new FatRef<>(false);

	/**
	 * Log debug info for network configuration.
	 */
	public final FatRef<Boolean> debugNetworkConfig = new FatRef<>(false);

	/**
	 * Send checkerboard pattern (instead of transparent gif) when blocking requests.
	 */
	public boolean useCheckerboardPatternWhenBlocking;

	/**
	 * Make html box with filter transform (javascript) errors.
	 */
	public boolean showFilterTransformErrorHtmlBox = false;

	/**
	 * The log handler.
	 */
	@NotNull protected final Logger logger;

	/**
	 * Are the filters/rules enabled, or are the requests always sent unmodified.
	 */
	public boolean filtersEnabled = true;

	/**
	 * The active rules.
	 */
	@NotNull public volatile RuleCollection ruleCollection;

	/**
	 * Rhino Javascript runtime cache.
	 */
	public final RhinoRuntimeCache rhinoRuntimeCache;

	@NotNull public final FilterSubscriptions filterSubscriptions;

	/**
	 * The secret paths used for admin requests.
	 * This is unique to each installation of glimmerblocker.
	 */
	@NotNull public String infoAdminKey;
	@NotNull public String secureAdminKey;
	@NotNull public final AdminDispatcher adminDispatcher;

	public static final File prefsPanelBundleF = new File("/Library/PreferencePanes/GlimmerBlocker.prefPane");
	public static final File prefsPanelResourcesF = new File(prefsPanelBundleF, "Contents/Resources");

	public static final File proxyBundleF = new File(prefsPanelBundleF, "Contents/GlimmerBlockerProxy.app");
	public static final File proxyResourcesF = new File(proxyBundleF, "Contents/Resources");

	public static final File networkConfigListenerBundleF =
			new File(prefsPanelBundleF, "Contents/NetworkConfigurationListener.app");
	public static final File networkConfigListenerBinary =
			new File(networkConfigListenerBundleF, "Contents/MacOS/NetworkConfigurationListener");

	public volatile XmlArgs panelSettingsXA;

	public final JavaSettings javaSettings;

	public final DnsResolver dnsResolver;

	public final HttpClientCache httpClientCache;

	public final NetworkManager networkManager;

	public final HttpServer httpServer;

	public final String gbVersionString;

	public final String serverUserAgent;

	public final SuspectsManager suspects;

	public final HistoryCache historyCache;

	public final SubscriptionsAutoUpdater subscriptionsAutoUpdater;

	/**
	 * Initialize the manager. Must be called after instance creation.
	 */
	public GlimmerGlobals(boolean logToStderr) throws Exception
	{
		if (logToStderr)
			logger = new DatePrefixedLogger();
		else
			logger = new RollingLogfile(this);
		String s = getOptTextResource("/gb-version.txt");
		if (s == null)
			s = "1.4.0"; // cmd-line
		gbVersionString = s;
		serverUserAgent = makeUserAgent();
		printBanner();
		fixLogfilePermissions();
		privateLoadSettingsXml(); // so init of e.g. networkConfig has updated debugFlag.
		//
		this.rhinoRuntimeCache = new RhinoRuntimeCache(this);
		this.dnsResolver = new DnsResolver(this, debugNetworkConfig);
		this.httpClientCache = new HttpClientCache(dnsResolver, this, debugHttpClient);
		this.networkManager = new NetworkManager(this);
		this.httpServer = new HttpServer(this);
		this.ruleCollection = new RuleCollection(this);
		this.javaSettings = new JavaSettings(this);
		this.secureAdminKey = javaSettings.getSecureAdminKey();
		this.filterSubscriptions = new FilterSubscriptions(this);
		this.suspects = new SuspectsManager(this);
		this.historyCache = new HistoryCache(this);
		this.adminDispatcher = new AdminDispatcher(this);
		sendLoadedSettingsNotifications();
		this.subscriptionsAutoUpdater = new SubscriptionsAutoUpdater(this);
		writeJavaVersionToFile();
		if (1 == 2)
			if (debugHttpClient.get())
				new TestServer(this);
	}

	private void printBanner() throws Exception
	{
		SimpleUnixExec sue = new SimpleUnixExec("/bin/sh", "-c", "echo $PPID");
		sue.setLogStdoutToSystemErr(false);
		sue.setWantsStdoutAsString(true);
		sue.run();
		String pid = (sue.getResultCode() == 0) ? sue.getStdoutAsString().trim() : "???";
		//
		sue = new SimpleUnixExec("/usr/bin/ulimit", "-n");
		sue.setLogStdoutToSystemErr(false);
		sue.setWantsStdoutAsString(true);
		sue.run();
		String ulimitn = (sue.getResultCode() == 0) ? sue.getStdoutAsString().trim() : "???";
		//
		ln("GlimmerBlocker proxy server starting.  Pid: " + pid + ".  Max fds: " + ulimitn + "." +
				"  " + System.getProperty("sun.arch.data.model") + "-bit jvm.");
		System.err.flush();
		String s = getOptTextResource("/gb-fingerprint.txt");
		if (s != null)
			ln("   *.class fingerprint: " + s);
		else
			ln("  Missing *.class fingerprint resource.");
	}

	/**
	 * See {@link Logger#ln}
	 */
	public void ln(@Nullable String s)
	{
		if (s == null)
			return;
		// remove sensitive information:
		if (infoAdminKey != null)
			s = StringUtils.replaceAll(s, infoAdminKey, "«-info-admin-»");
		if (secureAdminKey != null)
			s = StringUtils.replaceAll(s, secureAdminKey, "«-secure-admin-»");
		logger.ln(s);
	}

	public synchronized void reloadPanelSettings() throws Exception
	{
		privateLoadSettingsXml();
		sendLoadedSettingsNotifications();
	}

	private void privateLoadSettingsXml() throws Exception
	{
		boolean initialLoad = (infoAdminKey == null);
		boolean hadDebug = debugFlag;
		if (initialLoad) {
			ln("Loads settings from disk.  " + serverUserAgent);
		} else if (debugFlag) {
			ln("-------------------------------------");
			ln("Reloads settings from disk.  " + serverUserAgent);
		}
		for (int i = 0; i < 5; i++) {
			System.gc();
			Thread.sleep(10);
		}
		File f = new File("/Library/GlimmerBlocker/PanelSettings.xml");
		if (!f.exists()) {
			ln("Missing panel settings file: " + f.getPath());
			this.panelSettingsXA = new XmlArgs(XmlParserHelper.createDocument());
		} else {
			panelSettingsXA = new XmlArgs(XmlParserHelper.parse(f));
			debugFlag = panelSettingsXA.getOptBoolean("debug/@enabled");
		}
		if (debugFlag && !initialLoad && !hadDebug) {
			ln("-------------------------------------"); // didn't make this message above.
			ln("Reloads settings from disk.");
		}
		debugVerbose = debugFlag && panelSettingsXA.getOptBoolean("debug/@verbose");
		debugHttpClient.set(debugFlag && panelSettingsXA.getOptBoolean("debug/@http-client"));
		debugHttpServer = debugFlag && panelSettingsXA.getOptBoolean("debug/@http-server");
		debugNetworkConfig.set(debugFlag && panelSettingsXA.getOptBoolean("debug/@network-config"));
		debugFilters = debugFlag && panelSettingsXA.getOptBoolean("debug/@filters");
		showFilterTransformErrorHtmlBox = panelSettingsXA.getOptBoolean("debug/@transform-html-box");
		useCheckerboardPatternWhenBlocking = panelSettingsXA.getOptBoolean("debug/@checkerboard-pattern");
		if (debugFlag || hadDebug) {
			ln("  debug = " + debugFlag + ", verbose = " + debugVerbose +
					", filters = " + debugFilters +
					", filter-html-box = " + showFilterTransformErrorHtmlBox +
					"");
			ln("  http-client = " + debugHttpClient.get() +
					", http-server = " + debugHttpServer +
					", network-cfg = " + debugNetworkConfig.get() +
					"");
			logMemoryInfo();
		}
	}

	private void sendLoadedSettingsNotifications() throws Exception
	{
		infoAdminKey = panelSettingsXA.getOptString("proxy-server/@info-admin-key",
				"impossible/default/value");
		suspects.loadSettings();
		historyCache.loadSettings();
		RuleCollection rc = new RuleCollection(logger);
		new FilterLoader(this, rc);
		ruleCollection = rc;
		rhinoRuntimeCache.clear();
		networkManager.notifyPrefsChanged();
		httpServer.notifyPrefsChanged();
	}

	@NotNull
	public ResultData curl(@NotNull String method,
	                       @NotNull URL url,
	                       @NotNull Map<String, List<String>> requestHeadersMap,
	                       @NotNull Logger parentLogger) throws Exception
	{
		HeadersBuilder hb = HeadersBuilder.makeEmpty();
		hb.copyAll(requestHeadersMap);
		hb.setAllowGzip(true);
		return curl(method, url, hb.toHeaders(), parentLogger);
	}

	@NotNull
	public ResultData curl(@NotNull String method,
	                       @NotNull URL url,
	                       @NotNull Headers headers,
	                       @NotNull Logger parentLogger) throws Exception
	{
		final ResultData[] result = new ResultData[1];
		CurlBase exec = new CurlBase(this, parentLogger, networkManager.getConfig(), null, method, url, headers)
		{
			protected void readBody(@NotNull HttpResponseInfo hri,
			                        @Nullable TimeoutableInputStream bodyStream) throws Exception
			{
				BodyCollector bc = new BodyCollector(hri);
				bc.readBody(bodyStream);
				result[0] = bc.getResultData();
			}

			protected void sendRequestRuleData(@NotNull ResultData data) throws Exception
			{
				result[0] = data;
			}
		};
		exec.execute();
		if (result[0] == null)
			throw new IOException("Did not get any response body");
		return result[0];
	}

	@SuppressWarnings({"UnusedCatchParameter"})
	private String getOptTextResource(String path)
	{
		try {
			return new String(ResourceUtils.getResource(path), "UTF-8");
		}
		catch (Exception ex) {
			return null;
		}
	}

	private String makeUserAgent()
	{
		if (1 == 2) {
			for (Map.Entry<Object, Object> e : System.getProperties().entrySet())
				ln("SYSPROP {" + e.getKey() + "} = {" + e.getValue() + "}");
		}
		StringBuilder sb = new StringBuilder(20);
		addVersionItem(sb, "GlimmerBlocker/", gbVersionString);
		addVersionItem(sb, "Java/", System.getProperty("java.version"));
		addVersionItem(sb, "OSX/", System.getProperty("os.version"));
		return sb.toString();
	}

	private void addVersionItem(StringBuilder sb, String key, String value)
	{
		if (value == null || value.length() == 0)
			return;
		if (sb.length() > 0)
			sb.append(" ");
		sb.append(key).append(value);
	}

	private void fixLogfilePermissions()
	{
		File f = new File("/Library/Logs/GlimmerBlocker/GlimmerBlocker.log");
		ResourceUtils.setPermissions(f, "ugo+r");
	}

	public void logMemoryInfo()
	{
		Runtime r = Runtime.getRuntime();
		long free = r.freeMemory();
		long total = r.totalMemory();
		ln("freeMemory = " + kb(free) + " KB" +
				", totalMemory = " + kb(total) + " KB" +
				", maxMemory = " + kb(r.maxMemory()) + " KB" +
				", used = " + kb(total - free) + " KB" +
				"");
	}

	private void writeJavaVersionToFile() throws Exception
	{
		File f = new File("/Library/GlimmerBlocker/ProxyJavaInfo.txt");
		if (f.exists() && !f.delete())
			ln("Couldn't delete " + f.getPath()); // make sure the timestamp is updated.
		FileOutputStream fos = new FileOutputStream(f);
		try {
			String s = System.getProperty("java.version");
			if (s != null && s.length() > 0)
				fos.write(s.getBytes("UTF-8"));
		}
		finally {
			fos.close();
		}
		ResourceUtils.setPermissions(f, "0644");
	}

	private static long kb(long x)
	{
		return (x + 512) / 1024;
	}
}
