/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.network;

import org.glimmerblocker.proxy.globals.network.dns.DnsResolverConfig;
import org.glimmerblocker.proxy.rhino.callback.CallbackRhinoBinder;
import org.glimmerblocker.proxy.rhino.callback.CallbackSubObject;
import org.glimmerblocker.proxy.rhino.callback.GlimmerGlobalObjectBase;
import org.glimmerblocker.proxy.rhino.callback.RhinoCallback;
import org.jetbrains.annotations.NotNull;
import org.mozilla.javascript.Context;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.ref.FatRef;

public class NetworkConfigSubObject extends CallbackSubObject
{
	FatRef<NetworkConfig> ref;

	private static final CallbackRhinoBinder subBinder = new CallbackRhinoBinder(NetworkConfigSubObject.class);

	public NetworkConfigSubObject(@NotNull GlimmerGlobalObjectBase callback)
	{
		super(callback, subBinder);
	}

	public void prepareNetworkConfigSubObject(@NotNull Context ctx,
	                                          @NotNull FatRef<NetworkConfig> ref) throws Exception
	{
		super.prepareObjectBase(ctx);
		if (this.ref != null)
			throw Context.reportRuntimeError("Called inside existing scope, has already NetworkConfig ref.");
		this.ref = ref;
	}

	@Override
	public void notifyLeftContext()
	{
		super.notifyLeftContext();
		ref = null;
	}

	//===========================================================================

	@RhinoCallback
	public String jsGet_dnsServers()
	{
		return ref.get().dnsConfig.javascriptText;
	}

	@RhinoCallback
	public void jsSet_dnsServers(String s) throws Exception
	{
		if (ref == null)
			ln("jsSet_dnsServers called outside context!\n" + ExceptionUtils.getStackTraceAsString());
		NetworkConfig cfg = ref.get();
		cfg = cfg.withDnsConfig(DnsResolverConfig.decode(s));
		ref.set(cfg);
		if (globals.debugFilters)
			ln("Script sets dns servers to: " + cfg.dnsConfig.description);
	}

	//===========================================================================

	@RhinoCallback
	public String jsGet_httpProxy()
	{
		NetworkConfig cfg = ref.get();
		if (cfg.httpProxy != null)
			return cfg.httpProxy.hostnamePort;
		else
			return "";
	}

	@RhinoCallback
	public void jsSet_httpProxy(String s) throws Exception
	{
		s = (s == null) ? "" : s.trim();
		if (globals.debugFilters)
			ln("Script sets http-proxy to '" + s + "'");
		HostnamePort hp = HostnamePort.parse(s);
		if (hp != null) {
			if (hp.port < 80 || hp.port == 443 || hp.port > 65535)
				throw Context.reportRuntimeError("Invalid http proxy port number = " + hp.port);
		}
		NetworkConfig cfg = ref.get();
		cfg = cfg.withHttpProxy(hp);
		ref.set(cfg);
	}

	//===========================================================================

	@RhinoCallback
	public String jsGet_socksProxy()
	{
		NetworkConfig cfg = ref.get();
		if (cfg.socksProxy != null)
			return cfg.socksProxy.hostnamePort;
		else
			return "";
	}

	@RhinoCallback
	public void jsSet_socksProxy(String s) throws Exception
	{
		s = (s == null) ? "" : s.trim();
		if (globals.debugFilters)
			ln("Script sets socks-proxy to '" + s + "'");
		HostnamePort hp = HostnamePort.parse(s);
		if (hp != null) {
			if (hp.port <= 80 || hp.port == 443 || hp.port > 65535)
				throw new Exception("Invalid socks proxy port number = " + hp.port);
		}
		NetworkConfig cfg = ref.get();
		cfg = cfg.withSocksProxy(hp);
		ref.set(cfg);
	}

	//===========================================================================

	@RhinoCallback
	public String jsGet_location()
	{
		return ref.get().locationName;
	}

	//===========================================================================

	@RhinoCallback
	public Object jsGet_interfaces()
	{
		NetworkConfig cfg = ref.get();
		if (!cfg.interfacesJSON.equals(interfacesJSON)) {
			try {
				String js = "" +
						"(function() {\n" +
						"return (" + cfg.interfacesJSON.trim() + ");\n" +
						"})();";
				//if (globals.debugVerbose)
				//	ln("network JS:\n" + js);
				interfacesObject = ctx.evaluateString(scope, js, "interfaces", 1, null);
				interfacesJSON = cfg.interfacesJSON;
			}
			catch (Exception ex) {
				System.err.println("Failed evaluating network json: " + ExceptionUtils.getExMsg(ex));
				if (globals.debugVerbose)
					ln(ExceptionUtils.getStackTraceAsString(ex));
			}
		}
		return interfacesObject;
	}

	private Object interfacesObject;
	private String interfacesJSON;

	// http://web.archive.org/web/20060424005037/http://wp.netscape.com/eng/mozilla/2.0/relnotes/demo/proxy-live.html#isInNet

	//===========================================================================
}