/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.globals.suspect;

import java.net.URL;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.http.data.Headers;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public class SuspectsManager implements Logger
{
	@NotNull public final GlimmerGlobals globals;
	@NotNull public final Logger logger;

	private boolean enabled;

	public SuspectsManager(@NotNull GlimmerGlobals globals)
	{
		this.globals = globals;
		this.logger = new PrefixedLogger(globals, "suspects");
	}

	public void loadSettings() throws Exception
	{
		synchronized (cache) {
			boolean wantsEnabled = globals.panelSettingsXA.getOptBoolean("suspects/@enabled");
			if (enabled == wantsEnabled)
				return;
			ln("enabled: " + enabled + " -> " + wantsEnabled);
			enabled = wantsEnabled;
			if (!enabled)
				cache.clear();
		}
	}

	@SuppressWarnings({"UnusedDeclaration"})
	public void process(@NotNull Logger parentLogger,
	                    @NotNull String requestMethod,
	                    @NotNull URL requestURL,
	                    @NotNull Headers requestHeaders,
	                    @NotNull Headers responseHeaders) throws Exception
	{
		if (!enabled)
			return;
		if (!requestMethod.equals("GET"))
			return;
		SuspectCandidate candidate = new SuspectCandidate(parentLogger, globals, this,
				requestMethod, requestURL, requestHeaders,
				responseHeaders);
		Suspect suspect = candidate.suspect;
		if (suspect == null) {
			if (globals.debugVerbose)
				ln("ignored: " + requestURL);
			return;
		}
		synchronized (cache) {
			if (globals.debugVerbose) {
				String prefix = cache.containsKey(suspect) ? "Known " : "New ";
				ln(prefix + suspect.description +
						": " + suspect.host + suspect.path +
						"\n      from " + suspect.referer);
			}
			cache.put(candidate.suspect, candidate.suspect);
		}
	}

	@NotNull
	public List<Suspect> getSuspects()
	{
		synchronized (cache) {
			return new ArrayList<>(cache.values());
		}
	}

	public void removeSuspect(@NotNull Suspect suspect)
	{
		synchronized (cache) {
			cache.remove(suspect);
		}
	}

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}


	private final LruCache cache = new LruCache();
	private static final int cacheSize = 150;

	private static class LruCache extends LinkedHashMap<Suspect, Suspect>
	{
		private LruCache()
		{
			super(cacheSize);
		}

		@Override
		protected boolean removeEldestEntry(Map.Entry<Suspect, Suspect> eldest)
		{
			return size() >= cacheSize;
		}
	}
}
