/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.connection;

import java.io.BufferedInputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.network.dns.DnsResolver;
import org.glimmerblocker.proxy.http.client.cache.HttpClientCacheKey;
import org.glimmerblocker.proxy.http.client.stream.RequestData;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.ref.FatRef;

public class HttpClientConnection implements Logger
{
	@NotNull public final DnsResolver dnsResolver;
	@NotNull public final HttpClientCacheKey connKey;
	@NotNull public final Socket skt;
	@NotNull public final BufferedInputStream is;
	@NotNull public final OutputStream os;

	@NotNull private Logger logger;
	@NotNull protected FatRef<Boolean> debugHttpClient;
	@Nullable protected HttpTransaction transaction;
	private long maxAllowedReuseTime = Long.MAX_VALUE;
	private boolean reuseAllowed;
	protected int transactionCount;
	protected boolean debugHttpTransaction;
	private boolean hasOpenSocket;
	final FatRef<String> connectionStateDescRef = new FatRef<>("initalizing");

	public HttpClientConnection(@NotNull DnsResolver dnsResolver,
	                            @NotNull HttpClientCacheKey connKey,
	                            boolean useCONNECT,
	                            @NotNull Logger logger,
	                            @NotNull FatRef<Integer> ipAddressIndex,
	                            @NotNull FatRef<Boolean> debugHttpClient) throws Exception
	{
		this.debugHttpClient = debugHttpClient;
		this.dnsResolver = dnsResolver;
		this.connKey = connKey;
		this.logger = logger;
		InetAddress[] hostAddresses = resolveHostname();
		HttpSocketConnector hsc = connect(hostAddresses, ipAddressIndex, useCONNECT);
		this.hasOpenSocket = true;
		this.skt = hsc.skt;
		this.is = hsc.is;
		this.os = hsc.os;
		this.reuseAllowed = true;
	}

	private HttpSocketConnector connect(@NotNull InetAddress[] hostAddresses,
	                                    @NotNull FatRef<Integer> ipAddressIndex,
	                                    boolean useCONNECT) throws Exception
	{
		final boolean debugFlag = debugHttpClient.get();
		int idx = Math.max(0, ipAddressIndex.get());
		for (int i = 0; true; i++, idx++) {
			if (idx >= hostAddresses.length)
				idx = 0;
			InetAddress iaddr = hostAddresses[idx];
			String debugAddressDesc = connKey.tcpHost.hostname + ":" + connKey.tcpHost.port +
					" using " + iaddr.getHostAddress() +
					" [" + (idx + 1) + "/" + hostAddresses.length + "]";
			try {
				boolean longTimeout = (i >= hostAddresses.length);
				HttpSocketConnector hsc = new HttpSocketConnector(this,
						iaddr, connKey, useCONNECT, longTimeout, debugAddressDesc, connectionStateDescRef, debugFlag);
				ipAddressIndex.set(idx);
				return hsc;
			}
			catch (Exception ex) {
				if (i++ == 2 * hostAddresses.length)
					throw ex;
				if (debugFlag) {
					ln("Failed to connect to " + debugAddressDesc +
							" due to " + ExceptionUtils.getExMsg(ex));
				}
			}
		}
	}

	@Override
	protected void finalize() throws Throwable
	{
		if (hasOpenSocket)
			ln("HttpClientConnection finalized with non-closed socket!");
		super.finalize();
	}

	private InetAddress[] resolveHostname() throws Exception
	{
		long st0 = System.currentTimeMillis();
		final Boolean debugFlag = debugHttpClient.get();
		this.connectionStateDescRef.set("resolving " + connKey.tcpHost.hostname);
		InetAddress[] addrArray = null;
		try {
			addrArray = dnsResolver.getAllByName(connKey.dnsConfig, connKey.tcpHost.hostname);
		}
		catch (Exception ex) {
			if (debugFlag)
				ln("DNS lookup failed (" + connKey.tcpHost.hostname + "): " + ExceptionUtils.getExMsg(ex));
			throw ex;
		}
		finally {
			if (debugFlag) {
				long dur = System.currentTimeMillis() - st0;
				if (dur > 2) {
					// not cached.
					StringBuilder sb = new StringBuilder(50);
					sb.append("DNS lookup time = ").append(dur).append(" msecs");
					sb.append(" for ").append(connKey.tcpHost.hostname);
					if (addrArray != null) {
						for (int i = 0; i < addrArray.length; i++)
							sb.append(", [").append(i).append("] = ").append(addrArray[i].getHostAddress());
					}
					ln(sb.toString());
				}
			}
		}
		if (addrArray == null || addrArray.length == 0)
			throw new UnknownHostException("Invalid address array returned from getAllByName.");
		return addrArray;
	}

	public HttpTransaction makeTransaction(@NotNull RequestData requestData) throws KeepAliveException
	{
		if (transaction != null)
			throw new KeepAliveException("HttpConnection is not reusable: has already transaction");
		if (maxAllowedReuseTime < System.currentTimeMillis())
			throw new KeepAliveException("HttpConnection is not reusable: too long idle time");
		transactionCount++;
		debugHttpTransaction = debugHttpClient.get();
		if (debugHttpTransaction)
			ln("Transaction-" + transactionCount + " started");
		transaction = new HttpTransaction(this, requestData, this, debugHttpTransaction);
		return transaction;
	}

	public boolean isReusable()
	{
		// skt.isConnected() doesn't seem to return false if the remove site has closed
		// the connection, so it seems like an "is bound".
		long now = System.currentTimeMillis();
		boolean tooOld = (maxAllowedReuseTime < now);
		boolean result = reuseAllowed && (transaction == null) && !tooOld;
		//noinspection ConstantConditions
		if (false && debugHttpClient.get()) {
			ln("isReusable: reuseAllowed = " + reuseAllowed +
					(transaction != null ? ", has transaction" : ", no transaction") +
					", " +
					(maxAllowedReuseTime == 0
							? "zero max age"
							: Math.abs((maxAllowedReuseTime - now) / 1000.0) +
							(tooOld ? " too old" : " secs still usable")) +
					" => " + result);
		}
		return result;
	}

	void transactionDone(long maxAllowedReuseTime)
	{
		this.maxAllowedReuseTime = maxAllowedReuseTime;
		this.transaction = null;
		if (debugHttpTransaction) {
			String s = "Transaction-" + transactionCount + " done.";
			if (maxAllowedReuseTime > 0 && reuseAllowed) {
				long now = System.currentTimeMillis();
				s += " Connection is reusable for next " + (maxAllowedReuseTime - now) / 1000.0 + " secs.";
			}
			ln(s);
		}
		if (reuseAllowed)
			connectionStateDescRef.set("In keep-alive.");
		else
			connectionStateDescRef.set("Dead, not reusable.");
	}

	void disableReuse(String why)
	{
		if (!reuseAllowed)
			return;
		reuseAllowed = false;
		if (debugHttpClient.get())
			ln("Disables reuse: " + why);
	}

	/**
	 * Should only be called by CacheableHttpConnection.abort
	 */
	public void forcedSocketClose()
	{
		if (debugHttpClient.get() && !skt.isClosed())
			ln("Closing connection");
		try {
			hasOpenSocket = false;
			skt.close();
		}
		catch (Exception ex) {
			System.err.println("Got exception while closing http socket: " + ExceptionUtils.getExMsg(ex));
		}
	}

	public String getState()
	{
		HttpTransaction t = transaction;
		String state;
		if (t != null && !t.isDone) {
			boolean isSendingRequest = t.hasStartedSendingRequest && !t.hasCompletedSendingRequest.get();
			long secs = (System.currentTimeMillis() - t.transactionStartTime) / 1000;
			state = String.format("%02d:%02d %s%s. %s",
					secs / 60, secs % 60,
					(isSendingRequest ? "Sending request. " : ""),
					t.transactionStateDescRef.get(),
					t.requestData.url);
		} else {
			state = connectionStateDescRef.get();
		}
		if (!reuseAllowed)
			state += ", no keep-alive.";
		return state;
	}

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}
}
