/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.connection;

import java.io.ByteArrayInputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.util.List;
import org.glimmerblocker.proxy.http.data.Headers;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.io.Base64InputStream;
import dk.vitality.lib.foundation.misc.ByteBuffer;

class HttpTransactionSender
{
	@NotNull private final HttpTransaction t;
	@NotNull private final HttpClientConnection conn;
	private final boolean chunked;

	HttpTransactionSender(@NotNull HttpTransaction t) throws Exception
	{
		this.t = t;
		this.conn = t.conn;
		chunked = !t.isQuicktimeRtspPost && t.requestData.body.hasUnknownSize;
		final URL url = t.requestData.url;
		String pathArgs = url.getPath();
		if (pathArgs == null || pathArgs.length() == 0)
			pathArgs = "/";
		if (url.getQuery() != null)
			pathArgs += "?" + url.getQuery();
		//
		StringBuilder sb = new StringBuilder(40);
		sb.append(t.requestData.method);
		sb.append(' ');
		if (conn.connKey.httpProxy != null && !conn.connKey.ssl) {
			sb.append(url.getProtocol());
			sb.append("://");
			sb.append(url.getHost());
			if (url.getPort() >= 80) {
				sb.append(":");
				sb.append(url.getPort());
			}
		}
		sb.append(pathArgs);
		if (t.isQuicktimeRtspPost)
			sb.append(" HTTP/1.0");
		else
			sb.append(" HTTP/1.1");
		String requestLine = sb.toString();
		rawSendString(requestLine);
		if (t.debugHttpClient)
			t.ln(requestLine);
		sendCRLF();
		//
		if (!t.requestData.headers.headerMap.containsKey("host")) {
			// allow client to override the Host.
			String host = url.getHost();
			int port = url.getPort();
			sendOptHeader("Host", (port > 0) ? host + ":" + port : host);
		}
		//
		sendOptHeader("Content-Type", t.requestData.headers.contentType);
		sendOptHeader("Content-Encoding", t.requestData.headers.contentEncoding);
		//
		if (t.requestData.body.contentSize >= 0)
			sendOptHeader("Content-Length", Long.toString(t.requestData.body.contentSize));
		//
		for (String nameLC : t.requestData.headers.headerMap.keySet()) {
			String nameMX = Headers.getMixedCaseHeaderName(nameLC);
			List<String> list = t.requestData.headers.headerMap.get(nameLC);
			for (String value : list)
				sendOptHeader(nameMX, value);
			if (nameLC.equals("upgrade") && !list.isEmpty())
				conn.disableReuse("Has Upgrade: " + list);
		}
		if (!t.allowKeepAlive())
			sendOptHeader("Connection", "close");
		if (chunked)
			sendOptHeader("Transfer-Encoding", "chunked");
		sendCRLF(); // END of headers.
		//
		if (t.requestData.body.isEmpty) {
			conn.os.flush();
			t.hasCompletedSendingRequest.set(true);
			return;
		}
		if (t.requestData.body.contentSize > 0 && t.requestData.body.contentSize <= 32768) {
			if (t.debugHttpClient)
				t.ln("Sends sync request body with content-length = " + t.requestData.body.contentSize);
			sendRequestBody();
			return;
		}
		if (t.debugHttpClient) {
			if (t.requestData.body.hasUnknownSize)
				t.ln("Starts RequestBodySender for unknown request body size.");
			else
				t.ln("Starts RequestBodySender with req-content-length = " + t.requestData.body.contentSize);
		}
		new Thread(new RequestBodySender(), "HttpTransaction.ReqBodySender").start();
	}

	private void sendOptHeader(@NotNull String name, @Nullable String value) throws IOException
	{
		if (value == null)
			return;
		rawSendString(name);
		conn.os.write(':');
		conn.os.write(' ');
		rawSendString(value);
		sendCRLF();
		if (t.debugHttpClient)
			t.ln("  " + name + ": " + value);
	}

	private void rawSendString(String s) throws IOException
	{
		OutputStream os = conn.os;
		int len = s.length();
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch <= 127) {
				os.write(ch);
				continue;
			}
			conn.os.write(s.substring(i).getBytes("UTF-8"));
			return;
		}
	}

	private void sendCRLF() throws IOException
	{
		conn.os.write(13);
		conn.os.write(10);
	}

	void sendRequestBody()
	{
		try {
			OutputStream os = chunked ? new ChunkedOutputStream(conn.os) : conn.os;
			if (t.requestData.body.hasUnknownSize || t.isQuicktimeRtspPost) {
				// can't trust content-length with QuickTime RTSP, Apple says so!
				forwardUntilEof(os);
			} else {
				forwardFixedSize(os);
			}
			if (chunked)
				os.close();
			else
				os.flush();
		}
		catch (Exception ex) {
			// QuickTime sends bogous content-length in request header (way too big), so just
			// ignore "unexpected" errors.
			t.killKeepAlive("Failed sending request body: " + ExceptionUtils.getExMsg(ex));
			if (t.debugHttpClient && !t.isQuicktimeRtspPost)
				t.ln(ExceptionUtils.getStackTraceAsString(ex));
		}
		if (t.debugHttpClient)
			t.ln("Done sending request body");
		t.hasCompletedSendingRequest.set(true);
		t.checkTransactionDone();
		if (t.isQuicktimeRtspPost) {
			try {
				// make sure we kill connection here, because the server will never respond.
				// Keep-alive is already disabled.
				Thread.sleep(1000);
				if (t.debugHttpClient)
					t.ln("RequestBodySender for RTSP POST force-closes connection.");
				conn.os.close();
			}
			catch (Exception ex) {
				if (t.debugHttpClient)
					t.ln("RequestBodySender got error closing output stream: " + ExceptionUtils.getExMsg(ex));
			}
		}
	}

	private void forwardUntilEof(@NotNull OutputStream os) throws IOException
	{
		InputStream is = t.requestData.body.stream;
		if (is == null)
			throw new NullPointerException();
		long sumSize = 0;
		long startTime = System.currentTimeMillis();
		byte[] buffer = new byte[16384];
		while (true) {
			int num = is.read(buffer, 0, buffer.length);
			if (num < 0)
				break;
			if (t.hasReadResponseHeaders)
				continue; // skip sending data.
			if (t.debugHttpClient) {
				sumSize += num;
				long now = System.currentTimeMillis();
				t.ln("  Sends request body bytes: " + num + " bytes => sum = " + sumSize +
						" Duration: " + (now - startTime) + " secs.");
				if (t.isQuicktimeRtspPost && t.debugHttpClient) {
					ByteArrayInputStream bais = new ByteArrayInputStream(buffer, 0, num);
					Base64InputStream b64 = new Base64InputStream(bais);
					ByteBuffer bb = new ByteBuffer(64);
					bb.readFully(b64);
					String s = bb.toString("ISO-8859-1"); // should be UTF-8, but avoid decoding error.
					t.ln("RTSP request:\n" + s);
				}
			}
			os.write(buffer, 0, num);
			if (t.isQuicktimeRtspPost) {
				// In case RTSP sends small chunks and expect the server to immediately respond to them.
				os.flush();
			}
		}
	}

	private void forwardFixedSize(@NotNull OutputStream os) throws IOException
	{
		InputStream is = t.requestData.body.stream;
		if (is == null)
			throw new NullPointerException();
		long remaining = t.requestData.body.contentSize;
		long startTime = System.currentTimeMillis();
		byte[] buffer = new byte[(int)Math.min(16384, remaining)];
		while (remaining > 0) {
			long max = Math.min(buffer.length, remaining);
			int num = is.read(buffer, 0, (int)max);
			if (num < 0)
				throw new EOFException("Unexpected EOF from request body read, expected " + max + " bytes.");
			if (t.hasReadResponseHeaders) {
				if (t.debugHttpClient) {
					t.ln("QT workaround: Skipped sending " + numToString(num) + "/" + numToString(remaining) +
							" request body bytes as response headers have been read.");
				}
				remaining -= num;
				continue;
			}
			if (t.debugHttpClient) {
				long sz = t.requestData.body.contentSize;
				long secs = (System.currentTimeMillis() - startTime) / 1000;
				String s = String.format("" +
						"  Sends request body chunk: %d/%d. Total sent: %s / %s bytes. Duration: %02d:%02d",
						num, max,
						numToString(sz - remaining), numToString(sz),
						secs / 60, secs % 60);
				t.ln(s);
			}
			os.write(buffer, 0, num);
			remaining -= num;
		}
	}

	private static String numToString(long x)
	{
		StringBuilder sb = new StringBuilder();
		sb.append(x);
		for (int i = sb.length() - 3; i > 0; i -= 3)
			sb.insert(i, ',');
		return sb.toString();
	}

	private class RequestBodySender implements Runnable
	{
		public void run()
		{
			sendRequestBody();
		}
	}
}
