/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.client.stream;

import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.Socket;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.io.FileContentsUtils;
import dk.vitality.lib.foundation.misc.ByteBuffer;

public class ChunkedInputStream extends InputStream implements Logger
{
	@NotNull private final Socket skt;
	@NotNull private final InputStream src;
	@Nullable private final Logger logger;
	private long remainingInChunk = -1;
	@NotNull private final StringBuilder sb = new StringBuilder(10);
	@NotNull private State state = State.chunkLengthLine;

	@SuppressWarnings("ConstantConditions")
	@Nullable private final ByteBuffer dumpBuffer = debugChunks ? new ByteBuffer(256) : null;

	enum State
	{
		chunkBody,
		postChunkBodyCRLF,
		chunkLengthLine,
		trailer,
		EOF
	}

	private static final boolean debugChunks = false;
	private static final boolean debugVerbose = false;

	public ChunkedInputStream(@NotNull InputStream src, @NotNull Socket skt, @NotNull Logger parentLogger)
	{
		this.src = src;
		this.skt = skt;
		if (debugChunks)
			this.logger = new PrefixedLogger(parentLogger, "chunked");
		else
			this.logger = null;
	}

	public int read() throws IOException
	{
		handleState();
		if (state == State.EOF)
			return -1;
		int x = src.read();
		if (x < 0)
			handleUnexpectedEndOfStream();
		remainingInChunk--;
		if (dumpBuffer != null)
			dumpBuffer.append(x);
		if (debugChunks)
			ln("Reads char = " + x + ", remaining in chunk = " + remainingInChunk);
		return x;
	}

	public int read(@NotNull byte[] b, int off, int len) throws IOException
	{
		handleState();
		if (state == State.EOF)
			return -1;
		int max = (int)Math.min(len, remainingInChunk);
		if (debugChunks)
			ln("read[], len = " + len + ", max = " + max);
		int num = src.read(b, off, max);
		if (num < 0)
			handleUnexpectedEndOfStream();
		remainingInChunk -= num;
		if (dumpBuffer != null)
			dumpBuffer.append(b, off, num);
		if (debugChunks) {
			ln("  did read " + num + " bytes, remaining in chunk = " + remainingInChunk);
			//ln("{{{" + new String(b, off, len, "ISO-8859-1") + "}}}");
		}
		return num;
	}

	protected void handleState() throws IOException
	{
		if (debugChunks)
			ln("handleState, remainingInChunk = " + remainingInChunk + ", state = " + state);
		switch (state) {
			case EOF:
				return;
			case chunkBody:
				if (remainingInChunk > 0)
					return;
				state = State.postChunkBodyCRLF;
				break;
			default:
				break;
		}
		while (true) {
			int ch = readChunkHeaderByte();
			switch (state) {
				case chunkLengthLine:
					if (ch == 13)
						continue;
					if (ch != 10) {
						sb.append((char)ch);
						continue;
					}
					String s = sb.toString();
					if (debugChunks)
						ln("chunk length line: '" + s + "'");
					int idx = s.indexOf(';');
					if (idx > 0) {
						ln("Ignores chunk extension: " + s);
						s = s.substring(0, idx);
					}
					remainingInChunk = Long.parseLong(s.trim(), 16);
					if (debugChunks)
						ln("remainingInChunk = " + remainingInChunk);
					if (remainingInChunk == 0) {
						skt.setSoTimeout(60 * 1000);
						sb.setLength(0);
						state = State.trailer;
					} else {
						state = State.chunkBody;
					}
					return;
				case postChunkBodyCRLF:
					if (ch == 13)
						continue;
					if (ch != 10)
						throw new IOException("Invalid post-chunk-body char = " + ch);
					sb.setLength(0);
					state = State.chunkLengthLine;
					continue;
				case trailer:
					if (ch == 13)
						continue;
					if (ch != 10) {
						sb.append((char)ch);
						continue;
					}
					if (sb.length() == 0) {
						state = State.EOF;
						if (debugChunks)
							ln("Got EOF line in trailer chunk => state = " + state);
						if (dumpBuffer != null) {
							File f = new File("/tmp/chunked-dump.bin");
							ln("Dumps " + dumpBuffer.length() + " bytes to " + f.getPath());
							FileContentsUtils.saveFile(f, dumpBuffer.toBytes());
							dumpBuffer.clear();
						}
						notifyReachedEndOfExpectedSize();
						return;
					}
					ln("Ignores chunked trailer line: " + sb);
					sb.setLength(0);
					continue;
				case chunkBody:
				default:
					throw new IOException("Invalid state = " + state + " inside handleState loop");
			}
		}
	}

	private int readChunkHeaderByte() throws IOException
	{
		int ch = src.read();
		if (debugVerbose)
			ln("Chunk header byte: " + ch);
		if (ch < 0)
			handleUnexpectedEndOfStream();
		return ch;
	}

	public void ln(@Nullable String s)
	{
		if (logger != null)
			logger.ln(s);
	}

	protected void notifyReachedEndOfExpectedSize() throws IOException
	{
	}

	/**
	 * MUST throw IOException
	 */
	protected void handleUnexpectedEndOfStream() throws IOException
	{
		throw new EOFException("Server closed connection inside chunked stream");
	}
}
