/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.data;

import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.html.URLUtils;

public class UrlParams implements Iterable<String>
{
	@NotNull
	private final Map<String, String[]> args = new HashMap<>();

	/**
	 * Creates empty hash.
	 */
	public UrlParams()
	{
	}

	/**
	 * Gets an array of parameter values for non-empty parameter values.
	 *
	 * @return null if no such (non-empty) parameter was specified.
	 */
	public String[] optStringArray(@NotNull String name)
	{
		return args.get(name);
	}

	/**
	 * Is the named parameter specified and non-empty.
	 */
	public boolean has(String name)
	{
		return optStringArray(name) != null;
	}

	/**
	 * Get a parameter value and default to some value if the parameter was not specified (or was empty).
	 */
	@Nullable
	public String optString(@NotNull String name, @Nullable String defaultValue)
	{
		String[] array = optStringArray(name);
		if (array == null || array.length == 0)
			return defaultValue;
		return array[0];
	}

	/**
	 * Get a parameter value and default to null if the parameter was not specified (or was empty).
	 */
	@Nullable
	public String optString(@NotNull String name)
	{
		return optString(name, null);
	}


	/**
	 * Get a required string parameter. Throws MissingServletArgumentException if not specified (or is empty).
	 */
	@NotNull
	public String reqString(@NotNull String name) throws UrlParams.NoSuchParameterException
	{
		String s = optString(name);
		if (s == null)
			throw new UrlParams.NoSuchParameterException(name);
		return s;
	}

	//-------------------------------------------------------------

	/**
	 * Get a required boolean parameter. Throws MissingServletArgumentException if not specified (or is empty).
	 */
	@SuppressWarnings("unused")
	public boolean reqBoolean(@NotNull String name) throws Exception
	{
		return convertStringToBoolean(reqString(name));
	}

	/**
	 * Get an optional boolean parameter. Returns false if not specified.
	 */
	@SuppressWarnings("unused")
	public boolean optBoolean(@NotNull String name) throws Exception
	{
		return optBoolean(name, false);
	}

	/**
	 * Get an optional boolean parameter. Returns the specified defaultValue if not specified.
	 */
	public boolean optBoolean(@NotNull String name, boolean defaultValue) throws Exception
	{
		String s = optString(name);
		return (s == null) ? defaultValue : convertStringToBoolean(s);
	}

	/**
	 * Convert a string value into a boolean. Accepts null, "false" and "0" as false.
	 */
	private boolean convertStringToBoolean(@Nullable String s) throws Exception
	{
		return s != null && !"0".equals(s) && !"false".equalsIgnoreCase(s);
	}

	//-------------------------------------------------------------

	public int reqInt(@NotNull String name) throws Exception
	{
		return Integer.parseInt(reqString(name));
	}

	public int optInt(@NotNull String name, int defaultValue) throws Exception
	{
		String s = optString(name);
		return (s == null) ? defaultValue : Integer.parseInt(s);
	}

	//-------------------------------------------------------------

	/**
	 * Similar to Collection.isEmpty: true if no parameters is specified.
	 */
	public boolean isEmpty()
	{
		return args.isEmpty();
	}

	/**
	 * Returns iterator for the parameter names.
	 */
	public Iterator<String> iterator()
	{
		return args.keySet().iterator();
	}

	/**
	 * Returns iterator for the parameter names.
	 */
	public Iterator<String> names()
	{
		return args.keySet().iterator();
	}

	//=========================================================================================
	/**
	 * Decode and add parameter (and values) from a query-string.
	 *
	 * @throws java.io.UnsupportedEncodingException
	 *
	 */
	public void addFromQueryString(@Nullable String qs) throws UnsupportedEncodingException
	{
		if (qs == null)
			return;
		while (qs.length() > 0) {
			int sepIdx = qs.indexOf('&');
			String s = (sepIdx < 0) ? qs : qs.substring(0, sepIdx);
			qs = (sepIdx < 0) ? "" : qs.substring(sepIdx + 1);
			int eq = s.indexOf('=');
			if (eq < 0)
				continue;
			String name = decodeAndCleanString(s.substring(0, eq));
			String value = decodeAndCleanString(s.substring(eq + 1));
			if (value.length() == 0)
				continue;
			//System.err.println("'" + name + "' -> ' " + value + "'");
			String[] oldArray = args.get(name);
			String[] newArray;
			if (oldArray == null) {
				newArray = new String[1];
				newArray[0] = value;
			} else {
				newArray = new String[oldArray.length + 1];
				System.arraycopy(oldArray, 0, newArray, 0, oldArray.length);
				newArray[oldArray.length] = value;
			}
			args.put(name, newArray);
		}
	}

	/**
	 * Clean the specified parameter value: the string is trim()ed and CRLF is changed into LF.
	 */
	private static String decodeAndCleanString(String s)
	{
		s = URLUtils.decodeURL(s, true);
		s = s.replace('\u00A0', ' ');  // make nbsp into normal space so it is zapped by trim()
		if (s.indexOf('\r') < 0)
			return s.trim();
		for (int i = s.length() - 2; i >= 0; i--) {
			char ch = s.charAt(i);
			if (ch != '\r')
				continue;
			if (s.charAt(i + 1) != '\n')
				continue;
			s = i > 0 ? s.substring(0, i) + s.substring(i + 1) : s.substring(1);
		}
		s = s.replace('\r', '\n').trim();
		return s;
	}

	@SuppressWarnings("unused")
	static class NoSuchParameterException extends Exception
	{
		NoSuchParameterException()
		{
		}

		NoSuchParameterException(String string)
		{
			super(string);
		}

		NoSuchParameterException(String string, Throwable throwable)
		{
			super(string, throwable);
		}

		NoSuchParameterException(Throwable throwable)
		{
			super(throwable);
		}
	}
}
