/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.http.server;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.globals.network.NetworkConfig;
import org.glimmerblocker.proxy.http.client.cache.HttpClientCache;
import org.glimmerblocker.proxy.http.client.connection.KeepAliveException;
import org.glimmerblocker.proxy.http.client.stream.SocketTimeoutableInputStream;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.exception.ExceptionUtils;

class HttpServerConnection implements Logger, Runnable
{
	@NotNull final GlimmerGlobals globals;
	@NotNull final Socket skt;
	@NotNull final Logger logger;
	final int connectionId = nextConnectionId++;

	boolean useKeepAlive = true;
	SocketTimeoutableInputStream is;
	OutputStream os;
	int requestCount;

	private HttpClientCache clientCache;
	private NetworkConfig clientCacheNetworkConfig;

	private static int nextConnectionId;

	HttpServerConnection(@NotNull GlimmerGlobals globals, @NotNull Socket skt)
	{
		this.globals = globals;
		this.skt = skt;
		String name = "httpsrv-" + connectionId;
		this.logger = new PrefixedLogger(globals, name);
		Thread t = new Thread(this, name);
		t.start();
	}

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}

	public void run()
	{
		try {
			skt.setTcpNoDelay(true);
			is = new SocketTimeoutableInputStream(new BufferedInputStream(skt.getInputStream(), 4096), skt);
			os = new BufferedOutputStream(skt.getOutputStream(), 4096);
			while (useKeepAlive) {
				requestCount++;
				new HttpServerRequest(this, requestCount).main();
			}
		}
		catch (KeepAliveException ex) {
			if (globals.debugHttpServer)
				ln("Got KeepAliveException: " + ExceptionUtils.getExMsg(ex));
		}
		catch (IOException ex) {
			if (globals.debugHttpServer) {
				ln("Got IOException: " + ExceptionUtils.getExMsg(ex));
				ln(ExceptionUtils.getStackTraceAsString(ex));
			}
		}
		catch (Throwable ex) {
			ln("HttpServerConnection got exception: " + ExceptionUtils.getExMsg(ex));
			if (globals.debugHttpServer)
				ln(ExceptionUtils.getStackTraceAsString(ex));
		}
		try {
			skt.close();
		}
		catch (Throwable ex) {
			ln("Got exception closing socket: " + ExceptionUtils.getExMsg(ex));
		}
		if (clientCache != null)
			clientCache.close();
	}

	public void disableKeepalive(@NotNull String why)
	{
		if (!useKeepAlive)
			return;
		useKeepAlive = false;
		if (globals.debugHttpServer)
			ln("disables keep-alive: " + why);
	}

	HttpClientCache getHttpClientCache()
	{
		NetworkConfig c = globals.networkManager.getConfig();
		if (clientCache == null || !c.equals(clientCacheNetworkConfig)) {
			if (clientCache != null) {
				clientCache.close();
				clientCache = null;
			}
			clientCache = new HttpClientCache(globals.dnsResolver, this, globals.debugHttpClient);
			clientCacheNetworkConfig = c;
		}
		return clientCache;
	}
}
