/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.proxyRequest;

import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.util.ArrayList;
import java.util.List;
import org.glimmerblocker.proxy.http.client.BodyCollector;
import org.glimmerblocker.proxy.http.client.connection.HttpResponseInfo;
import org.glimmerblocker.proxy.http.client.stream.TimeoutableInputStream;
import org.glimmerblocker.proxy.http.data.CharsetDeterminer;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.proxyRequest.transform.MimeTypeMatcher;
import org.glimmerblocker.proxy.proxyRequest.transform.Transformer;
import org.glimmerblocker.proxy.rhino.callback.HtmlHelper;
import org.glimmerblocker.proxy.rule.AbstractRule;
import org.glimmerblocker.proxy.rule.BlockerRule;
import org.glimmerblocker.proxy.rule.ModificationRule;
import org.glimmerblocker.proxy.rule.RequestRule;
import org.glimmerblocker.proxy.rule.WhitelistRule;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

class ForwardStreamerTransformers extends ForwardStreamerBase
{
	@NotNull List<RequestRule> origRequestRules;
	@NotNull List<AbstractRule> ruleList;

	ForwardStreamerTransformers(@NotNull ProxyRequestHandler rh,
	                            @NotNull Headers headers,
	                            @NotNull List<RequestRule> origRequestRules,
	                            @NotNull List<AbstractRule> ruleList) throws Exception
	{
		super(rh, headers);
		this.origRequestRules = origRequestRules;
		this.ruleList = ruleList;
	}

	protected void sendResponse(@NotNull HttpResponseInfo hri,
	                            @Nullable TimeoutableInputStream bodyStream) throws Exception
	{
		if (ruleList.isEmpty() && globals.debugFilters)
			ln("Bug: ruleList is empty when entering ForwardStreamerTransformers.sendReponse");
		boolean isHtml = MimeTypeMatcher.Html.matches(hri.headers.status, hri.headers.contentTypeBaseLC);
		List<ModificationRule> matchingTransformers = new ArrayList<>(ruleList.size());
		int minUserPriority = Integer.MIN_VALUE;
		for (AbstractRule rule : ruleList) {
			if (rule.userPriority < minUserPriority) {
				if (globals.debugFilters)
					ln("  Ignores remaining rules due to lower priority than matched transformers: " + rule);
				break;
			}
			if (rule instanceof ModificationRule) {
				ModificationRule trule = (ModificationRule)rule;
				boolean hasSimpleTransform = isHtml && (trule.css != null || trule.js != null);
				MimeTypeMatcher mtm = trule.transformContentTypeMatcher;
				boolean hasJavascripTransform = (mtm != null)
						&& mtm.matches(hri.headers.status, hri.headers.contentTypeBaseLC);
				if (hasSimpleTransform || hasJavascripTransform) {
					matchingTransformers.add(trule);
					minUserPriority = trule.userPriority;
					if (globals.debugFilters)
						ln("  Transformer matched: " + trule);
				} else {
					if (globals.debugFilters)
						ln("  Transformer didn't match. Ignored: " + trule);
				}
				continue;
			}
			if (rule instanceof WhitelistRule) {
				if (!matchingTransformers.isEmpty()) {
					if (globals.debugFilters)
						ln("  Matched whitelist after transformer => uses found transformers.");
					break;
				}
				if (globals.debugFilters)
					ln("  First match is whitelist => simple stream of response.");
				history.addAction(rule, "whitelisted");
				history.commit();
				streamResponse(hri, bodyStream);
				return;
			}
			if (rule instanceof BlockerRule) {
				if (!matchingTransformers.isEmpty()) {
					if (globals.debugFilters)
						ln("  Matched blocker after transformer => uses found transformers.");
					break;
				}
				if (globals.debugFilters)
					ln("  First match is blocker => ignore response data and send simple block.");
				prh.sendBlockerResult((BlockerRule)rule);
				return;
			}
			ln("  Unsupported rule class: " + rule.getClass().getName());
		}
		if (matchingTransformers.isEmpty()) {
			if (globals.debugFilters)
				ln("  Found no matching transformers => simple stream of response.");
			streamResponse(hri, bodyStream);
			return;
		}
		if (hri.isContentLessResponse) {
			if (globals.debugFilters)
				ln("  Is content-less response => simple stream of response.");
			streamResponse(hri, bodyStream);
			return;
		}
		BodyCollector bc = new BodyCollector(hri);
		bc.readBody(bodyStream);
		ResultData rd = bc.getResultData();
		String text;
		boolean debugCharset = globals.debugFilters && globals.debugVerbose;
		CharsetDeterminer cd;
		try {
			cd = new CharsetDeterminer(rd, this, debugCharset);
			text = cd.getContentAsText();
		}
		catch (UnsupportedEncodingException ex) {
			if (globals.debugFlag)
				ln("Got UnsupportedEncodingException(" + ex.getMessage() + "): sends content unmodified.");
			history.addAction(matchingTransformers.get(0), "failed-charset-decoding");
			history.commit();
			prh.sendResultData(rd);
			return;
		}
		if (globals.debugFilters) {
			ln("Got " + text.length() + " chars content" +
					" for " + matchingTransformers.size() + " transformers." +
					" isHtml = " + isHtml + ", base content-type = '" + rd.headers.contentTypeBaseLC + "'.");
			if (rd.headers.contentEncoding != null)
				ln(" using content-encoding: " + rd.headers.contentEncoding);
		}
		//
		Transformer t = new Transformer(globals, this, matchingTransformers, prh, this, rd.headers, text, history);
		text = t.rd.getContentAsText(this, false);
		if (globals.debugFilters && t.didTransformContent)
			ln("Base content-type after transformers: " + t.rd.headers.contentTypeBaseLC);
		//
		if (isHtml) {
			HtmlHelper hc = new HtmlHelper(text);
			// Do not force insert as many php sites sends out javascript with content-type text/html
			final boolean forceInsert = false;
			for (ModificationRule rule : matchingTransformers) {
				if (rule.cssElement != null) {
					history.addAction(rule, "css-added");
					hc.insertAtHeadEnd(rule.cssElement, forceInsert);
				}
				if (rule.jsElement != null) {
					history.addAction(rule, "javascript-added");
					hc.insert(rule.jsElement, rule.jsPlacement, forceInsert);
				}
			}
			text = hc.html;
		}
		if (globals.debugFilters) {
			ln("Got " + text.length() + " chars content after " +
					matchingTransformers.size() + " transformers.");
		}
		history.commit();
		try {
			// The result must be sent in the same charset as the server originally used
			// as the browser will use this charset when submitting forms, and many sites have hardcoded
			// assumptions about which charset is used in form submission.
			// http://web.archive.org/web/20060427015200/ppewww.ph.gla.ac.uk/~flavell/charset/form-i18n.html
			Headers h = t.rd.headers;
			if (cd.hasCharsetInHttpHeader()) {
				h = h.copyWithContentType(h.contentTypeBaseLC + "; charset=" + cd.httpCharset);
			} else {
				// As original response from server did not have charset option, make
				// sure the response doesn't add a charset option as some browsers doesn't
				// default to ISO-8559-1, but uses the charset of the html page (IE6).
				// So even though we 'know' the charset, let browsers use their default
				// instead of telling them a truth that's wrong.
				h = h.copyWithContentType(h.contentTypeBaseLC);
			}
			boolean shouldEncode = shouldEncode16bits(cd.javaCharset);
			if (debugCharset) {
				ln("Using http-charset = '" + cd.httpCharset + "', java-charset = '" + cd.javaCharset + "'" +
						", encodes 16-bit = " + shouldEncode + "" +
						". Content-type: " + h.contentType);
			}
			if (shouldEncode)
				text = encode16bits(text);
			rd = new ResultData(h, text.getBytes(cd.javaCharset));
			prh.sendResultData(rd);
		}
		catch (UnsupportedEncodingException ex) {
			if (globals.debugFlag)
				ln("Got UnsupportedEncodingException(" + ex.getMessage() + "): sends content unmodified.");
			prh.sendResultData(rd);
		}
	}

	protected boolean processRequestRules() throws Exception
	{
		return (redirectCount == 0) ? super.processRequestRules(origRequestRules) : super.processRequestRules();
	}

	@SuppressWarnings({"SimplifiableIfStatement"})
	private static boolean shouldEncode16bits(String charset)
	{
		if (charset.equalsIgnoreCase("UTF-8")
				|| charset.equalsIgnoreCase("UTF8")
				|| charset.equalsIgnoreCase("euc-cn")) {
			return false;
		}
		return charset.length() < 3
				|| Character.toUpperCase(charset.charAt(0)) != 'G'
				|| Character.toUpperCase(charset.charAt(1)) != 'B';
	}

	/**
	 * transformation might have added 16 bit chars which needs &#...; encoding for ISO-8859-1
	 */

	private static String encode16bits(String s)
	{
		StringBuilder sb = new StringBuilder(s.length() + 16);
		int len = s.length();
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch <= 255)
				sb.append(ch);
			else
				sb.append('&').append('#').append((int)ch).append(';');
		}
		return sb.toString();
	}
}
