/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.proxyRequest;

import java.io.IOException;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.SocketException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import org.glimmerblocker.proxy.admin.AdminDispatcher;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.history.HistoryEntryBuilder;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.globals.network.NetworkConfig;
import org.glimmerblocker.proxy.http.client.connection.HttpResponseInfo;
import org.glimmerblocker.proxy.http.client.stream.TimeoutableInputStream;
import org.glimmerblocker.proxy.http.data.Headers;
import org.glimmerblocker.proxy.http.data.HeadersBuilder;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.http.server.HttpServerRequest;
import org.glimmerblocker.proxy.proxyRequest.errorPage.ErrorPage;
import org.glimmerblocker.proxy.proxyRequest.keyword.KeywordExecutor;
import org.glimmerblocker.proxy.rule.AbstractRule;
import org.glimmerblocker.proxy.rule.BlockerRule;
import org.glimmerblocker.proxy.rule.RequestRule;
import org.glimmerblocker.proxy.util.HttpException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import dk.vitality.lib.foundation.http.AcceptEncodingUtils;
import static dk.vitality.lib.foundation.exception.ExceptionUtils.getExMsg;

/**
 * The main class for handling HTTP proxy requests.
 */
public class ProxyRequestHandler implements Logger
{
	/**
	 * Preferences etc.
	 */
	@NotNull public final GlimmerGlobals globals;

	/**
	 * The server/Safari http request object.
	 */
	@NotNull public final HttpServerRequest request;

	/**
	 * If this is method "HEAD".
	 */
	public final boolean isHEAD;

	/**
	 * Network configuration: http proxy, socks proxy, dns server.
	 */
	@NotNull public NetworkConfig networkConfig;

	/**
	 * Does the client (e.g. Safari) allow gzip?
	 */
	public final boolean clientAcceptEncodingAllowsGzip;

	/**
	 * The request serial id.
	 * Used in logging prefix to make it easy to see which requests which made which log lines.
	 */
	public final int id = nextHandlerid++;
	//
	private static int nextHandlerid = 100;

	private final Logger logger;
	/**
	 * Is the initial log line (containing HTTP method and request URL)
	 * for this request been emitted?
	 */
	private boolean hasMadeLogHeader;

	/**
	 * URL of the requested content.
	 */
	public URL requestURL;

	/**
	 * The path part of {@link #requestURL}. Contains "/" instead of empty string if no path was specified.
	 */
	public String requestPath;

	/**
	 * Is this a proxy request (request path starts with e.g. "http://") or a local request
	 * such as "GET /favicon.ico HTTP/1.1".
	 */
	boolean isProxyRequest;

	public ProxyRequestHandler(@NotNull HttpServerRequest request,
	                           @NotNull GlimmerGlobals globals) throws Exception
	{
		this.request = request;
		this.globals = globals;
		this.isHEAD = request.method.equals("HEAD");
		this.logger = new PrefixedLogger(globals, Integer.toString(id));
		this.clientAcceptEncodingAllowsGzip =
				AcceptEncodingUtils.allowsGzip(request.headers.getLC("accept-encoding"));
		this.networkConfig = globals.networkManager.getConfig();
		if (globals.debugFlag)
			ln(null);
		if (globals.debugVerbose)
			ln("RequestHandler: " + Thread.currentThread().getName());
	}

	/**
	 * The main method for executing the request.
	 */
	public void main() throws Exception
	{
		try {
			checkMethodIsAllowed();
			int idx = request.path.indexOf('?');
			if (handleAdmin(idx >= 0 ? request.path.substring(0, idx) : request.path))
				return;
			parseUri();
			if (requestURL.getPort() == globals.httpServer.getListenPort()
					&& requestURL.getHost().equals("127.0.0.1")
					&& handleAdmin(requestPath)) {
				return; // proxied admin request.
			}
			execute();
		}
		catch (ClientAbortedException ex) {
			if (globals.debugFlag)
				ln("Got ClientAbortedException: " + ex.getMessage());
		}
		catch (HttpException ex) {
			if (!request.hasSentResponseHeaders()) {
				String ua = request.headers.getLC("user-agent");
				ResultData rd = new ErrorPage(globals, ex, ua, requestURL).getResultData();
				sendResultData(rd);
			} else {
				String msg = "Got HttpException after headers have been sent: " + ex.getMessage();
				if (globals.debugFlag || !(ex.getCause() instanceof SocketException))
					ln(msg);
				request.disableKeepalive(msg);
			}
		}
		catch (Exception ex) {
			ln("GOT exception: " + ex.getMessage());
			ex.printStackTrace();
		}
		finally {
			if (globals.debugVerbose)
				ln("All done.");
		}
	}

	/**
	 * Do we allow the requested method, or do we outright deny it (such as TRACE).
	 */
	protected void checkMethodIsAllowed() throws HttpException
	{
		if (request.method.equals("TRACE")) {
			Headers h = new Headers(HttpURLConnection.HTTP_BAD_METHOD, "Method " + request.method + " not allowed");
			throw new HttpException(h);
		}
		String s = request.headers.getLC("upgrade");
		if (s != null) {
			ln("Does not support Upgrade: " + s);
			Headers h = new Headers(HttpURLConnection.HTTP_NOT_ACCEPTABLE,
					"Does not support upgrading protocol: 'Upgrade: " + s + "'");
			throw new HttpException(h);
		}
	}

	protected boolean handleAdmin(String path) throws Exception
	{
		if (!path.startsWith(AdminDispatcher.pathPrefix))
			return false;
		String host = request.headers.getLC("host");
		if (host == null || (!host.equals("127.0.0.1") && !host.startsWith("127.0.0.1:")))
			return false;
		globals.adminDispatcher.handle(this, path);
		return true;
	}

	/**
	 * Parse the request path sent from the client. It must be a full URL as this
	 * is a proxy server.
	 *
	 * @return true if it's a proxy request and false if it's an admin request
	 */
	protected void parseUri() throws Exception
	{
		try {
			if (request.path.startsWith("/")) {
				isProxyRequest = false;
				String host = request.headers.getLC("host");
				if (host == null || host.length() == 0)
					host = "127.0.0.1:8228";
				requestURL = new URL("http://" + host + request.path);
			} else {
				isProxyRequest = true;
				requestURL = new URL(request.path);
			}
			String s = requestURL.getPath();
			if (s == null || s.length() == 0)
				s = "/";
			requestPath = s;
		}
		catch (Exception ex) {
			Headers h = new Headers(HttpURLConnection.HTTP_BAD_REQUEST,
					"Bad request URL (" + getExMsg(ex) + "): " + request.path);
			throw new HttpException(h, ex);
		}
		String scheme = requestURL.getProtocol();
		if (scheme == null || (!scheme.equals("http") && !scheme.equals("https"))) {
			Headers h = new Headers(HttpURLConnection.HTTP_FORBIDDEN,
					"Disallowed scheme = '" + scheme + "'");
			throw new HttpException(h);
		}
	}

	/**
	 * Execute the request after method and URL has been validated/decoded.
	 * <p>
	 * Handles blocked URLs.
	 */
	protected void execute() throws Exception
	{
		String host = requestURL.getHost();
		if (!globals.filtersEnabled) {
			forwardResponseUnmodified(Collections.<RequestRule>emptyList());
			return;
		}
		if (host.startsWith(globals.javaSettings.getAttachmentDownloadHostPrefix())) {
			new ForwardStreamAttachmentDownloader(this, getForwardRequestHeaders(), requestPath).execute();
			return;
		}
		if (KeywordExecutor.handleKeyword(this, host))
			return;
		String query = requestURL.getQuery();
		if (query == null)
			query = "";
		boolean mobile = request.headers.hasMobileUserAgent();
		List<AbstractRule> ruleList = globals.ruleCollection.getRules(host, requestPath, query, mobile);
		if (ruleList.isEmpty()) {
			forwardResponseUnmodified(Collections.emptyList());
			return;
		}
		if (globals.debugFilters) {
			ln(ruleList.size() == 1 ? "1 rule matched:" : ruleList.size() + " rules matched:");
			for (int i = 0; i < ruleList.size(); i++)
				ln("   [" + i + "]: " + ruleList.get(i));
		}
		AbstractRule r0 = ruleList.get(0);
		if (r0 instanceof BlockerRule) {
			if (globals.debugFilters)
				ln("  First matching rule is blocker => simple block of URL.");
			sendBlockerResult((BlockerRule)r0);
			return;
		}
		List<RequestRule> requestRules = new ArrayList<>(ruleList.size());
		int idx = 0;
		int priority = -1;
		while (idx < ruleList.size()) {
			AbstractRule rule = ruleList.get(idx);
			if (rule instanceof RequestRule) {
				ruleList.remove(idx);
				if (priority < 0 || rule.userPriority == priority) {
					requestRules.add((RequestRule)rule);
					priority = rule.userPriority;
				}
			} else {
				idx++;
			}
		}
		if (ruleList.isEmpty())
			forwardResponseUnmodified(requestRules);
		else
			new ForwardStreamerTransformers(this, getForwardRequestHeaders(), requestRules, ruleList).execute();
	}

	void sendBlockerResult(BlockerRule rule) throws Exception
	{
		HistoryEntryBuilder heb = makeHistoryEntryBuilder();
		heb.addAction(rule, "blocked");
		heb.commit();
		String p = requestURL.getPath();
		if (p == null)
			p = "/";
		if (p.endsWith(".js") || requestURL.getHost().endsWith(".adtech.de")) {
			sendResultData(new ResultData(new Headers("application/x-javascript")));
			return;
		}
		if (p.endsWith(".swf")) {
			sendResultData(ResultData.NOT_FOUND);
			return;
		}
		if (p.endsWith(".html") || p.endsWith(".htm") || p.endsWith(".php")) {
			String q = requestURL.getQuery();
			if (q != null && q.contains(".gif"))
				sendImageBlocker();
			else
				sendResultData(new ResultData(new Headers("text/html")));
			return;
		}
		if (p.endsWith(".css")) {
			sendResultData(new ResultData(new Headers("text/css")));
			return;
		}
		sendImageBlocker();
	}

	private void sendImageBlocker() throws Exception
	{
		if (globals.useCheckerboardPatternWhenBlocking)
			sendResultData(ResultData.checkerboardPatternGif);
		else
			sendResultData(ResultData.transparentGif);
	}

	public HistoryEntryBuilder makeHistoryEntryBuilder()
	{
		return globals.historyCache.makeBuilder(requestURL.toExternalForm(), request.headers.getLC("referer"));
	}

	private void forwardResponseUnmodified(@NotNull List<RequestRule> origRequestRules) throws Exception
	{
		if (!isProxyRequest && origRequestRules.isEmpty()) {
			sendResultData(ResultData.NOT_FOUND);
			return;
		}
		ForwardStreamerBase streamer = new UnmodifiedStreamer(this, getForwardRequestHeaders(), origRequestRules);
		streamer.execute();
	}

	class UnmodifiedStreamer extends ForwardStreamerBase
	{
		@NotNull List<RequestRule> origRequestRules;

		protected UnmodifiedStreamer(@NotNull ProxyRequestHandler prh,
		                             @NotNull Headers requestHeaders,
		                             @NotNull List<RequestRule> origRequestRules) throws Exception
		{
			super(prh, requestHeaders);
			this.origRequestRules = origRequestRules;
		}

		protected void sendResponse(@NotNull HttpResponseInfo hri,
		                            @Nullable TimeoutableInputStream bodyStream) throws Exception
		{
			globals.suspects.process(this, request.method, requestURL, request.headers, responseHeaders);
			streamResponse(hri, bodyStream);
		}

		/*
		Safari (and Chrome) now uses combined URL/search bar, so don't expand simple host names.
		@Override
		protected void executeWithoutExceptionHandling(HttpClientCache clientCache) throws Exception
		{
			try {
				super.executeWithoutExceptionHandling(clientCache);
			}
			catch (UnknownHostException ex) {
				if (requestURL.getPort() > 0 && requestURL.getPort() != 80)
					throw ex;
				if (!simpleHostRE.matcher(requestURL.getHost()).matches())
					throw ex;
				String ua = request.headers.getLC("user-agent");
				if (ua == null || !ua.contains("Safari/") || ua.contains("Chrome/"))
					throw ex;
				// Some users report that Safari doesn't always do the www...com expansion.
				String host = "www." + requestURL.getHost() + ".com";
				HeadersBuilder hb = HeadersBuilder.makeErrorTextPlain(
						HttpURLConnection.HTTP_MOVED_TEMP, "moved to " + host);
				hb.add("location", "http://" + host);
				// can't just throw HttpException as the error page drops the 'location' header.
				sendResultData(new ResultData(hb.toHeaders(), "Moved to " + host + "\n"));
			}
		}
		*/

		protected boolean processRequestRules() throws Exception
		{
			return (redirectCount == 0) ? super.processRequestRules(origRequestRules) : super.processRequestRules();
		}
	}

	private Headers getForwardRequestHeaders() throws Exception
	{
		HeadersBuilder hb = HeadersBuilder.makeEmpty();
		hb.contentType = request.headers.contentType;
		hb.contentEncoding = request.headers.contentEncoding;
		hb.copyBrowserHeaders(request.headers, HeadersBuilder.HeaderSelection.Maximal);
		String host = requestURL.getHost();
		if (host != null && !host.isEmpty()) {
			int port = requestURL.getPort();
			String scheme = requestURL.getProtocol();
			if (scheme != null && port > 0) {
				int defaultPort = (scheme.equals("http") ? 80 : scheme.equals("https") ? 443 : -1);
				if (port != defaultPort)
					host += ":" + port;
			}
		} else {
			List<String> list = request.headers.headerMap.get("host");
			if (!list.isEmpty())
				host = list.get(0);
		}
		if (host != null && !host.isEmpty())
			hb.add("host", host);
		hb.setAllowGzip(clientAcceptEncodingAllowsGzip);
		return hb.toHeaders();
	}

	/**
	 * See {@link Logger#ln}
	 */
	public void ln(@Nullable String s)
	{
		if (!hasMadeLogHeader) {
			hasMadeLogHeader = true;
			logger.ln(request.method + "  " + request.path);
		}
		if (s == null)
			return;
		logger.ln(s);
	}

	/**
	 * Send the specified HTTP headers and content to the client.
	 */
	public void sendResultData(@NotNull ResultData rd) throws Exception
	{
		try {
			byte[] b = rd.getContent();
			request.sendResponseHeaders(rd.headers, (b == null) ? 0 : b.length);
			if (b == null || b.length == 0)
				return;
			OutputStream os = request.getOutputStream();
			os.write(b);
			os.close();
		}
		catch (IOException ex) {
			throw new ClientAbortedException(ex);
		}
	}


}
