/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.proxyRequest.keyword;

import java.net.HttpURLConnection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.history.HistoryEntryBuilder;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.http.data.HeadersBuilder;
import org.glimmerblocker.proxy.http.data.ResultData;
import org.glimmerblocker.proxy.proxyRequest.ProxyRequestHandler;
import org.glimmerblocker.proxy.rhino.runtime.ExecutorDelegate;
import org.glimmerblocker.proxy.rule.KeywordRule;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.RhinoException;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;
import dk.vitality.lib.foundation.exception.ExceptionUtils;
import dk.vitality.lib.foundation.html.URLUtils;
import dk.vitality.lib.foundation.misc.StringUtils;

// xn--wiki%20%20letters-tfbo20a
public class KeywordExecutor implements Logger
{
	@NotNull final GlimmerGlobals globals;
	@NotNull final ProxyRequestHandler prh;
	@NotNull final String keyword;
	@NotNull final String term;

	public static boolean handleKeyword(@NotNull ProxyRequestHandler prh,
	                                    @NotNull String host) throws Exception
	{
		int spaceIdx = host.indexOf("%20");
		if (spaceIdx > 0) {
			String h = (host.startsWith("www.") && host.endsWith(".com")) ? host.substring(4, host.length() - 4) : host;
			h = URLUtils.decodeURL(h, true);
			if (h.startsWith("xn--"))
				h = Punycode.decode(h.substring(4));
			spaceIdx = h.indexOf(' ');
			if (spaceIdx < 0) {
				prh.ln("Lost space from keyword host: '" + host + "' -> '" + h + "'");
				return false;
			}
			String keyword = h.substring(0, spaceIdx);
			String term = h.substring(spaceIdx + 1);
			KeywordExecutor ke = new KeywordExecutor(prh, keyword, term);
			return ke.execute();
		}
		if (!prh.requestURL.getHost().equals("www.google.com"))
			return false;
		if (prh.request.headers.getLC("referer") != null) // search from Google page.
			return false;
		boolean debugFlag = prh.globals.debugFlag;
		String q = "&" + prh.requestURL.getQuery() + "&";
		Matcher m = qRE.matcher(q);
		if (!m.find()) {
			if (debugFlag)
				prh.ln("Keyword: missing q=  in query.");
			return false;
		}
		String s = URLUtils.decodeURL(m.group(1), true).trim();
		spaceIdx = s.indexOf(' ');
		if (spaceIdx < 0) {
			if (debugFlag)
				prh.ln("No space in query: '" + s + "'");
			return false;
		}
		String keyword = s.substring(0, spaceIdx).trim();
		String term = s.substring(spaceIdx + 1).trim();
		KeywordExecutor ke = new KeywordExecutor(prh, keyword, term);
		return ke.execute();
	}

	private static final Pattern qRE = Pattern.compile("&q=([^&]+)&");

	public KeywordExecutor(@NotNull ProxyRequestHandler prh,
	                       @NotNull String keyword,
	                       @NotNull String term) throws Exception
	{
		this.globals = prh.globals;
		this.prh = prh;
		this.keyword = keyword;
		this.term = term;
	}

	private boolean execute() throws Exception
	{
		if (globals.debugFlag)
			ln("Execute keyword: '" + keyword + "' using term '" + term + "'");
		for (KeywordRule rule : globals.ruleCollection.getKeywordRules(keyword)) {
			if (executeRule(rule))
				return true;
		}
		if (globals.debugFlag)
			ln("Had no (successful) keyword handlers for '" + keyword + "'");
		return false;
	}

	private boolean executeRule(KeywordRule rule) throws Exception
	{
		if (globals.debugFlag)
			ln("Executes rule: " + rule.keywordMatchType.debugDescription(rule.keyword));
		if (!rule.useJavascript) {
			String url = StringUtils.replaceAll(rule.urlTemplate, "%s", URLUtils.encodeURL(term));
			url = StringUtils.replaceAll(url, "%S", term); // capital-S = unencoded.
			url = StringUtils.replaceAll(url, "%t", latin1encode(term));
			makeHistoryEntry(rule);
			sendRedirect(url);
			return true;
		}
		KeywordExecutorDelegate delegate = new KeywordExecutorDelegate();
		try {
			globals.rhinoRuntimeCache.execute(delegate, rule.runtimeId, rule.javascript, rule.toString());
		}
		catch (Throwable ex) {
			HistoryEntryBuilder history = prh.makeHistoryEntryBuilder();
			history.addAction(rule, "keyword-failed");
			ln("Keyword filter failed with exception: " + ExceptionUtils.getExMsg(ex));
			ln("   " + ex.getClass().getName());
			ln("   filter: " + rule.filterName);
			if (ex instanceof RhinoException) {
				RhinoException rex = (RhinoException)ex;
				ln("  line " + rex.lineNumber() + ", col " + rex.columnNumber());
			}
			ln("   " + rule);
			if (globals.debugFilters)
				ln(ExceptionUtils.getStackTraceAsString(ex));
			return false;
		}
		if (delegate.rd != null) {
			makeHistoryEntry(rule);
			prh.sendResultData(delegate.rd);
			return true;
		}
		if (globals.debugFlag)
			ln("Keyword rule (" + keyword + ") did not return url nor html: " + rule.filterName);
		return false;
	}

	private void sendRedirect(String url) throws Exception
	{
		if (globals.debugFlag)
			ln("Got keyword redir to " + url);
		HeadersBuilder hb = HeadersBuilder.makeErrorTextPlain(HttpURLConnection.HTTP_MOVED_TEMP, "moved to " + url);
		hb.add("location", url);
		prh.sendResultData(new ResultData(hb.toHeaders(), "Moved to " + url + "\n"));
	}

	private void makeHistoryEntry(KeywordRule rule)
	{
		HistoryEntryBuilder history = globals.historyCache.makeBuilder(keyword + " " + term, null);
		history.addAction(rule, "keyword");
		history.commit();
	}

	class KeywordExecutorDelegate implements ExecutorDelegate<KeywordGlobalObject>
	{
		ResultData rd;

		@NotNull
		public KeywordGlobalObject create(@NotNull ScriptableObject scope,
		                                  @NotNull Scriptable defaultPrototype) throws Exception
		{
			return new KeywordGlobalObject(globals, scope, defaultPrototype);
		}

		public void prepare(@NotNull KeywordGlobalObject callback,
		                    @NotNull Context ctx) throws Exception
		{
			callback.prepareKeywordCallbacks(KeywordExecutor.this, ctx, prh, keyword, term);
		}

		public void executionDone(@NotNull KeywordGlobalObject gb) throws Exception
		{
			rd = gb.getOptResultData();
		}
	}

	public void ln(@Nullable String s)
	{
		prh.ln(s);
	}

	private static String latin1encode(String s)
	{
		int len = s.length();
		StringBuilder sb = new StringBuilder(len + 5);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch == '_' || ch == '-' || ch == '.'
					|| (ch >= '0' && ch <= '9')
					|| (ch >= 'A' && ch <= 'Z')
					|| (ch >= 'a' && ch <= 'z')) {
				sb.append(ch);
			} else if (ch <= 255) {
				sb.append('%');
				sb.append(tohex(ch >> 4));
				sb.append(tohex(ch));
			} else {
				sb.append('?');
			}
		}
		return sb.toString();
	}

	private static char tohex(int i)
	{
		i &= 15;
		return i < 10 ? (char)('0' + i) : (char)('A' + i - 10);
	}
}
