/* Copyright (C) 2009-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.rhino.callback;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.FunctionObject;
import org.mozilla.javascript.ScriptableObject;

public class CallbackRhinoBinder
{
	@NotNull public final Class<? extends ScriptableObject> cls;
	@NotNull public final Set<String> boundNames;

	@NotNull private final List<Binder> binderList;
	@NotNull private final Method[] methods;

	private static final String functionPrefix = "jsFunction_";
	private static final String getterPrefix = "jsGet_";
	private static final String setterPrefix = "jsSet_";

	public CallbackRhinoBinder(@NotNull Class<? extends ScriptableObject> cls)
	{
		this.cls = cls;
		methods = cls.getMethods();
		List<Binder> list = new ArrayList<>();
		Set<String> seenNames = new HashSet<>();
		for (Method m : methods) {
			String methodName = m.getName();
			if (methodName.startsWith(functionPrefix)) {
				String name = methodName.substring(functionPrefix.length());
				if (!checkSeenMethod(seenNames, name))
					continue;
				list.add(new FunctionBinder(name, m));
				continue;
			}
			if (methodName.startsWith(getterPrefix)) {
				String name = methodName.substring(getterPrefix.length());
				if (!checkSeenMethod(seenNames, name))
					continue;
				Method setter = findOptMethod(setterPrefix + name);
				list.add(new GetterSetterBinder(name, m, setter));
			}
		}
		boundNames = Collections.unmodifiableSet(seenNames);
		binderList = Collections.unmodifiableList(list);
	}

	public void bind(@NotNull ScriptableObject callback)
	{
		for (Binder b : binderList)
			b.bind(callback);
	}

	interface Binder
	{
		void bind(@NotNull ScriptableObject callback);
	}

	static class GetterSetterBinder implements Binder
	{
		@NotNull String name;
		@NotNull Method getter;
		@Nullable Method setter;

		GetterSetterBinder(@NotNull String name, @NotNull Method getter, @Nullable Method setter)
		{
			this.name = name;
			this.getter = getter;
			this.setter = setter;
		}

		public void bind(@NotNull ScriptableObject callback)
		{
			//System.err.println("Create getter" + (setter != null ? "+setter" : "") + ":  " + name);
			int attr = ScriptableObject.PERMANENT | ScriptableObject.DONTENUM;
			if (setter == null)
				attr |= ScriptableObject.READONLY;
			callback.defineProperty(name, null, getter, setter, attr);
		}
	}

	static class FunctionBinder implements Binder
	{
		@NotNull String name;
		@NotNull Method method;

		FunctionBinder(@NotNull String name, @NotNull Method method)
		{
			this.name = name;
			this.method = method;
		}

		public void bind(@NotNull ScriptableObject callback)
		{
			FunctionObject f = new FunctionObject(name, method, callback);
			int attrs = ScriptableObject.PERMANENT | ScriptableObject.DONTENUM;
			ScriptableObject.defineProperty(callback, name, f, attrs);
			f.sealObject();
		}
	}

	@Nullable
	private Method findOptMethod(@NotNull String methodName)
	{
		for (Method m : methods) {
			if (m.getName().equals(methodName))
				return m;
		}
		return null;
	}

	private boolean checkSeenMethod(@NotNull Set<String> seen, @NotNull String methodName)
	{
		if (seen.contains(methodName)) {
			String msg = "Duplicate Javascript -> Java method: " + cls.getName() + "." + methodName;
			System.err.println(msg);
			return false;
		}
		seen.add(methodName);
		return true;
	}
}
