/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.rhino.runtime;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import org.glimmerblocker.proxy.globals.GlimmerGlobals;
import org.glimmerblocker.proxy.globals.log.Logger;
import org.glimmerblocker.proxy.globals.log.PrefixedLogger;
import org.glimmerblocker.proxy.rhino.callback.GlimmerGlobalObjectBase;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.ContextFactory;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.ScriptableObject;
import dk.vitality.lib.foundation.misc.KeyedMutexExecutor;
import dk.vitality.lib.foundation.misc.RunnableX;

public class RhinoRuntimeCache implements Logger
{
	@NotNull private final GlimmerGlobals globals;
	@NotNull private final Logger logger;
	private static int previousRuntimeUniqueId = 1000;

	public RhinoRuntimeCache(@NotNull GlimmerGlobals globals) throws Exception
	{
		this.globals = globals;
		this.logger = new PrefixedLogger(globals, "transform-rt-cache");
		synchronized (ContextFactory.class) {
			if (!ContextFactory.hasExplicitGlobal())
				ContextFactory.initGlobal(new GbContextFactory());
		}
	}

	public <CB extends GlimmerGlobalObjectBase> void execute(@NotNull final ExecutorDelegate<CB> delegate,
	                                                         final int runtimeId,
	                                                         @NotNull final String rawScriptText,
	                                                         @NotNull final String runtimeDebugName) throws Exception
	{
		RunnableX r = () -> {
			doExecute(delegate, runtimeId, rawScriptText, runtimeDebugName);
		};
		rhinoMutex.execute("Rhino-" + runtimeId, r);
	}

	private static final KeyedMutexExecutor rhinoMutex = new KeyedMutexExecutor("rhino");

	public void clear()
	{
		synchronized (cache) {
			cache.clear();
		}
	}

	public static int makeNewRuntimeId()
	{
		return ++previousRuntimeUniqueId;
	}

	@NotNull
	private ScriptableObject createScope(Context ctx) throws Exception
	{
		return ctx.initStandardObjects(null, false);
	}

	/**
	 * It seems like a single scope uses 40KB memory.
	 * 25 * 40 KB = 1 MB.
	 * The scopes can't be shared using a sealed super-scope because
	 * RegExp needs to be able to modify RegExp.$1, ... $n when using
	 * "abdefghi".replace(/abc(def)ghi/, "$1");
	 */
	private static final int cacheSize = 25;
	private final LruCache cache = new LruCache();

	public void ln(@Nullable String s)
	{
		logger.ln(s);
	}

	private class LruCache extends LinkedHashMap<Integer, GlimmerGlobalObjectBase>
	{
		private LruCache()
		{
			super(cacheSize);
		}

		@Override
		protected boolean removeEldestEntry(Map.Entry<Integer, GlimmerGlobalObjectBase> eldest)
		{
			boolean flush = size() >= cacheSize;
			if (globals.debugFilters && flush)
				ln("Flushes: " + eldest.getKey());
			return flush;
		}
	}

	void testMemoryUsage() throws Exception
	{
		Context ctx = ContextFactory.getGlobal().enterContext();
		try {
			List<ScriptableObject> list = new ArrayList<>(10000);
			java.lang.Runtime r = java.lang.Runtime.getRuntime();
			for (int i = 0; i < 10; i++) {
				long m1 = r.totalMemory() - r.freeMemory();
				list.add(createScope(ctx));
				long m2 = r.totalMemory() - r.freeMemory();
				System.err.println("New scope: " + m1 + " -> " + m2 + " = " + (m2 - m1) +
						" from " + r.totalMemory() + " - " + r.freeMemory() + " of " + r.maxMemory());
			}
			for (int i = 0; i < 10; i++) {
				long m1 = r.totalMemory() - r.freeMemory();
				for (int x = 0; x < 100; x++)
					list.add(createScope(ctx));
				long m2 = r.totalMemory() - r.freeMemory();
				System.err.println("New 100 scope: " + m1 + " -> " + m2 + " = " + (m2 - m1));
			}
			System.err.println("Num: " + list.size());
		}
		finally {
			Context.exit();
		}
	}


	@SuppressWarnings({"unchecked"})
	private <GB extends GlimmerGlobalObjectBase> void doExecute(@NotNull ExecutorDelegate<GB> delegate,
	                                                            int runtimeId,
	                                                            @NotNull String rawScriptText,
	                                                            @NotNull String runtimeDebugName) throws Exception
	{
		if (globals.debugFilters)
			ln("Executes javascript for: " + runtimeId + " = " + runtimeDebugName);
		long st0 = System.currentTimeMillis();
		Context ctx = ContextFactory.getGlobal().enterContext();
		long st1 = 0, st2 = 0;
		try {
			GB gb;
			synchronized (cache) {
				gb = (GB)cache.get(runtimeId);
			}
			boolean virgin = (gb == null);
			if (virgin) {
				ScriptableObject scope = ctx.initStandardObjects(null, false);
				Scriptable globalObjectPrototype = ctx.newObject(scope).getPrototype();
				gb = delegate.create(scope, globalObjectPrototype);
				if (gb == null)
					throw new NullPointerException("createCallbackObject() returned null.");
				gb.bind();
				if (gb.binder == null)
					throw new Exception("cb.bind() did not bind the java methods.");
				scope.put("gb", scope, gb);
			}
			st1 = System.currentTimeMillis();
			//
			delegate.prepare(gb, ctx);
			//
			st1 = System.currentTimeMillis();
			if (gb.script == null) {
				if (globals.debugFilters)
					ln("Compiles script for: " + runtimeDebugName);
				gb.script = ctx.compileString(rawScriptText, "script", 1, null);
			}
			if (virgin) {
				synchronized (cache) {
					cache.put(runtimeId, gb);
				}
			}
			//
			st2 = System.currentTimeMillis();
			try {
				gb.script.exec(ctx, gb.scope);
				delegate.executionDone(gb);
			}
			finally {
				gb.notifyLeftContext();
			}
		}
		finally {
			Context.exit();
			if (globals.debugFilters) {
				long st3 = System.currentTimeMillis();
				ln("Time:" +
						" scope=" + (st1 - st0) +
						" + compile=" + (st2 - st1) +
						" + exec=" + (st3 - st2) +
						" = " + (st3 - st0) + " msecs for " + runtimeId + " = " + runtimeDebugName);
			}
		}
	}
}
