/* Copyright (C) 2008-2015 Peter Speck
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.glimmerblocker.proxy.util;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public class HtmlUtils
{
	@NotNull
	public static String encodeHtml(@Nullable String txt)
	{
		if (txt == null)
			return "";
		if (!needHtmlEncode(txt))
			return txt;
		int len = txt.length();
		StringBuilder enc = new StringBuilder(len * 2);
		for (int i = 0; i < len; i++) {
			char ch = txt.charAt(i);
			switch (ch) {
				case '&':
					enc.append("&amp;");
					break;
				case '<':
					enc.append("&lt;");
					break;
				case '>':
					enc.append("&gt;");
					break;
				case '\n':
					enc.append("<br>\n");
					break;
				default:
					if (ch > 255) {
						enc.append('&');
						enc.append('#');
						enc.append(Integer.toString(ch));
						enc.append(';');
					} else {
						enc.append(ch);
					}
					break;
			}
		}
		return enc.toString();
	}

	private static boolean needHtmlEncode(@NotNull String s)
	{
		int len = s.length();
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			switch (ch) {
				case '\n':
				case '&':
				case '<':
				case '>':
					return true;
				default:
					if (ch > 255)
						return true;
			}
		}
		return false;
	}

	// Compared to java.net.URLEncoder.encode:
	// It doesn't encode with the platform encoding, aka doesn't trash
	// URLs with 8-bit chars....
	public static String encodeURL(String s)
	{
		if (s == null)
			return "";
		if (!needUrlEncode(s))
			return s;
		int len = s.length();
		StringBuilder sb = new StringBuilder(len + 25);
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch == '_' || ch == '-' || ch == '.'
					|| (ch >= '0' && ch <= '9')
					|| (ch >= 'A' && ch <= 'Z')
					|| (ch >= 'a' && ch <= 'z')) {
				sb.append(ch);
			} else if (ch <= 255) {
				sb.append('%');
				sb.append(hex[(ch >> 4) & 15]);
				sb.append(hex[ch & 15]);
			} else {
				sb.append("%u");
				sb.append(hex[(ch >> 12) & 15]);
				sb.append(hex[(ch >> 8) & 15]);
				sb.append(hex[(ch >> 4) & 15]);
				sb.append(hex[ch & 15]);
			}
		}
		return sb.toString();
	}

	private static final char[] hex = {
			'0', '1', '2', '3', '4', '5', '6', '7',
			'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

	private static boolean needUrlEncode(String s)
	{
		int len = s.length();
		for (int i = 0; i < len; i++) {
			char ch = s.charAt(i);
			if (ch != '_' && ch != '-' && ch != '.'
					&& (ch < '0' || ch > '9')
					&& (ch < 'A' || ch > 'Z')
					&& (ch < 'a' || ch > 'z')) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Clean a string so it is suitable as a safe client filename: ensure it
	 * <ul>
	 * <li>doesn't contain slashes or colons</li>
	 * <li>start with a dot</li>
	 * <li>contains ascii control characters</li>
	 * <li>is the empty string</li>
	 * </ul>
	 */
	@NotNull
	public static String cleanFilename(@NotNull String title)
	{
		StringBuilder sb = new StringBuilder(title.length());
		for (int i = 0; i < title.length(); i++) {
			char ch = title.charAt(i);
			if (ch < 32)
				continue;
			if (ch == '.' && sb.length() == 0)
				continue;
			if (ch == '/' || ch == ':')
				continue;
			sb.append(ch);
		}
		if (sb.length() == 0)
			sb.append("without-name");
		return sb.toString();
	}
}
